 /*
 *Driver for the card pcie bridge.

 *Copyright (C) 2019 AVerMedia TECHNOLOGIES, Inc.

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#include <linux/kernel.h>
#include "cxt_mgr.h"
#include "framegrabber.h"
#include "mem_model.h"
#include "v4l2_model.h"
#include "board_v4l2.h"
#include "debug.h"
#include "pci_model.h"
#include "i2c_model.h"
#include "task_model.h"
#include "fpga_xilinx.h"
#include "adv7619.h"
#include "board_alsa.h"

static int cnt_retry=0;
enum cx311h_input_e
{
	CX311H_HDMI_INPUT=0,
	CX311H_INPUT_COUNT,
};

enum cx311h_i2c_chip_e
{
    CX311H_I2C_CHIP_ADV7619_0,
    CX311H_I2C_CHIP_COUNT,
};
typedef struct
{
	BASIC_CXT_HANDLE_DECLARE;
    framegrabber_handle_t fg_handle;
    v4l2_model_handle_t v4l2_handle;
    task_model_handle_t task_model_handle;
    handle_t fpga_xilinx_handle;
    handle_t i2c_chip_handle[CX311H_I2C_CHIP_COUNT];
    adv7619_frameinfo_t cur_fe_frameinfo;
    enum adv7619_audio_sample cur_fe_audiorate;
    int cur_bchs_value;
    int cur_bchs_selection;
    task_handle_t check_signal_task_handle;
}board_v4l2_context_t;

typedef struct
{
    const char *name;
    int index;
}board_chip_desc_t;

static board_chip_desc_t cx311h_chip_desc[CX311H_I2C_CHIP_COUNT]=
{
	[CX311H_I2C_CHIP_ADV7619_0]=
	{
		.name=ADV7619_DRVNAME,
		.index=1,
    },
};

static framegrabber_setup_input_info_t cx311h_input_info[] =
{
    [CX311H_HDMI_INPUT] =
    {
        .name = "HDMI",
        .support_framesize_info =
        {
			[FRAMESIZE_640x480] = REFRESHRATE_60_BIT,
            [FRAMESIZE_720x480] = REFRESHRATE_60_BIT,
            [FRAMESIZE_720x576] = REFRESHRATE_50_BIT,
            [FRAMESIZE_1280x720] = REFRESHRATE_50_BIT \
                                    | REFRESHRATE_60_BIT,
            [FRAMESIZE_1920x1080] = REFRESHRATE_60_BIT \
                                    | REFRESHRATE_50_BIT \
                                    | REFRESHRATE_30_BIT \
                                    | REFRESHRATE_25_BIT \
                                    | REFRESHRATE_24_BIT,
            [FRAMESIZE_3840x2160] = REFRESHRATE_30_BIT, 
            [FRAMESIZE_4096x2160] = REFRESHRATE_30_BIT,                       
        },
    },
    [CX311H_INPUT_COUNT] =
    {
		.name = NULL,
	},
};

static framegrabber_property_t  cx311h_property=
{
	.name="CX311H",
	.input_setup_info=cx311h_input_info,
	.support_out_pixfmt_mask=FRAMEGRABBER_PIXFMT_BITMSK,
	//.max_supported_line_width=3840,
	//.max_supported_line_width=4096,
    .max_frame_size=4096 *2160,

};

static v4l2_model_device_setup_t device_info=
{
    .type = DEVICE_TYPE_GRABBER,
    .driver_name="CX311H",
    .card_name="CN311H",
    .capabilities= V4L2_MODEL_CAPS_CAPTURE_BIT | V4L2_MODEL_CAPS_READWRITE_BIT | V4L2_MODEL_CAPS_STREAMING_BIT,
    .buffer_type=V4L2_MODEL_BUF_TYPE_DMA_SG,
};

static void *board_v4l2_alloc(void);
static void board_v4l2_release(void *cxt);
static void cx311h_v4l2_stream_on(v4l2_model_callback_parameter_t *cb_info);
static void cx311h_v4l2_stream_off(v4l2_model_callback_parameter_t *cb_info);
static void cx311h_stream_on(framegrabber_handle_t handle);
static void cx311h_stream_off(framegrabber_handle_t handle);
//static void cx311h_bchs_read(framegrabber_handle_t handle);
static void cx311h_bchs_write(framegrabber_handle_t handle);
static void cx311h_brightness_read(framegrabber_handle_t handle,int *brightness);
static void cx311h_contrast_read(framegrabber_handle_t handle,int *contrast);
static void cx311h_hue_read(framegrabber_handle_t handle,int *hue);
static void cx311h_saturation_read(framegrabber_handle_t handle,int *saturation);

static int cx311h_flash_dump(framegrabber_handle_t handle,int start_block, int blocks, U8_T *flash_dump); 
static int cx311h_flash_update(framegrabber_handle_t handle,int start_block, int blocks, U8_T *flash_dump); 

static int cx311h_edid_get(framegrabber_handle_t handle, U8_T* val, int* size);
static int cx311h_edid_set(framegrabber_handle_t handle, U8_T* val, int size);

static framegrabber_interface_t cx311h_ops={
    .stream_on      = cx311h_stream_on,
    .stream_off     = cx311h_stream_off,
    //.bchs_get     = cx311h_bchs_read,
    .brightness_get = cx311h_brightness_read,
    .contrast_get   = cx311h_contrast_read,
    .hue_get        = cx311h_hue_read,
    .saturation_get = cx311h_saturation_read,
    .bchs_set       = cx311h_bchs_write,
    .flash_read     = cx311h_flash_dump,
    .flash_update   = cx311h_flash_update,
    .edid_val_get	= cx311h_edid_get,
    .edid_val_set	= cx311h_edid_set
};

static void *board_v4l2_alloc()
{
	board_v4l2_context_t *cxt;
	cxt=mem_model_alloc_buffer(sizeof(*cxt));

	return cxt;
}

static void board_v4l2_release(void *cxt)
{
	board_v4l2_context_t *board_v4l2_cxt=cxt;

	if(board_v4l2_cxt)
	{
		if(board_v4l2_cxt->check_signal_task_handle)
        {
            task_model_release_task(board_v4l2_cxt->task_model_handle,board_v4l2_cxt->check_signal_task_handle);
            board_v4l2_cxt->check_signal_task_handle=NULL;
        }
        cxt_manager_unref_context(board_v4l2_cxt->task_model_handle);
        cxt_manager_unref_context(board_v4l2_cxt->fpga_xilinx_handle);
        mem_model_free_buffer(board_v4l2_cxt);
	}
}

static void cx311h_v4l2_stream_on(v4l2_model_callback_parameter_t *cb_info)
{
    board_v4l2_context_t *board_v4l2_cxt=cb_info->asso_data;
    
    debug_msg("%s %p\n",__func__,board_v4l2_cxt);

}

static void cx311h_v4l2_stream_off(v4l2_model_callback_parameter_t *cb_info)
{
    board_v4l2_context_t *board_v4l2_cxt=cb_info->asso_data;
    debug_msg("%s %p\n",__func__,board_v4l2_cxt);
    
}
static void cx311h_stream_on(framegrabber_handle_t handle)
{
    board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
    fpga_xilinx_video_process_cfg_t vip_cfg;
    int width,height;
    int out_width,out_height;
    int frameinterval=0;
    int audio_rate=0;
    int dual_pixel=0;
    int ioctl_bypass_disable=0;
    int yuv422=0;
    framegrabber_framemode_e framemode;
    const framegrabber_pixfmt_t *pixfmt=framegrabber_g_out_pixelfmt(handle);
    
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0]; 
    enum adv7619_audio_sample fe_audiorate=0;
    adv7619_frameinfo_t *fe_frameinfo=&board_v4l2_cxt->cur_fe_frameinfo; 
    
    //fpga_xilinx_frame_info_t detected_frameinfo;
    
    //fpga_xilinx_get_frameinfo(board_v4l2_cxt->fpga_xilinx_handle,&detected_frameinfo);
    framemode=framegrabber_g_input_framemode(board_v4l2_cxt->fg_handle);
    framegrabber_g_input_framesize(board_v4l2_cxt->fg_handle,&width,&height);

    //frameinterval=framegrabber_g_input_framerate(board_v4l2_cxt->fg_handle);
    frameinterval=framegrabber_g_out_framerate(board_v4l2_cxt->fg_handle); //one from ap setting
    if (frameinterval ==0)
    {
		frameinterval=framegrabber_g_input_framerate(board_v4l2_cxt->fg_handle);
	}
    
    framegrabber_g_out_framesize(board_v4l2_cxt->fg_handle,&out_width,&out_height);

    if ((width!=out_width) || (height!=out_height))
    {
		printk("w x h=%d x %d; out_w x out_h=%d x %d\n",width,height,out_width,out_height);
		printk(">>>%s only support bypass mode\n",__func__);
		//fpga_xilinx_enable_video_streaming(board_v4l2_cxt->fpga_xilinx_handle,FALSE);
		//return;
    }
 
    if(framemode==FRAMEGRABBER_FRAMEMODE_INTERLACED)
        vip_cfg.is_interlace=TRUE;
    else
        vip_cfg.is_interlace=FALSE;
        
    //audio_rate = framegrabber_g_input_audioinfo(board_v4l2_cxt->fg_handle);
    
    adv7619_get_audiorate(adv7619_handle,&fe_audiorate);
    
    adv7619_get_yuvmode(adv7619_handle,&yuv422);
    
    dual_pixel = framegrabber_g_input_dualmode(board_v4l2_cxt->fg_handle);
    
    //fpga_xilinx_dual_pixel(board_v4l2_cxt->fpga_xilinx_handle,dual_pixel);
    if (dual_pixel ==1)
        vip_cfg.dual_pixel = 1;
    else
        vip_cfg.dual_pixel = 0;   
    
    audio_rate = fe_audiorate;
    
    framegrabber_s_input_audiorate(board_v4l2_cxt->fg_handle,audio_rate);
    switch (audio_rate)
    {
		case ADV7619_AUDIO_RATE_32000:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_32000;
		break;
		case ADV7619_AUDIO_RATE_44100:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_44100;
		break;
		case ADV7619_AUDIO_RATE_48000:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_48000;
		break;
		case ADV7619_AUDIO_RATE_88200:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_88200;
		break;
		case ADV7619_AUDIO_RATE_96000:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_96000;
		break;
		case ADV7619_AUDIO_RATE_176400:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_176400;
		break;
		case ADV7619_AUDIO_RATE_192000:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_192000;
		break;
		default:
		vip_cfg.audio_rate=FPGA_XILINX_AUDIO_44100;
		break;
	}
    printk("%s...vip_cfg.audio_rate_index=%d\n",__func__,vip_cfg.audio_rate);
    adv7619_get_frameinfo(adv7619_handle,fe_frameinfo);
    
    vip_cfg.in_framerate = framegrabber_g_input_framerate(board_v4l2_cxt->fg_handle);
    
    if (vip_cfg.is_interlace==TRUE)
        vip_cfg.in_framerate = vip_cfg.in_framerate *2; //same as fpga detect
    else
        vip_cfg.in_framerate = vip_cfg.in_framerate;
    
    vip_cfg.YUV422_flag = yuv422;    
    vip_cfg.packet_colorspace = fe_frameinfo->packet_colorspace;    
    vip_cfg.in_framesize.width=width;
    vip_cfg.in_framesize.height=height;
    vip_cfg.out_framesize.width=out_width;
    vip_cfg.out_framesize.height=out_height;
    vip_cfg.clip_start.x=0;
    vip_cfg.clip_start.y=0;
    vip_cfg.clip_size.width=width;
    vip_cfg.clip_size.height=height;
//  vip_cfg.valid_mask=CLIP_CFG_VALID_MASK|SCALER_CFG_VALID_MASK;
//  vip_cfg.valid_mask=SCALER_CFG_VALID_MASK|SCALER_CFG_SHRINK_MASK;
    vip_cfg.valid_mask=SCALER_CFG_VALID_MASK|FRAMERATE_CFG_VALID_MASK;
    vip_cfg.pixel_clock=board_v4l2_cxt->cur_fe_frameinfo.pixel_clock;
    //vip_cfg.out_framerate= 0; //for test
    vip_cfg.out_framerate=frameinterval;
    printk("%s...vip_cfg.in_framerate=%d  vip_cfg.out_framerate=%d pixel_clock=%d\n",__func__,vip_cfg.in_framerate,vip_cfg.out_framerate,vip_cfg.pixel_clock);
    
    ioctl_bypass_disable = framegrabber_g_fpga_bypassmode(board_v4l2_cxt->fg_handle);
    
    vip_cfg.ioctl_YUYV = framegrabber_g_output_pixelformat(board_v4l2_cxt->fg_handle);
    
    #ifndef CONFIG_SONY
    ioctl_bypass_disable =1;
    vip_cfg.ioctl_YUYV =1;
    #endif
    
    printk("ioctl_bypass_disable=%d\n",ioctl_bypass_disable);
    
    if (1 == ioctl_bypass_disable) //control by ioctl
    {
        //enable video bypass
        if (((vip_cfg.in_framesize.width == 4096) && (vip_cfg.out_framesize.width == 4096)
        && (vip_cfg.out_framesize.height == 2160) && (vip_cfg.out_framesize.height == 2160)) || //4096x2160
        ((vip_cfg.in_framesize.width == 3840) && (vip_cfg.out_framesize.width == 3840)
        && (vip_cfg.out_framesize.height == 2160) && (vip_cfg.out_framesize.height == 2160)) || //3840x2160
        ((vip_cfg.in_framesize.width == 2560) && (vip_cfg.out_framesize.width == 2560)  
        && (vip_cfg.in_framesize.height == 1600) && (vip_cfg.out_framesize.height == 1600)) || //2560x1600
        ((vip_cfg.in_framesize.width == 2560) && (vip_cfg.out_framesize.width == 2560) 
        && (vip_cfg.in_framesize.height == 1440) && (vip_cfg.out_framesize.height == 1440)) || //2560x1440
        ((vip_cfg.in_framesize.width == 2560) && (vip_cfg.out_framesize.width == 2560) 
        && (vip_cfg.in_framesize.height == 1080) && (vip_cfg.out_framesize.height == 1080)) ||  //2560x1080
        ((vip_cfg.in_framesize.width == 1920) && (vip_cfg.out_framesize.width == 1920) 
        && (vip_cfg.in_framesize.height == 1440) && (vip_cfg.out_framesize.height == 1440)) || //1920x1440
        ((vip_cfg.in_framesize.width == 1856) && (vip_cfg.out_framesize.width == 1856)
        && (vip_cfg.in_framesize.height == 1392) && (vip_cfg.out_framesize.height == 1392)) || //1856x1392
        ((vip_cfg.in_framesize.width == 1792) && (vip_cfg.out_framesize.width == 1792)
        && (vip_cfg.in_framesize.height == 1344) && (vip_cfg.out_framesize.height == 1344)) || //1792x1344
        ((vip_cfg.in_framesize.width == 2048) && (vip_cfg.out_framesize.width == 2048)
        && (vip_cfg.in_framesize.height == 1152) && (vip_cfg.out_framesize.height == 1152)) || //2048x1152
        ((vip_cfg.in_framesize.width == 1920) && (vip_cfg.out_framesize.width == 1920)
        && (vip_cfg.in_framesize.height == 1200) && (vip_cfg.out_framesize.height == 1200)) || //1920x1200
        ((vip_cfg.in_framesize.width == 1920) && (vip_cfg.out_framesize.width == 1920)
        && (vip_cfg.in_framesize.height == 1080) && (vip_cfg.out_framesize.height == 1080)
        && (vip_cfg.in_framerate == 120)))
        {
            vip_cfg.video_bypass = 1;
	    }
	    else
	    {
		    vip_cfg.video_bypass = 0;
	    }
	}
	else
	{
		vip_cfg.out_framerate = vip_cfg.in_framerate;
		if (vip_cfg.is_interlace==TRUE)
		{
			printk("===ioctl enable video bypass mode and interlaced===\n");
		    vip_cfg.video_bypass = 0;
		}
		else
		{
			printk("===ioctl enable video bypass mode and progressive===\n");
	        vip_cfg.video_bypass = 1;
		}
	}
	
    switch(pixfmt->pixfmt_out)
    {
        case FPGA_XILINX_FMT_YUYV:
        case FPGA_XILINX_FMT_UYVY:
        case FPGA_XILINX_FMT_YVYU:
        case FPGA_XILINX_FMT_VYUY:
            vip_cfg.pixel_format=FPGA_XILINX_FMT_YUYV;
            printk("%s...vip_cfg.pixel_format_output=FPGA_XILINX_FMT_YUYV\n",__func__);
            break;
        case FPGA_XILINX_FMT_RGBP: //RGB565
        case FPGA_XILINX_FMT_RGBR: //RGB565X
        case FPGA_XILINX_FMT_RGBO:
        case FPGA_XILINX_FMT_RGBQ:
        case FPGA_XILINX_FMT_RGB3: //RGB24 4:4:4
        case FPGA_XILINX_FMT_BGR3: 
        case FPGA_XILINX_FMT_RGB4:
        case FPGA_XILINX_FMT_BGR4:
            vip_cfg.pixel_format=FPGA_XILINX_FMT_RGB3;
            printk("%s...vip_cfg.pixel_format=FPGA_XILINX_FMT_RGB3\n",__func__);
            break;
        default:
            vip_cfg.pixel_format=FPGA_XILINX_CS_YUV422;
            break;
    }       
    fpga_xilinx_config_video_process(board_v4l2_cxt->fpga_xilinx_handle,&vip_cfg);
    
    fpga_xilinx_enable_video_streaming(board_v4l2_cxt->fpga_xilinx_handle,TRUE);
   
    //adv7619_set_freerun_screen(adv7619_handle,FALSE);    
}

static void cx311h_stream_off(framegrabber_handle_t handle)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
	
	//adv7619_set_freerun_screen(adv7619_handle,TRUE);
	
    fpga_xilinx_enable_video_streaming(board_v4l2_cxt->fpga_xilinx_handle,FALSE);
}

static int cx311h_flash_dump(framegrabber_handle_t handle,int start_block, int blocks, U8_T *flash_dump) //
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
	int ret=0;
		
    ret = fpga_xilinx_flash_dump(board_v4l2_cxt->fpga_xilinx_handle,start_block,blocks,flash_dump);
    
    return ret;
}

static int cx311h_flash_update(framegrabber_handle_t handle,int start_block, int blocks, U8_T *flash_update)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
	int ret=0;
	
    ret = fpga_xilinx_flash_update(board_v4l2_cxt->fpga_xilinx_handle,start_block,blocks,flash_update);
    
    return ret;
}

static int cx311h_edid_get(framegrabber_handle_t handle, U8_T* val, int* size)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0];
	return adv7619_get_video_time_list(adv7619_handle, val, size);
}


static int cx311h_edid_set(framegrabber_handle_t handle, U8_T* val, int size)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0];
    printk("%s run!\n", __func__);
	return adv7619_set_video_time_list(adv7619_handle, val, size);
}

static void cx311h_brightness_read(framegrabber_handle_t handle,int *brightness)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);//
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0]; 
   
	adv7619_get_brightness(adv7619_handle,&board_v4l2_cxt->cur_bchs_value); 
	
	framegrabber_set_data(handle,board_v4l2_cxt);
	brightness = &board_v4l2_cxt->cur_bchs_value;
	handle->fg_bchs_value = *brightness;
	//printk("%s get brightness=%x\n",__func__,board_v4l2_cxt->cur_bchs_value);
}
static void cx311h_contrast_read(framegrabber_handle_t handle,int *contrast)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0]; 
   
	adv7619_get_contrast(adv7619_handle,&board_v4l2_cxt->cur_bchs_value); 
	
	framegrabber_set_data(handle,board_v4l2_cxt);
	contrast = &board_v4l2_cxt->cur_bchs_value;
	handle->fg_bchs_value = *contrast;
	//printk("%s get contrast=%x\n",__func__,board_v4l2_cxt->cur_bchs_value);
}

static void cx311h_hue_read(framegrabber_handle_t handle,int *hue)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0]; 
   
	adv7619_get_hue(adv7619_handle,&board_v4l2_cxt->cur_bchs_value); 
	
	framegrabber_set_data(handle,board_v4l2_cxt);
	hue = &board_v4l2_cxt->cur_bchs_value;
	handle->fg_bchs_value = *hue;
	//printk("%s get hue=%x\n",__func__,board_v4l2_cxt->cur_bchs_value);
}

static void cx311h_saturation_read(framegrabber_handle_t handle,int *saturation)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0]; 
   
	adv7619_get_saturation(adv7619_handle,&board_v4l2_cxt->cur_bchs_value); 
	
	framegrabber_set_data(handle,board_v4l2_cxt);
	saturation = &board_v4l2_cxt->cur_bchs_value;
	handle->fg_bchs_value = *saturation;
	//printk("%s get saturation=%x\n",__func__,board_v4l2_cxt->cur_bchs_value);
}

static void cx311h_bchs_write(framegrabber_handle_t handle)
{
	board_v4l2_context_t *board_v4l2_cxt=framegrabber_get_data(handle);
	handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0]; 
   
	int bchs_value = handle->fg_bchs_value;
	int bchs_select = handle->fg_bchs_selection;	
		
	adv7619_set_bchs(adv7619_handle,&bchs_value,&bchs_select); 
}

static void cx311h_video_buffer_done(void *data)
{
    board_v4l2_context_t *board_v4l2_cxt=data;
 
    if(board_v4l2_cxt)
    {
        v4l2_model_buffer_done(board_v4l2_cxt->v4l2_handle);
    }
}

static void cx311h_v4l2_buffer_prepare(v4l2_model_callback_parameter_t *cb_info)
{
    board_v4l2_context_t *board_v4l2_cxt=cb_info->asso_data;
    v4l2_model_buffer_info_t *buffer_info=cb_info->u.buffer_prepare_info.buffer_info;
 
    if(buffer_info)
    {
        int i;
        v4l2_model_buf_desc_t *desc;
        unsigned framebufsize,remain;
        int width,height;
        unsigned bytesperline;
        
        framegrabber_g_out_framesize(board_v4l2_cxt->fg_handle,&width,&height);
        bytesperline=framegrabber_g_out_bytesperline(board_v4l2_cxt->fg_handle);
        framebufsize=bytesperline*height; 
        //printk("%s %dx%d framesize %u\n",__func__,width,height,framebufsize);
   
        for(i=0,desc=buffer_info->buf_info[0],remain=framebufsize;i<buffer_info->buf_count[0];i++)
        {
            if(remain >= desc[i].size)
            {
                fpga_xilinx_add_to_cur_desclist(board_v4l2_cxt->fpga_xilinx_handle,desc[i].addr,desc[i].size);
                remain -= desc[i].size;
            }
            else
            {
                if(remain != 0) {
                    fpga_xilinx_add_to_cur_desclist(board_v4l2_cxt->fpga_xilinx_handle,desc[i].addr,remain);
                    remain =0;
                }
                break;
            }
        }
        fpga_xilinx_active_current_desclist(board_v4l2_cxt->fpga_xilinx_handle,cx311h_video_buffer_done,board_v4l2_cxt);   
    }  
}

static void cx311h_v4l2_buffer_init(v4l2_model_callback_parameter_t *cb_info)
{
    v4l2_model_buffer_info_t *buffer_info=cb_info->u.buffer_prepare_info.buffer_info;
  
    if(buffer_info)
    {
        int i;
        v4l2_model_buf_desc_t *desc;
        
        for(i=0,desc=buffer_info->buf_info[0];i<buffer_info->buf_count[0];i++)
        {
            //printk("addr %08x size %x\n",desc[i].addr,desc[i].size);
        }
    }  
}

static void check_signal_stable_task(void *data)
{
    board_v4l2_context_t *board_v4l2_cxt=data;
    int width,height;
    int dual_pixel;
    int dual_pixel_set;
    framegrabber_framemode_e framemode;
    fpga_xilinx_frame_info_t detected_frameinfo;
    //adv7619_frameinfo_t *fe_frameinfo=&board_v4l2_cxt->cur_fe_frameinfo; 
    BOOL_T is_interlace;
    int stable_check;
    
    debug_msg("%s\n",__func__);
        
    fpga_xilinx_get_frameinfo(board_v4l2_cxt->fpga_xilinx_handle,&detected_frameinfo,&stable_check); 
    framemode=framegrabber_g_input_framemode(board_v4l2_cxt->fg_handle);
    is_interlace=(framemode==FRAMEGRABBER_FRAMEMODE_INTERLACED) ? TRUE : FALSE;
    framegrabber_g_input_framesize(board_v4l2_cxt->fg_handle,&width,&height);
    dual_pixel = framegrabber_g_input_dualmode(board_v4l2_cxt->fg_handle);
    
    if (dual_pixel ==1)
        dual_pixel_set = detected_frameinfo.width*2;
    else
        dual_pixel_set = detected_frameinfo.width; 
#if 0  
    if((((detected_frameinfo.width<320 || detected_frameinfo.height<240)) || (width !=/*detected_frameinfo.width*/dual_pixel_set)|| (height !=detected_frameinfo.height)) && (cnt_retry<2))
    {
		cnt_retry++;
		//printk("%s retry...in %dx%d  get %dx%d\n",__func__,width,height,dual_pixel_set,detected_frameinfo.height);
        task_model_run_task_after(board_v4l2_cxt->task_model_handle,board_v4l2_cxt->check_signal_task_handle,50000);
        return;
    }
#endif  
    printk("%s stable_check=%d\n",__func__,stable_check);

    if ((stable_check == 0) && (cnt_retry<6))
    {
		cnt_retry++;
		task_model_run_task_after(board_v4l2_cxt->task_model_handle,board_v4l2_cxt->check_signal_task_handle,50000);
        return;
	} 
    printk("%s fpga detected size %dx%d\n",__func__,detected_frameinfo.width,detected_frameinfo.height);
#if 0   
    if (fe_frameinfo->is_interlace)
    {
		detected_frameinfo.width = fe_frameinfo->width;
		detected_frameinfo.height = fe_frameinfo->height/2;
	}
	else
	{
		detected_frameinfo.width = fe_frameinfo->width;
		detected_frameinfo.height = fe_frameinfo->height;
	}
      
    if(width==detected_frameinfo.width && height==detected_frameinfo.height && (is_interlace==detected_frameinfo.is_interlace))
    {
		printk("%s fix input %dx%d to fpga %dx%d\n",__func__,width,height,detected_frameinfo.width,detected_frameinfo.height);
        framegrabber_mask_s_status(board_v4l2_cxt->fg_handle,FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT,FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT);
        cnt_retry=0;
    }
    else
    {
        printk("%s input %dx%d match to fpga %dx%d\n",__func__,width,height,detected_frameinfo.width,detected_frameinfo.height);
      
        width=detected_frameinfo.width;
        height=detected_frameinfo.height;
        
        framegrabber_s_input_framesize(board_v4l2_cxt->fg_handle,width,height);
        if(is_interlace!=detected_frameinfo.is_interlace)
        {
            framemode=(detected_frameinfo.is_interlace) ? FRAMEGRABBER_FRAMEMODE_INTERLACED:FRAMEGRABBER_FRAMEMODE_PROGRESS;
            framegrabber_s_input_framemode(board_v4l2_cxt->fg_handle,framemode);
        }
        task_model_run_task_after(board_v4l2_cxt->task_model_handle,board_v4l2_cxt->check_signal_task_handle,50000); // run again to comfirm
    }
#else
        printk("%s fix input %dx%d to fpga %dx%d\n",__func__,width,height,detected_frameinfo.width,detected_frameinfo.height);
        //framegrabber_s_input_framesize(board_v4l2_cxt->fg_handle,width,height);
        framegrabber_mask_s_status(board_v4l2_cxt->fg_handle,FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT,FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT);
        cnt_retry=0;

#endif   
}

static void cx311h_adv7619_event(void *cxt,adv7619_event_e event) 
{
    board_v4l2_context_t *board_v4l2_cxt=cxt;
    adv7619_frameinfo_t *fe_frameinfo=&board_v4l2_cxt->cur_fe_frameinfo; 
    enum adv7619_audio_sample *fe_audiorate=&board_v4l2_cxt->cur_fe_audiorate; 
    handle_t adv7619_handle=board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0]; 
    adv7619_out_format_e out_format;
    unsigned int hdcp_flag;
    int ITU_flag;
    U8_T adv7619_infoframe[INFOFRAME_MAXSIZE];
    
    //printk("%s event %d\n",__func__,event);
    
    switch(event)
    {
		case ADV7619_LOCK:
		{
			printk("ADV7619_LOCK\n");
			//fpga_xilinx_get_frameinfo(fpga_xilinx_handle,&detected_frameinfo,fe_frameinfo->pixel_clock); 
        
			adv7619_get_aviinfoframe(adv7619_handle,adv7619_infoframe);
			
			framegrabber_s_input_aviinfoframe(board_v4l2_cxt->fg_handle,adv7619_infoframe);
			
			adv7619_get_vendorinfoframe(adv7619_handle,adv7619_infoframe);
			
			framegrabber_s_input_vendorinfoframe(board_v4l2_cxt->fg_handle,adv7619_infoframe);
			
			adv7619_get_audioinfoframe(adv7619_handle,adv7619_infoframe);
			
			framegrabber_s_input_audioinfoframe(board_v4l2_cxt->fg_handle,adv7619_infoframe);		
        
            adv7619_get_frameinfo(adv7619_handle,fe_frameinfo);
   
            //work around for ADV7619 detect issue      
            if ((fe_frameinfo->is_interlace==0) && ((fe_frameinfo->height == 240) || (fe_frameinfo->height == 288)))
            {
				printk("ADV7619 height adjustment\n");
				fe_frameinfo->is_interlace =1;
				fe_frameinfo->height += fe_frameinfo->height;	
				fe_frameinfo->framerate /= 2;
			} 
			if ((fe_frameinfo->is_interlace==1) && ((fe_frameinfo->height == 480) || (fe_frameinfo->height == 576)))
            {
				printk("ADV7619 framerate adjustment\n");
				
				fe_frameinfo->framerate /= 2;
			} 
			
			hdcp_flag = fe_frameinfo->adv7619_hdcp_flag;
			
			adv7619_get_audiorate(adv7619_handle,fe_audiorate);
			
			framegrabber_s_input_audiorate(board_v4l2_cxt->fg_handle,*fe_audiorate);
			
			if (fe_frameinfo->is_interlace==1)
			{
                framegrabber_s_input_framerate(board_v4l2_cxt->fg_handle,fe_frameinfo->framerate,2*fe_frameinfo->denominator);
                printk("frameinfo->denominator(interlaced) = %d\n",fe_frameinfo->denominator);
			}
			else
			{
				framegrabber_s_input_framerate(board_v4l2_cxt->fg_handle,fe_frameinfo->framerate,fe_frameinfo->denominator);
				printk("fe_frameinfo->denominator(progressive) = %d\n",fe_frameinfo->denominator);
			}
            
            framegrabber_s_input_interlace(board_v4l2_cxt->fg_handle,fe_frameinfo->is_interlace);
            
            framegrabber_s_input_framesize(board_v4l2_cxt->fg_handle,fe_frameinfo->width,fe_frameinfo->height);
            
            framegrabber_s_input_dualmode(board_v4l2_cxt->fg_handle,fe_frameinfo->dual_pixel); 
            
            framegrabber_s_hdcp_flag(board_v4l2_cxt->fg_handle,hdcp_flag);
            
            if ((fe_frameinfo->is_ITU601==1) && (fe_frameinfo->is_ITU709==0))
                ITU_flag =1;
            else if ((fe_frameinfo->is_ITU601==0) && (fe_frameinfo->is_ITU709==1))
                ITU_flag =2;
            else  
                ITU_flag =0;  
            
            framegrabber_s_input_ITU(board_v4l2_cxt->fg_handle,ITU_flag);
                    
            
            //fpga_xilinx_dual_pixel(board_v4l2_cxt->fpga_xilinx_handle,fe_frameinfo->dual_pixel);
     
            if((fe_frameinfo->pixel_clock>170000000) || (fe_frameinfo->dual_pixel_like ==1))
            {
                //adv7619_out_format_e out_format;
                if((fe_frameinfo->is_yuv422==1) && (fe_frameinfo->is_yuv444==0))
                    out_format=ADV7619_OUT_FORMAT_SDR_422_2X24_INTERLEAVE_MODE0;
                else if((fe_frameinfo->is_yuv444==1) && (fe_frameinfo->is_yuv422==0))
                    out_format=ADV7619_OUT_FORMAT_SDR_444_2X24_INTERLEAVE_MODE0;
                else
                    out_format=ADV7619_OUT_FORMAT_SDR_422_2X24_INTERLEAVE_MODE0;   
                adv7619_set_out_format(adv7619_handle,out_format);
                //adv7619_set_out_format(board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_1],out_format);
            }
            else
            {
				if((fe_frameinfo->is_yuv422==1) && (fe_frameinfo->is_yuv444==0))
                    out_format=ADV7619_OUT_FORMAT_SDR_ITU656_24_MODE0;
                else if((fe_frameinfo->is_yuv444==1) && (fe_frameinfo->is_yuv422==0))
                    out_format=ADV7619_OUT_FORMAT_SDR_444_36_MODE0;
                else
                    out_format=ADV7619_OUT_FORMAT_SDR_ITU656_24_MODE0;   
                adv7619_set_out_format(adv7619_handle,out_format);
            }
            //printk("=========== adv7619_set_out_format=%02x\n",out_format);
        
			//debug_msg("pixelclock %d detected %dx%d%c\n",fe_frameinfo->pixel_clock,detected_frameinfo.width,detected_frameinfo.height,(detected_frameinfo.is_interlace) ?'i':'p');
			framegrabber_s_input_status(board_v4l2_cxt->fg_handle,FRAMEGRABBER_INPUT_STATUS_OK);

			task_model_run_task_after(board_v4l2_cxt->task_model_handle,board_v4l2_cxt->check_signal_task_handle,100000);
		}
        break;
    case ADV7619_UNLOCK:
        printk("ADV7619_UNLOCK\n");
        framegrabber_s_input_status(board_v4l2_cxt->fg_handle,FRAMEGRABBER_INPUT_STATUS_NO_SIGNAL);
        framegrabber_s_input_framesize(board_v4l2_cxt->fg_handle,0,0);
        framegrabber_mask_s_status(board_v4l2_cxt->fg_handle,FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT,0);
        break;
    case ADV7619_UNLOCK_ENHANCE:
        printk("ADV7619_UNLOCK_ENHANCE\n");
        framegrabber_s_input_status(board_v4l2_cxt->fg_handle,FRAMEGRABBER_INPUT_STATUS_NO_SIGNAL);
        //framegrabber_s_input_framesize(board_v4l2_cxt->fg_handle,0,0);
        framegrabber_mask_s_status(board_v4l2_cxt->fg_handle,FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT,0);
        break;    
    case ADV7619_HDCP:
        //printk("ADV7619_HDCP\n");
        adv7619_get_frameinfo(adv7619_handle,fe_frameinfo);
        hdcp_flag = fe_frameinfo->adv7619_hdcp_flag;
		framegrabber_s_hdcp_flag(board_v4l2_cxt->fg_handle,hdcp_flag);
		break;    

	case ADV7619_AUDIO_RATE:
		adv7619_get_audiorate(adv7619_handle,fe_audiorate);		
		framegrabber_s_input_audiorate(board_v4l2_cxt->fg_handle,*fe_audiorate);
		if (*fe_audiorate ==0)
		    printk("ADV7619_AUDIO_RATE_CHANGE to 32000\n");	
		else if (*fe_audiorate ==1)
		    printk("ADV7619_AUDIO_RATE_CHANGE to 44100\n");	
		else if (*fe_audiorate ==2)
		    printk("ADV7619_AUDIO_RATE_CHANGE to 48000\n");	          
		else
		    printk("ADV7619_AUDIO_RATE_CHANGE to unsupported\n");	     
		break;
		
	case ADV7619_HOTPLUG:
		fpga_xilinx_hdmi_hotplug(board_v4l2_cxt->fpga_xilinx_handle);
		break;	
    } 
   
}

static void cx311h_adv7619_get_cardid(void *cxt, U32_T* cardid) 
{
	board_v4l2_context_t *board_v4l2_cxt = (board_v4l2_context_t*)cxt;

	if((board_v4l2_cxt == NULL) || (cardid == NULL))
	{
		return;
	}
	
	fpga_xilinx_get_cardid(board_v4l2_cxt->fpga_xilinx_handle, cardid);
}

void board_v4l2_init(cxt_mgr_handle_t cxt_mgr, unsigned long card_id)
{
    framegrabber_handle_t framegrabber_handle;
    board_v4l2_context_t *board_v4l2_cxt=NULL;
    v4l2_model_handle_t v4l2_handle;
    task_model_handle_t task_model_handle=NULL;
    handle_t i2c_mgr=NULL;
    handle_t fpga_xilinx_handle=NULL;
    task_handle_t task_handle=NULL;
    int i;
    
    enum
    {
        BOARD_V4L2_OK = 0,
        BOARD_V4L2_ERROR_CXT_MGR,
        BOARD_V4L2_ERROR_FRAMEGRABBER_INIT,
        BOARD_V4L2_ERROR_V4L2_MODEL_INIT,
        BOARD_V4L2_ERROR_ALLOC_CXT,
        BOARD_V4L2_ERROR_GET_MAIN_CHIP_HANDLE,
        BOARD_V4L2_ERROR_GET_I2C_MGR,
        BOARD_V4L2_ERROR_GET_TASK_MGR,
        BOARD_V4L2_ERROR_ALLOC_TASK,
    } err = BOARD_V4L2_OK;

    do
    {
        if (!cxt_mgr)
        {
            err = BOARD_V4L2_ERROR_CXT_MGR;
            break;
        }

        framegrabber_handle = framegrabber_init(cxt_mgr, &cx311h_property, &cx311h_ops);
        if (framegrabber_handle == NULL)
        {
            err = BOARD_V4L2_ERROR_FRAMEGRABBER_INIT;
            break;
        }
        board_v4l2_cxt = cxt_manager_add_cxt(cxt_mgr, BOARD_V4L2_CXT_ID, board_v4l2_alloc, board_v4l2_release);
        if (!board_v4l2_cxt)
        {
            err = BOARD_V4L2_ERROR_ALLOC_CXT;
            break;
        }
        board_v4l2_cxt->fg_handle = framegrabber_handle;
        framegrabber_set_data(framegrabber_handle,board_v4l2_cxt);
        
        v4l2_handle = v4l2_model_init(cxt_mgr, &device_info,framegrabber_handle,card_id);
        if(v4l2_handle==NULL)
	    {
            err=BOARD_V4L2_ERROR_V4L2_MODEL_INIT;
		    break;
	    }
	    board_v4l2_cxt->v4l2_handle=v4l2_handle;
        fpga_xilinx_handle=cxt_manager_get_context(cxt_mgr,FPGA_XILINX_CXT_ID,0);
	    if(fpga_xilinx_handle==NULL)
	    {
            err=BOARD_V4L2_ERROR_GET_MAIN_CHIP_HANDLE;
            break;
	    }
            
        board_v4l2_cxt->fpga_xilinx_handle=fpga_xilinx_handle;
        i2c_mgr=cxt_manager_get_context(cxt_mgr,I2C_CXT_ID,0);
	    if(i2c_mgr==NULL)
	    {
            err=BOARD_V4L2_ERROR_GET_I2C_MGR;
            break;
	    }
        for(i=0;i<CX311H_I2C_CHIP_COUNT;i++)
	    {
            board_v4l2_cxt->i2c_chip_handle[i]=i2c_model_get_nth_driver_handle(i2c_mgr,cx311h_chip_desc[i].name,cx311h_chip_desc[i].index);
            debug_msg("i2c_chip_handle[%d] %p\n",i,board_v4l2_cxt->i2c_chip_handle[i]);
	    }
        task_model_handle=cxt_manager_get_context(cxt_mgr,TASK_MODEL_CXT_ID,0);
        if(task_model_handle==NULL)
        {
            err=BOARD_V4L2_ERROR_GET_TASK_MGR;
            break;
        }
        board_v4l2_cxt->task_model_handle=task_model_handle;
        cxt_manager_ref_context(task_model_handle);
        
        task_handle=task_model_create_task(task_model_handle,check_signal_stable_task,board_v4l2_cxt,"check_signal_task");
        if(task_handle==NULL)
        {
            err=BOARD_V4L2_ERROR_ALLOC_TASK;
            break;
        }
        
        board_v4l2_cxt->check_signal_task_handle=task_handle;
        
        v4l2_model_register_callback(v4l2_handle,V4L2_MODEL_CALLBACK_STREAMON ,&cx311h_v4l2_stream_on, board_v4l2_cxt);
        v4l2_model_register_callback(v4l2_handle,V4L2_MODEL_CALLBACK_STREAMOFF,&cx311h_v4l2_stream_off, board_v4l2_cxt);
        v4l2_model_register_callback(v4l2_handle,V4L2_MODEL_CALLBACK_BUFFER_PREPARE,&cx311h_v4l2_buffer_prepare, board_v4l2_cxt);
        v4l2_model_register_callback(v4l2_handle,V4L2_MODEL_CALLBACK_BUFFER_INIT,&cx311h_v4l2_buffer_init, board_v4l2_cxt);
      
        adv7619_register_callback(board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0],cx311h_adv7619_event,board_v4l2_cxt);
        adv7619_register_callback_for_getcardid(board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0],cx311h_adv7619_get_cardid, board_v4l2_cxt);
        framegrabber_start(framegrabber_handle);
        cxt_manager_ref_context(fpga_xilinx_handle);
        
        fpga_xilinx_hdmi_hotplug(board_v4l2_cxt->fpga_xilinx_handle);
        //board_v4l2_cxt->board_id = board_id;
        //adv7619_get_board_id(board_v4l2_cxt->i2c_chip_handle[CX311H_I2C_CHIP_ADV7619_0],&board_v4l2_cxt->board_id); 
        
    }
    while (0);
    if (err != BOARD_V4L2_OK)
    {
        switch (err)
        {
        case BOARD_V4L2_ERROR_ALLOC_TASK:
            cxt_manager_unref_context(task_model_handle);
        case BOARD_V4L2_ERROR_GET_TASK_MGR:
        case BOARD_V4L2_ERROR_GET_I2C_MGR:
        case BOARD_V4L2_ERROR_V4L2_MODEL_INIT:
        case BOARD_V4L2_ERROR_ALLOC_CXT:
        case BOARD_V4L2_ERROR_FRAMEGRABBER_INIT:
        case BOARD_V4L2_ERROR_CXT_MGR:
        default:
            break;
        }
        debug_msg("%s err %d\n", __func__, err);
    }
}
