 /*
 *Driver for the card pcie bridge.

 *Copyright (C) 2019 AVerMedia TECHNOLOGIES, Inc.

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#include <linux/kernel.h>
#include "cxt_mgr.h"
#include "pci_model.h"
#include "mem_model.h"
#include "sys.h"
#include "debug.h"
#include <linux/i2c.h>
#include "i2c_model.h"
#include "gpio_model.h"
#include "mem_model.h"
#include "task_model.h"
#include "trace.h"
#include "fpga_xilinx.h"
#include "xilinx750_reg.h"

const unsigned int  FPGA_XILINX_FREQ=125*1000*1000;
enum
{
    FPGA_XILINX_MAX_VIDEO_DESC_LIST_COUNT=4,
    FPGA_XILINX_USR_VIDEO_DESC_LIST_COUNT=8,
    FPGA_XILINX_MAX_DESC_PER_DESC_LIST=2048,
    FPGA_XILINX_DEFAULT_AUDIO_BUF_SIZE=7680,
    FPGA_XILINX_AUDIO_BUF_COUNT=2,
};

enum
{
    DOMAIN_GENERAL,
    DOMAIN_INT,
    DOMAIN_SGDMA,
    DOMAIN_VIDEO,
    DOMAIN_VIDEO_CLIP,
    DOMAIN_VIDEO_SCALER,
    DOMAIN_FRAME_CTRL,
    DOMAIN_FRAME_CTRL_4K,
    DOMAIN_VIDEO_SCALER_CH0,
    DOMAIN_VIDEO_SCALER_CH1,
    DOMAIN_VIDEO_SCALER_CH2,
    DOMAIN_VIDEO_SCALER_CH3,
    DOMAIN_AUDIO_DMA,
    DOMAIN_IO,
    FPGA_XILINX_DOMAIN_COUNT,
};

typedef struct
{
    U32_T phys_L;
    U32_T phys_H;
    U32_T length;
    U32_T ctrl;
}fpga_xilinx_desc_t;

typedef struct
{
    queue_t queue;
    mem_model_dma_buffer_t *dma_buffer;
    fpga_xilinx_desc_t *descs;
    int desc_count;
    int use_ch;
    fpga_xilinx_callback_t callback;
    void *cb_cxt;
}fpga_xilinx_desclist_item_t;

typedef struct
{
    mem_model_dma_buffer_t *dma_buffer[FPGA_XILINX_AUDIO_BUF_COUNT];
    U32_T audio_buf_size;
    fpga_xilinx_audio_cb_t audio_cb;
    int cur_done_index;
    void *audio_cb_cxt;
    
}fpga_xilinx_audio_info_t;

typedef struct
{
    BASIC_CXT_HANDLE_DECLARE;
    pci_model_handle_t pci_handle;
    trace_attr_handle_t attr_handles[FPGA_XILINX_DOMAIN_COUNT];
    handle_t mem_mgr;
    handle_t task_mgr;
    int desc_list_count;
    queue_t free_desc_queue;
    queue_t ready_desc_queue;
    queue_t done_desc_queue;
    fpga_xilinx_desclist_item_t *tx_desclist[FPGA_XILINX_MAX_VIDEO_DESC_LIST_COUNT];
    sys_spin_lock_t *lock;
    fpga_xilinx_desclist_item_t *cur_desclist;
    int last_tx_ch;
    task_handle_t video_DPC_task;
    task_handle_t video_term_DPC_task;
    task_handle_t audio_DPC_task;
    fpga_xilinx_colorspace_e colorspace;
    BOOL_T is_streaming;
    sys_atomic_t atomic_VDPC_scheduling;
    BOOL_T is_ADPC_scheduling;
    BOOL_T can_be_terminated;
    fpga_xilinx_audio_info_t audio_info;
    int dual_pixel;
    
}fpga_xilinx_cxt_t;

typedef struct
{
    u32 offset;
    u32 length;
}reg_range_t;

#define BWFS_EVENT_NO_ERROR        0
#define BWFS_EVENT_NO_DEVICE       1
#define BWFS_EVENT_DATA_ERROR      2
#define BWFS_EVENT_TIMEOUT         3
#define BWFS_EVENT_PARMS_ERROR     4

static void *fpga_xilinx_alloc(void);
static void fpga_xilinx_release(void *data);
static u32 fpga_xilinx_ReadReg(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg);
static void  fpga_xilinx_WriteReg(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u32 value);
static void  fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u32 mask,u32 value);
static u8 fpga_xilinx_ReadRegByte(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg);
static void  fpga_xilinx_WriteRegByte(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u8 value);
static void  fpga_xilinx_MaskWriteRegByte(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u8 mask,u8 value);
static void fpga_xilinx_set_i2c_speed(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,int speed);
bool fpga_xilinx_I2CReadData(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,u8 *data,u8 cmd);
bool fpga_xilinx_I2CDoneCheck(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,bool checkack);
bool fpga_xilinx_I2CWriteData(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,u8 data,u8 cmd);
static int fpga_xilinx_i2c_xfer(void *ref_cxt,void *bus_data, i2c_model_msg_t *msg,int num);
static void fpga_xilinx_setgpio(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 mask,u32 value);
static u32 fpga_xilinx_getgpio(fpga_xilinx_cxt_t *fpga_xilinx_cxt);

static int fpga_xilinx_gpio_op_set(void *context,unsigned no, gpio_model_pin_value_t value);
static gpio_model_pin_value_t fpga_xilinx_gpio_op_get(void *context,unsigned no);
static void fpga_xilinx_videoch_done(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int ch);
static fpga_xilinx_desclist_item_t *fpga_xilinx_get_free_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt);
static fpga_xilinx_desclist_item_t *fpga_xilinx_get_ready_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt);
static fpga_xilinx_desclist_item_t *fpga_xilinx_get_done_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt);
static void fpga_xilinx_add_free_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,fpga_xilinx_desclist_item_t *desclist_item);
static int fpga_xilinx_tranfer_desclist(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_desclist_item_t *desclist_item);
static void fpga_xilinx_enable_int(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,BOOL_T enable);
static void  fpga_xilinx_csc_in_out(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,fpga_xilinx_colorspace_e colorspace_in,fpga_xilinx_colorspace_e colorspace_out,BOOL_T BT709_out,int ioctl_yuyv);
static void fpga_xilinx_audio_rate_input(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,fpga_xilinx_audio_e audio_rate);
static BOOL_T fpga_xilinx_is_TX_available(fpga_xilinx_cxt_t  *fpga_xilinx_cxt);
static int fpga_xilinx_flash_pageread(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,u8 *buffer,u32 offset);
static int fpga_xilinx_spi_done(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int wait);
static int fpga_xilinx_isp_enable(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int enable);
//static int fpga_xilinx_flash_erase(fpga_xilinx_cxt_t  *fpga_xilinx_cxt);
static int fpga_xilinx_flash_pageerase(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int type,U32_T address);
static int fpga_xilinx_flash_pagewrite(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,U8_T *buffer,U32_T offset);

static reg_range_t reg_domains[]=
{
    [DOMAIN_GENERAL]={
        .offset=0x4000,
        .length=0x100,
    },
    [DOMAIN_INT]={
        .offset=0x40,
        .length=0x20,
    },
    [DOMAIN_SGDMA]={
        .offset=0x4140,
        .length=0x40,
    },
    [DOMAIN_VIDEO]={
        .offset=0x4100,
        .length=0x20,
    },
    [DOMAIN_VIDEO_CLIP]={
        .offset=0x4900,
        .length=0x20,
    },
    [DOMAIN_VIDEO_SCALER]={
        .offset=0x4A00,
        .length=0x200,
    },
    [DOMAIN_FRAME_CTRL]={
        .offset=0x4f00,
        .length=0x20,
    },
    [DOMAIN_FRAME_CTRL_4K]={
        .offset=0x4800,
        .length=0x90,
    },
    [DOMAIN_VIDEO_SCALER_CH0]={
        .offset=0x5000,
        .length=0x200,
    },
    [DOMAIN_VIDEO_SCALER_CH1]={
        .offset=0x5200,
        .length=0x200,
    },
    [DOMAIN_VIDEO_SCALER_CH2]={
        .offset=0x5400,
        .length=0x200,
    },
    [DOMAIN_VIDEO_SCALER_CH3]={
        .offset=0x5600,
        .length=0x200,
    },
    [DOMAIN_AUDIO_DMA]={
        .offset=0x4184,
        .length=0x20,
    },
    [DOMAIN_IO]={
        .offset=0x41C0,
        .length=0x10,
    },   
};

typedef struct 
{
    U32_T Y_Coef;
    U32_T Cb_Coef;
    U32_T Cr_Coef;
}fpga_xilinx_csc_coef_t;

#define FPGA_XILINX_TRACE_FUNC_DECLARE(x)\
static unsigned int x##_show(void *cxt,  char *buf) \
{\
	fpga_xilinx_cxt_t *fpga_xilinx_cxt=cxt;\
        int i;\
	int c;\
	int columns=16; \
        reg_range_t *reg_range=&reg_domains[DOMAIN_##x];\
\
\
	c=0;\
	c+=sys_sprintf(&buf[c],"     ");\
	for(i=0;i<columns;i+=4) \
            c+=sys_sprintf(&buf[c],"\t%08x",i); \
	c+=sys_sprintf(&buf[c],"\n"); \
	c+=sys_sprintf(&buf[c],"_____"); \
	for(i=0;i<columns;i+=4) \
            c+=sys_sprintf(&buf[c],"\t________"); \
	c+=sys_sprintf(&buf[c],"\n");\
	for(i=0;i<reg_range->length;i+=4) \
	{ \
            if((i% columns)==0) \
            {\
		if(i>0 ) \
                    c+=sys_sprintf(&buf[c],"\n");\
		c+=sys_sprintf(&buf[c],"%04x: ",reg_range->offset+i); \
            } \
            c+=sys_sprintf(&buf[c],"\t%08x",fpga_xilinx_ReadReg(fpga_xilinx_cxt,reg_range->offset+i)); \
	}\
	c+=sys_sprintf(&buf[c],"\n");\
  return c;\
}\
\
static unsigned int  x##_store(void *cxt, const char *buf, unsigned int  count) \
{ \
	fpga_xilinx_cxt_t *fpga_xilinx_cxt=cxt;\
	unsigned reg;\
	unsigned val; \
         \
	if(sys_sscanf(buf,"%x=%x", &reg,&val)==2) \
	{ \
		printk("%s set reg %x = %x\n",#x,reg,val); \
                    fpga_xilinx_WriteReg(fpga_xilinx_cxt,reg,val);\
		}\
    return count;\
}\


static gpio_ic_ops_t fpga_xilinx_gpio_ops=
{
        .set=fpga_xilinx_gpio_op_set,
        .get=fpga_xilinx_gpio_op_get,
};

static void *fpga_xilinx_alloc()
{
	fpga_xilinx_cxt_t *fpga_xilinx_cxt=mem_model_alloc_buffer(sizeof(fpga_xilinx_cxt_t));

	if(fpga_xilinx_cxt)
	{

	}
	return fpga_xilinx_cxt;
}

static void fpga_xilinx_release(void *data)
{
    fpga_xilinx_cxt_t *fpga_xilinx_cxt=data;
    
    if(fpga_xilinx_cxt)
    {
        int i;
        fpga_xilinx_enable_int(fpga_xilinx_cxt,FALSE);
        if(fpga_xilinx_cxt->audio_DPC_task)
            task_model_release_task(fpga_xilinx_cxt->task_mgr,fpga_xilinx_cxt->audio_DPC_task);
        if(fpga_xilinx_cxt->video_DPC_task)
            task_model_release_task(fpga_xilinx_cxt->task_mgr,fpga_xilinx_cxt->video_DPC_task);
        if(fpga_xilinx_cxt->video_term_DPC_task)
            task_model_release_task(fpga_xilinx_cxt->task_mgr,fpga_xilinx_cxt->video_term_DPC_task);
        if(fpga_xilinx_cxt->atomic_VDPC_scheduling)
            sys_release_atomic(fpga_xilinx_cxt->atomic_VDPC_scheduling);
        cxt_manager_unref_context(fpga_xilinx_cxt->task_mgr);
        cxt_manager_unref_context(fpga_xilinx_cxt->pci_handle);
        for(i=0;i<sizeof(fpga_xilinx_cxt->attr_handles)/sizeof(void *);i++)
            trace_attr_free(fpga_xilinx_cxt->attr_handles[i]);
        mem_model_free_buffer(fpga_xilinx_cxt);
    }
}

static u32 fpga_xilinx_ReadReg(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg)
{
    u32 value=0;
    
    if(fpga_xilinx_cxt)
    {
        value=pci_model_mmio_read(fpga_xilinx_cxt->pci_handle,0,reg);
    }
    return value;
}

static void  fpga_xilinx_WriteReg(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u32 value)
{
    if(fpga_xilinx_cxt)
    {
        pci_model_mmio_write(fpga_xilinx_cxt->pci_handle,0,reg,value);
    }
    
}
// write only bits with mask bit is 1
static void  fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u32 mask,u32 value)
{
    if(fpga_xilinx_cxt)
    {
        u32 old=fpga_xilinx_ReadReg(fpga_xilinx_cxt,reg);
        u32 new=(old & ~mask) | (value & mask);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,reg,new);
    }
}

static u8 fpga_xilinx_ReadRegByte(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg)
{
    u8 value=0;
    
    if(fpga_xilinx_cxt)
    {
        value=pci_model_mmio_readb(fpga_xilinx_cxt->pci_handle,0,reg);
    }
    return value;
}

static void  fpga_xilinx_WriteRegByte(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u8 value)
{
    if(fpga_xilinx_cxt)
    {
        pci_model_mmio_writeb(fpga_xilinx_cxt->pci_handle,0,reg,value);
    } 
}

// write only bits with mask bit is 1
static __attribute__ ((unused)) void  fpga_xilinx_MaskWriteRegByte(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 reg,u8 mask,u8 value)
{
    if(fpga_xilinx_cxt)
    {
        u8 old=fpga_xilinx_ReadRegByte(fpga_xilinx_cxt,reg);
        u8 new=(old & ~mask) | (value & mask);
        fpga_xilinx_WriteRegByte(fpga_xilinx_cxt,reg,new);
    }
}

static void fpga_xilinx_set_i2c_speed(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,int speed)
{
    u32 reg=0;
    if((bus<0) || (bus >=FPGA_XILINX_I2C_BUS_NUM))
        return;
    if(fpga_xilinx_cxt)
    {
        switch(bus)
        {
            case FPGA_XILINX_I2C_BUS_0:
                reg=REG_FPGA_BRIDGE_750_I2C0_SPEED_L;
                break;        
            //case FPGA_XILINX_I2C_BUS_1:
            //    reg=REG_FPGA_BRIDGE_750_I2C1_SPEED_L;
            //    break;
            //case FPGA_XILINX_I2C_BUS_2:
            //    reg=REG_FPGA_BRIDGE_750_I2C2_SPEED_L;
            //    break;
            //case FPGA_XILINX_I2C_BUS_3:
            //    reg=REG_FPGA_BRIDGE_750_I2C_SEC_SPEED_L; 
            //    break;
            //case FPGA_XILINX_I2C_BUS_4:
            //    reg=REG_FPGA_EXTERNAL_EDID_I2C_SPEED_L;
            //    break;
            default:
                debug_msg("%s wrong bus %d,debug!!! \n",__func__,bus);
        }
        if(reg)
        {
            unsigned int value= (FPGA_XILINX_FREQ/5)/speed-1;
            printk("%s bus %d value %x\n",__func__,bus,value);
            fpga_xilinx_WriteReg(fpga_xilinx_cxt,reg,value & 0xff);
            fpga_xilinx_WriteReg(fpga_xilinx_cxt,reg+4,(value >> 8) & 0xff );
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,reg+8,I2C_ENABLE_MASK,I2C_ENABLE_MASK);
        }      
    }  
}

bool fpga_xilinx_I2CDoneCheck(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,bool checkack)
{
    u32 regStatus=0;
    u8 status;
    int timeout;
    if((bus<0) || (bus >=FPGA_XILINX_I2C_BUS_NUM))
        return false;
    if(fpga_xilinx_cxt)
    {
        switch(bus)
        {
            case FPGA_XILINX_I2C_BUS_0:
                regStatus=REG_FPGA_BRIDGE_750_I2C0_STATUS;
                break;        
            //case FPGA_XILINX_I2C_BUS_1:
            //    regStatus=REG_FPGA_BRIDGE_750_I2C1_STATUS;
            //    break;
            //case FPGA_XILINX_I2C_BUS_2:
            //    regStatus=REG_FPGA_BRIDGE_750_I2C2_STATUS;
            //    break;
            //case FPGA_XILINX_I2C_BUS_3:
            //    regStatus=REG_FPGA_BRIDGE_750_I2C_SEC_STATUS;
            //    break;
            //case FPGA_XILINX_I2C_BUS_4:
            //    regStatus=REG_FPGA_EXTERNAL_EDID_I2C_STATUS;
            //    break;
            default:
                debug_msg("%s wrong bus %d,debug!!! \n",__func__,bus);
                return false;
                break;
        }       
    }
    status=fpga_xilinx_ReadRegByte(fpga_xilinx_cxt,regStatus);
    timeout=200;
    while( (status & I2C_STATUS_READ_MASK) && timeout)
    {
        status=fpga_xilinx_ReadRegByte(fpga_xilinx_cxt,regStatus);
        timeout--;
    }
    if(timeout==0)
    {
        debug_msg("%s timeout status %02x\n",__func__,status);
        return false;
    }
    return (checkack ? ((status & I2C_STATUS_NACK_MASK)==0) : true);  
}

bool fpga_xilinx_I2CWriteData(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,u8 data,u8 cmd)
{
    u32 regData=0;
    u32 regCmd=0;
    
    if((bus<0) || (bus >=FPGA_XILINX_I2C_BUS_NUM))
        return false;
    
    if(fpga_xilinx_cxt)
    {
        switch (bus) {
            case FPGA_XILINX_I2C_BUS_0:
                regData = REG_FPGA_BRIDGE_750_I2C0_DATA;
                regCmd = REG_FPGA_BRIDGE_750_I2C0_CMD;
                break;
            //case FPGA_XILINX_I2C_BUS_1:
            //    regData = REG_FPGA_BRIDGE_750_I2C1_DATA;
            //    regCmd = REG_FPGA_BRIDGE_750_I2C1_CMD;
            //    break;
            //case FPGA_XILINX_I2C_BUS_2:
            //    regData = REG_FPGA_BRIDGE_750_I2C2_DATA;
            //    regCmd = REG_FPGA_BRIDGE_750_I2C2_CMD;
            //    break;
            //case FPGA_XILINX_I2C_BUS_3:
            //    regData = REG_FPGA_BRIDGE_750_I2C_SEC_DATA;
            //    regCmd = REG_FPGA_BRIDGE_750_I2C_SEC_CMD;
            //    break;
            //case FPGA_XILINX_I2C_BUS_4:
            //    regData = REG_FPGA_EXTERNAL_EDID_I2C_DATA;
            //    regCmd = REG_FPGA_EXTERNAL_EDID_I2C_CMD;
            //    break;
            default:
                debug_msg("%s wrong bus %d,debug!!! \n", __func__, bus);
                return false;
                break;
        }
        fpga_xilinx_WriteRegByte(fpga_xilinx_cxt,regData,data);
        fpga_xilinx_WriteRegByte(fpga_xilinx_cxt,regCmd,cmd);
        if(!fpga_xilinx_I2CDoneCheck(fpga_xilinx_cxt,bus,true))
        {
            return false;
        }
    }
    return true;
}

static int fpga_xilinx_i2c_xfer(void *ref_cxt,void *bus_data, i2c_model_msg_t *msg,int num)
{
    fpga_xilinx_cxt_t *fpga_xilinx_cxt=ref_cxt;
    int i=0,j;
    u8 cmd;
    long bus=(long)bus_data;
    
    for(i=0;i<num;i++)
    {
        cmd=0;
        cmd |= I2C_CMD_START_MASK | I2C_CMD_WRITE_MASK;
        if(msg[i].flags & I2C_MODEL_READ)
        {
            if(!fpga_xilinx_I2CWriteData(fpga_xilinx_cxt,bus,(msg[i].addr<<1|0x1)&0xff,cmd))
              break;
        }
        else
        {
            if(!fpga_xilinx_I2CWriteData(fpga_xilinx_cxt,bus,(msg[i].addr<<1)&0xff,cmd))
              break;
        }
           
        for(j=0;j< msg[i].len;j++)
        {
            cmd=0;
            
            if((i==num-1) && (j==msg[i].len-1))
            {
                cmd |= I2C_CMD_STOP_MASK;
                if(msg[i].flags & I2C_MODEL_READ)
                    cmd |= I2C_CMD_NACK_MASK;
            }
            if(msg[i].flags & I2C_MODEL_READ)
            {
                cmd |= I2C_CMD_READ_MASK;
                if(!fpga_xilinx_I2CReadData(fpga_xilinx_cxt,bus,&msg[i].buf[j],cmd))
                    break;
            }    
            else
            {
                cmd |= I2C_CMD_WRITE_MASK;
                if(!fpga_xilinx_I2CWriteData(fpga_xilinx_cxt,bus,msg[i].buf[j],cmd))
                    break;
            }
        }
            
    }
#if 0    
    for(i=0;i<num;i++)
    {
        debug_msg("%s %02x\t:",(msg[i].flags & I2C_MODEL_READ)?"RD":"WR",msg[i].addr);
     //   if(!(msg[i].flags & I2C_MODEL_READ))
        {
            for(j=0;j< msg[i].len;j++)
                debug_msg("%02x ",msg[i].buf[j]);
        }
        debug_msg("\n");
            
    }
    debug_msg("\n");
#endif    
    
    return i;
}

bool fpga_xilinx_I2CReadData(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_i2c_bus_e bus,u8 *data,u8 cmd)
{
    u32 regData=0;
    u32 regCmd=0;
    
    if((bus<0) || (bus >=FPGA_XILINX_I2C_BUS_NUM))
        return false;
    
    if(fpga_xilinx_cxt)
    {
        switch (bus) {
            case FPGA_XILINX_I2C_BUS_0:
                regData = REG_FPGA_BRIDGE_750_I2C0_DATA;
                regCmd = REG_FPGA_BRIDGE_750_I2C0_CMD;
                break;
            //case FPGA_XILINX_I2C_BUS_1:
            //    regData = REG_FPGA_BRIDGE_750_I2C1_DATA;
            //    regCmd = REG_FPGA_BRIDGE_750_I2C1_CMD;
            //    break;
            //case FPGA_XILINX_I2C_BUS_2:
            //    regData = REG_FPGA_BRIDGE_750_I2C2_DATA;
            //    regCmd = REG_FPGA_BRIDGE_750_I2C2_CMD;
            //    break;
            //case FPGA_XILINX_I2C_BUS_3:
            //    regData = REG_FPGA_BRIDGE_750_I2C_SEC_DATA;
            //    regCmd = REG_FPGA_BRIDGE_750_I2C_SEC_CMD;
            //    break;
            //case FPGA_XILINX_I2C_BUS_4:
            //    regData = REG_FPGA_EXTERNAL_EDID_I2C_DATA;
            //    regCmd = REG_FPGA_EXTERNAL_EDID_I2C_CMD;
            //    break;
            default:
                debug_msg("%s wrong bus %d,debug!!! \n", __func__, bus);
                return false;
                break;
        }
        fpga_xilinx_WriteRegByte(fpga_xilinx_cxt,regCmd,cmd);
        
        if(!fpga_xilinx_I2CDoneCheck(fpga_xilinx_cxt,bus,false))
        {
            return false;
        }
        *data=fpga_xilinx_ReadRegByte(fpga_xilinx_cxt,regData);
    }
    return true;
}

static void fpga_xilinx_setgpio(fpga_xilinx_cxt_t *fpga_xilinx_cxt,u32 mask,u32 value)
{
    debug_msg("%s mask %08x value %08x\n",__func__,mask,value);
    fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GPIO,mask,value);
}

static u32 fpga_xilinx_getgpio(fpga_xilinx_cxt_t *fpga_xilinx_cxt)
{
    return fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GPIO);
}

static int fpga_xilinx_gpio_op_set(void *context,unsigned no, gpio_model_pin_value_t value)
{
    fpga_xilinx_cxt_t *fpga_xilinx_cxt=context;
    
    if(fpga_xilinx_cxt)
    {
        u32 gpio_mask=0x1<< no;
        u32 gpio_value=(value==GPIOIC_VALUE_HIGH)? gpio_mask:0;
        fpga_xilinx_setgpio(fpga_xilinx_cxt,gpio_mask,gpio_value);
    }
    return 0;
}

static gpio_model_pin_value_t fpga_xilinx_gpio_op_get(void *context,unsigned no)
{
    fpga_xilinx_cxt_t *fpga_xilinx_cxt=context;
    u32 gpio_value;
    gpio_model_pin_value_t value=GPIOIC_VALUE_UNKNOWN;
    
    if(fpga_xilinx_cxt)
    {
        gpio_value=fpga_xilinx_getgpio(fpga_xilinx_cxt);
        if(gpio_value & (0x1 << no))
            value=GPIOIC_VALUE_HIGH;
        else
            value=GPIOIC_VALUE_LOW;
    }
    return value;
}

void fpga_xilinx_get_i2c_bus_cfg(handle_t fpga_xilinx_handle,fpga_xilinx_i2c_bus_e bus,i2c_model_bus_cfg_t *i2c_bus_cfg)
{
    if(i2c_bus_cfg)
    {
        i2c_bus_cfg->ref_cxt=fpga_xilinx_handle;
        i2c_bus_cfg->bus_data=(void *)(long)bus;
        i2c_bus_cfg->i2c_transfer_func=fpga_xilinx_i2c_xfer;
    }
}

FPGA_XILINX_TRACE_FUNC_DECLARE(GENERAL);
FPGA_XILINX_TRACE_FUNC_DECLARE(INT);
FPGA_XILINX_TRACE_FUNC_DECLARE(SGDMA);
FPGA_XILINX_TRACE_FUNC_DECLARE(VIDEO);
FPGA_XILINX_TRACE_FUNC_DECLARE(VIDEO_CLIP);
FPGA_XILINX_TRACE_FUNC_DECLARE(VIDEO_SCALER);
FPGA_XILINX_TRACE_FUNC_DECLARE(FRAME_CTRL);
FPGA_XILINX_TRACE_FUNC_DECLARE(FRAME_CTRL_4K);
FPGA_XILINX_TRACE_FUNC_DECLARE(VIDEO_SCALER_CH0);
FPGA_XILINX_TRACE_FUNC_DECLARE(VIDEO_SCALER_CH1);
FPGA_XILINX_TRACE_FUNC_DECLARE(VIDEO_SCALER_CH2);
FPGA_XILINX_TRACE_FUNC_DECLARE(VIDEO_SCALER_CH3);
FPGA_XILINX_TRACE_FUNC_DECLARE(AUDIO_DMA);
FPGA_XILINX_TRACE_FUNC_DECLARE(IO);

void fpga_xilinx_add_trace(handle_t fpga_xilinx_handle,void *trace_handle)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    trace_attr_setup_t setup;
    trace_attr_handle_t attr_handle;
    int index=0;

    if(!fpga_xilinx_cxt || !trace_handle)
            return;
#define FPGA_XILINX_ADD_TRACE(x) \
    setup.name = "FPGA_XILINX_"#x; \
    setup.show = x##_show; \
    setup.store = x##_store; \
    setup.cxt = fpga_xilinx_cxt; \
    attr_handle=trace_attr_create(&setup); \
    if(!attr_handle)\
    { \
       break;\
    }\
    trace_model_add_attr(trace_handle,attr_handle); 
    
    
    index=0;
    do
    {
        FPGA_XILINX_ADD_TRACE(GENERAL);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;
        FPGA_XILINX_ADD_TRACE(INT);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;
        FPGA_XILINX_ADD_TRACE(SGDMA);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;
        FPGA_XILINX_ADD_TRACE(VIDEO);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;
        FPGA_XILINX_ADD_TRACE(VIDEO_CLIP);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;
        FPGA_XILINX_ADD_TRACE(VIDEO_SCALER);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;
        FPGA_XILINX_ADD_TRACE(FRAME_CTRL);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;        
        FPGA_XILINX_ADD_TRACE(FRAME_CTRL_4K);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;        
        FPGA_XILINX_ADD_TRACE(VIDEO_SCALER_CH0);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;            
        FPGA_XILINX_ADD_TRACE(VIDEO_SCALER_CH1);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;            
        FPGA_XILINX_ADD_TRACE(VIDEO_SCALER_CH2);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;                
        FPGA_XILINX_ADD_TRACE(VIDEO_SCALER_CH3);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;                
        FPGA_XILINX_ADD_TRACE(AUDIO_DMA);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;                
        FPGA_XILINX_ADD_TRACE(IO);
        fpga_xilinx_cxt->attr_handles[index++]=attr_handle;                        
        
    }while(0);
}

static fpga_xilinx_desclist_item_t *fpga_xilinx_new_desclist_item(fpga_xilinx_cxt_t *fpga_xilinx_cxt)
{
    fpga_xilinx_desclist_item_t *desc_item = mem_model_alloc_buffer(sizeof (fpga_xilinx_desclist_item_t));
    if(desc_item)
    {
        mem_model_dma_pool_cfg_t dma_pool_cfg ={
            .name = "fpga_xilinx",
            .size = sizeof (fpga_xilinx_desc_t) * FPGA_XILINX_MAX_DESC_PER_DESC_LIST,
            .align = 0,
            .boundary = 0,
        };
        mem_model_dma_buffer_t *dma_buffer;
        dma_buffer = mem_model_dma_pool_alloc_atomic(fpga_xilinx_cxt->mem_mgr, &dma_pool_cfg);
        init_queue(&desc_item->queue);
        desc_item->desc_count = 0;
        desc_item->descs = (fpga_xilinx_desc_t *) dma_buffer->vaddr;
        desc_item->dma_buffer = dma_buffer;
    }    
    
    return desc_item;
}

static void fpga_xilinx_schedule_audio_DPC(fpga_xilinx_cxt_t   *fpga_xilinx_cxt)
{
    if(!fpga_xilinx_cxt->is_ADPC_scheduling)
    {
        fpga_xilinx_cxt->is_ADPC_scheduling=TRUE;
        task_model_schedule_DPC(fpga_xilinx_cxt->task_mgr,fpga_xilinx_cxt->audio_DPC_task);
    }
}

static void fpga_xilinx_schedule_video_DPC(fpga_xilinx_cxt_t   *fpga_xilinx_cxt)
{
    if(!sys_atomic_get(fpga_xilinx_cxt->atomic_VDPC_scheduling))
    {
        debug_msg("&");
        sys_atomic_set(fpga_xilinx_cxt->atomic_VDPC_scheduling,1);
        task_model_schedule_DPC(fpga_xilinx_cxt->task_mgr,fpga_xilinx_cxt->video_DPC_task);
    }
}

static int fpga_xilinx_irq_func(void *cxt)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=cxt;
    fpga_xilinx_desclist_item_t *ready_desclist;
    int ret=0;
    U32_T index1;
    U32_T index2;
    
    U32_T status = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_INTRPT_TYPE);
    
    //debug_msg("%s status %08x\n",__func__,status);
    
    if(0 == (status & 0xff))
    {
        ret= -1;
        return ret;
	}
    if(status & VIDEO_DMA_DONE_MASK)
    {
        index1=(fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_STATUS) & BITMASK(3)) -1;
        //fpga_xilinx_desclist_item_t *ready_desclist;
      
        debug_msg("V%d\n",index1);
        
        if (index1 >= 0 && index1 < 4)
        {
            fpga_xilinx_videoch_done(fpga_xilinx_cxt,index1);
        
        // to clear the related valid bit
        
            if(fpga_xilinx_is_TX_available(fpga_xilinx_cxt))
            {
                ready_desclist=fpga_xilinx_get_ready_desclist(fpga_xilinx_cxt);
                if(ready_desclist)
                {
                    fpga_xilinx_tranfer_desclist(fpga_xilinx_cxt,ready_desclist);
                    debug_msg("*");
                }
            }
        
            fpga_xilinx_schedule_video_DPC(fpga_xilinx_cxt);
        // ack
	    }
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_ACK,ACK_VIDEO_DMA_DONE_MASK,ACK_VIDEO_DMA_DONE_MASK);
        fpga_xilinx_cxt->can_be_terminated=FALSE; 
    }
    
    if(status & VIDEO_DMA_TERM_MASK)
    {
        debug_msg("%s VIDEO TERM\n",__func__);
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_ACK,ACK_VIDEO_DMA_TERM_MASK,ACK_VIDEO_DMA_TERM_MASK);
        task_model_schedule_DPC(fpga_xilinx_cxt->task_mgr,fpga_xilinx_cxt->video_term_DPC_task);
        fpga_xilinx_cxt->can_be_terminated=TRUE; 
    }
    
    if(status & AUDIO_DMA_DONE_MASK)
    {
        index2=((fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_STATUS)) & BITMASK(2))-1;
        
        if (index2 == 0 || index2 ==1)
        {
            fpga_xilinx_cxt->audio_info.cur_done_index=index2;    
            fpga_xilinx_schedule_audio_DPC(fpga_xilinx_cxt);
		}
        // ack
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_ACK,ACK_AUDIO_DMA_DONE_MASK,ACK_AUDIO_DMA_DONE_MASK);
    }
    
    if(status & AUDIO_DMA_TERM_MASK)
    {
        debug_msg("%s AUDIO TERM\n",__func__);
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_ACK,ACK_AUDIO_DMA_TERM_MASK,ACK_AUDIO_DMA_TERM_MASK);
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_ACK,ACK_AUDIO_DMA_DONE_MASK,ACK_AUDIO_DMA_DONE_MASK);
    }
    
    if(status & VIDEO_DMA_EOF_MASK)
	{
	    debug_msg("VE\n");
	    index1=(fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_STATUS) & BITMASK(3)) -1;
	    //if (index1 >= 0 && index1 < 4)
	        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_ACK,ACK_VIDEO_DMA_EOF_MASK,VIDEO_DMA_EOF_MASK);
		fpga_xilinx_cxt->can_be_terminated=TRUE; 	
	}

    if(status & VIDEO_DMA_LESS_MASK)
	{
		debug_msg("VL\n");
		index1=(fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_STATUS) & BITMASK(3)) -1;
		//if (index1 >= 0 && index1 < 4)
	        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_ACK,ACK_VIDEO_DMA_LESS_MASK,VIDEO_DMA_LESS_MASK);
	    fpga_xilinx_cxt->can_be_terminated=TRUE; 
	}
	
    if(status & OTHER_DMA_TERM_MASK)
	{
	    fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_ACK,ACK_OTHER_DMA_TERM_MASK,OTHER_DMA_TERM_MASK);
        fpga_xilinx_cxt->can_be_terminated=TRUE;
	}
    
    return ret;
}

static fpga_xilinx_desclist_item_t *fpga_xilinx_get_free_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt)
{
    fpga_xilinx_desclist_item_t *desclist_item=NULL;
    if(fpga_xilinx_cxt)
    {      
        sys_spin_lock(fpga_xilinx_cxt->lock);
        desclist_item=queue_first_entry_or_null(&fpga_xilinx_cxt->free_desc_queue,fpga_xilinx_desclist_item_t,queue);
        if(desclist_item)
        {
            queue_del(&desclist_item->queue);
        }
        sys_spin_unlock(fpga_xilinx_cxt->lock);
    }
    return desclist_item;
}

static void fpga_xilinx_add_free_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,fpga_xilinx_desclist_item_t *desclist_item)
{ 
    if(fpga_xilinx_cxt)
    {    
        sys_spin_lock(fpga_xilinx_cxt->lock);
        queue_add_tail(&desclist_item->queue,&fpga_xilinx_cxt->free_desc_queue);
        sys_spin_unlock(fpga_xilinx_cxt->lock);
    }
}

static fpga_xilinx_desclist_item_t *fpga_xilinx_get_ready_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt)
{
    fpga_xilinx_desclist_item_t *desclist_item=NULL;
    if(fpga_xilinx_cxt)
    {
        
        sys_spin_lock(fpga_xilinx_cxt->lock);
        desclist_item=queue_first_entry_or_null(&fpga_xilinx_cxt->ready_desc_queue,fpga_xilinx_desclist_item_t,queue);
        if(desclist_item)
        {
            queue_del(&desclist_item->queue);
        }
        sys_spin_unlock(fpga_xilinx_cxt->lock);
    }
    return desclist_item;
}
static void fpga_xilinx_add_ready_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,fpga_xilinx_desclist_item_t *desclist_item)
{
    
    if(fpga_xilinx_cxt)
    {
        
        sys_spin_lock(fpga_xilinx_cxt->lock);
        queue_add_tail(&desclist_item->queue,&fpga_xilinx_cxt->ready_desc_queue);
        sys_spin_unlock(fpga_xilinx_cxt->lock);
    }
}

static fpga_xilinx_desclist_item_t *fpga_xilinx_get_done_desclist(fpga_xilinx_cxt_t  *fpga_xilinx_cxt)
{
    fpga_xilinx_desclist_item_t *desclist_item=NULL;
    if(fpga_xilinx_cxt)
    {
        
        sys_spin_lock(fpga_xilinx_cxt->lock);
        desclist_item=queue_first_entry_or_null(&fpga_xilinx_cxt->done_desc_queue,fpga_xilinx_desclist_item_t,queue);
        if(desclist_item)
        {
            queue_del(&desclist_item->queue);
        }
        sys_spin_unlock(fpga_xilinx_cxt->lock);
    }
    return desclist_item;
}

void fpga_xilinx_add_to_cur_desclist(handle_t fpga_xilinx_handle,unsigned long phys_addr,unsigned size)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    fpga_xilinx_desclist_item_t *cur_desclist;
    
    if(fpga_xilinx_cxt->cur_desclist==NULL)
    {
        if(fpga_xilinx_cxt->desc_list_count<FPGA_XILINX_MAX_VIDEO_DESC_LIST_COUNT)
        {
            fpga_xilinx_cxt->cur_desclist=fpga_xilinx_new_desclist_item(fpga_xilinx_cxt);
            fpga_xilinx_cxt->desc_list_count++;
        }
        else
        {
            fpga_xilinx_cxt->cur_desclist=fpga_xilinx_get_free_desclist(fpga_xilinx_cxt);
        }
    }
    cur_desclist=fpga_xilinx_cxt->cur_desclist;
    if(cur_desclist)
    {
        fpga_xilinx_desc_t *desc;
        desc=&cur_desclist->descs[cur_desclist->desc_count];
        desc->phys_L=phys_addr & 0xffffffff;
        desc->phys_H=(phys_addr >> 32);
        desc->length=size/4; 
        desc->ctrl=0x80006000;
        cur_desclist->desc_count++;
    }
}

void fpga_xilinx_active_current_desclist(handle_t fpga_xilinx_handle,fpga_xilinx_callback_t callback,void *cb_cxt)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    
    if ((fpga_xilinx_cxt==NULL) || (cb_cxt ==NULL) || (callback==NULL))
        printk("%s ini fail\n",__func__);
    if(fpga_xilinx_cxt)
    {
        if(fpga_xilinx_cxt->cur_desclist)
        {
            fpga_xilinx_desclist_item_t *desclist_item=fpga_xilinx_cxt->cur_desclist;
                               
            desclist_item->callback=callback;
            desclist_item->cb_cxt=cb_cxt;
            fpga_xilinx_cxt->cur_desclist=NULL;
            fpga_xilinx_add_ready_desclist(fpga_xilinx_cxt,desclist_item);
            
        }
        if(fpga_xilinx_cxt->is_streaming)
        {
            debug_msg("!");
            fpga_xilinx_schedule_video_DPC(fpga_xilinx_cxt); 
        }
    }
}
static int fpga_xilinx_tranfer_desclist(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_desclist_item_t *desclist_item)
{
    int ret=0;
    if(fpga_xilinx_cxt && desclist_item)
    {
        int ch;
        U32_T mask;
        
        ch=fpga_xilinx_cxt->last_tx_ch+1;
        if(ch==FPGA_XILINX_MAX_VIDEO_DESC_LIST_COUNT)
            ch=0;
        if(fpga_xilinx_cxt->tx_desclist[ch]==NULL)
        {  
            sys_spin_lock(fpga_xilinx_cxt->lock);
            fpga_xilinx_cxt->tx_desclist[ch]=desclist_item;
            desclist_item->use_ch=ch;
            fpga_xilinx_cxt->last_tx_ch=ch;
            sys_spin_unlock(fpga_xilinx_cxt->lock);
            fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_DESPTRLIST_ADDR_L(ch),desclist_item->dma_buffer->phys_addr& 0xffffffff);
            fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_DESPTRLIST_ADDR_H(ch),(desclist_item->dma_buffer->phys_addr >> 32));
            fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_DESPTRLIST_NUM(ch),desclist_item->desc_count);
            mask=0x1<<(ch+1);
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_CTRL,mask,mask);
        }
        else
            ret=-1;
    }
    return ret;
}

static void fpga_xilinx_videoch_done(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int ch)
{
    if(fpga_xilinx_cxt)
    {
        fpga_xilinx_desclist_item_t *done_desclist=fpga_xilinx_cxt->tx_desclist[ch];
        fpga_xilinx_cxt->tx_desclist[ch]=NULL;
        if(done_desclist)
        {            
            sys_spin_lock(fpga_xilinx_cxt->lock);
            done_desclist->use_ch=-1;
            queue_add_tail(&done_desclist->queue,&fpga_xilinx_cxt->done_desc_queue);
            sys_spin_unlock(fpga_xilinx_cxt->lock);          
        }      
    }   
}

static void fpga_xilinx_enable_int(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,BOOL_T enable)
{
    if(fpga_xilinx_cxt)
    {
        U32_T mask=VIDEO_DMA_DONE_MASK | AUDIO_DMA_DONE_MASK | VIDEO_DMA_TERM_MASK | AUDIO_DMA_TERM_MASK | VIDEO_DMA_EOF_MASK | VIDEO_DMA_LESS_MASK | OTHER_DMA_TERM_MASK;
        if(enable)
        {
            fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_INTRPT,mask);        
        }
        else
        {
            fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_INTRPT,0x00);
        }     
    }   
}
static void fpga_xilinx_audio_DPC_func(void *data)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=data;
    
    if(fpga_xilinx_cxt)
    {
        int index=fpga_xilinx_cxt->audio_info.cur_done_index;
        
        if(fpga_xilinx_cxt->audio_info.audio_cb)
        {
            fpga_xilinx_cxt->audio_info.audio_cb(fpga_xilinx_cxt->audio_info.audio_cb_cxt,(U8_T *)fpga_xilinx_cxt->audio_info.dma_buffer[index]->vaddr ,fpga_xilinx_cxt->audio_info.audio_buf_size);
        }
        fpga_xilinx_cxt->is_ADPC_scheduling=FALSE;
    }
}
static BOOL_T fpga_xilinx_is_TX_available(fpga_xilinx_cxt_t  *fpga_xilinx_cxt)
{
    int ch;
    ch=fpga_xilinx_cxt->last_tx_ch+1;
    if(ch==FPGA_XILINX_MAX_VIDEO_DESC_LIST_COUNT)
          ch=0;
    return (fpga_xilinx_cxt->tx_desclist[ch]==NULL);
}

static void fpga_xilinx_video_DPC_func(void *data)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=data;
    
    debug_msg("^");
    if(!queue_empty(&fpga_xilinx_cxt->done_desc_queue))
    {
        fpga_xilinx_desclist_item_t *desclist_item;
        do
        {
            desclist_item=fpga_xilinx_get_done_desclist(fpga_xilinx_cxt);

            if(desclist_item)
            {
                debug_msg("-"); 
#if 1
                desclist_item->desc_count=0;
                fpga_xilinx_add_free_desclist(fpga_xilinx_cxt,desclist_item); 
#endif
                if(desclist_item->callback)
                    desclist_item->callback(desclist_item->cb_cxt);
#if 0
                desclist_item->desc_count=0;
                fpga_xilinx_add_free_desclist(fpga_xilinx_cxt,desclist_item); 
#endif
            }
        }while(desclist_item);
    }
    if(!queue_empty(&fpga_xilinx_cxt->ready_desc_queue))
    {
        
        while(fpga_xilinx_is_TX_available(fpga_xilinx_cxt))
        {
            fpga_xilinx_desclist_item_t *ready_desclist=NULL;
            ready_desclist=fpga_xilinx_get_ready_desclist(fpga_xilinx_cxt);
            if(ready_desclist)
            {
                fpga_xilinx_tranfer_desclist(fpga_xilinx_cxt,ready_desclist);
                debug_msg("+");
            }
            else
            {
                break;
            }
        }       
    }
    sys_atomic_set(fpga_xilinx_cxt->atomic_VDPC_scheduling,0);
    debug_msg("%0x$",(fpga_xilinx_ReadRegByte(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_CTRL)>>1));
}

static void fpga_xilinx_video_term_DPC_func(void *data)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=data;
    fpga_xilinx_desclist_item_t *desclist_item;
    int ch;
    
    for(ch=0;ch<FPGA_XILINX_MAX_VIDEO_DESC_LIST_COUNT;ch++)
        {
           desclist_item=fpga_xilinx_cxt->tx_desclist[ch];
           if(desclist_item)
           {
               desclist_item->desc_count=0;
                fpga_xilinx_cxt->tx_desclist[ch]=NULL;
                if(desclist_item->callback)
                    desclist_item->callback(desclist_item->cb_cxt);
                fpga_xilinx_add_free_desclist(fpga_xilinx_cxt,desclist_item);
           }
        }
        while(!queue_empty(&fpga_xilinx_cxt->ready_desc_queue))
        {
            desclist_item=fpga_xilinx_get_ready_desclist(fpga_xilinx_cxt);
            if(desclist_item)
            {
                if(desclist_item->callback)
                    desclist_item->callback(desclist_item->cb_cxt);
                fpga_xilinx_add_free_desclist(fpga_xilinx_cxt,desclist_item);
            }
        }
        while(!queue_empty(&fpga_xilinx_cxt->done_desc_queue))
        {
            desclist_item=fpga_xilinx_get_done_desclist(fpga_xilinx_cxt);
            if(desclist_item)
            {
                if(desclist_item->callback)
                    desclist_item->callback(desclist_item->cb_cxt);
                fpga_xilinx_add_free_desclist(fpga_xilinx_cxt,desclist_item);
            }
        }
        fpga_xilinx_cxt->last_tx_ch=-1;
        fpga_xilinx_cxt->is_streaming=FALSE;
}

static void fpga_xilinx_vip_set_scaler(fpga_xilinx_cxt_t *fpga_xilinx_cxt,U32_T width_out,U32_T height_out,U32_T width_in,U32_T height_in,BOOL_T is_interlace)
{
	U32_T Data;
	
	if (width_in == 0 || height_in == 0)
		    printk("%s error parameters\n",__func__);
	
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_CTRL,0);
	printk("%s in %dx%d ==> out %dx%d\n",__func__,width_in,height_in,width_out,height_out);
	if(is_interlace)
	{
		height_in = height_in;
		height_out = height_out;
	}
	else
	{
		height_in = height_in;
		height_out = height_out;
	}
		    	   
	Data = 0x100000 * width_in / width_out;
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_HSHRINK_FACTOR,Data);
	
	printk("%s HSHRINK=%x\n",__func__,fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_HSHRINK_FACTOR));
	
	if(is_interlace)
	{ 
		Data = 0x100000 * (height_in*2) / height_out;  
	}
	else
	{
		Data = 0x100000 * height_in / height_out;
	}
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_VSHRINK_FACTOR,Data);
	
	printk("%s VSHRINK=%x\n",__func__,fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_VSHRINK_FACTOR));
	
	if(is_interlace)
	{ 
	    //set input Size
	    Data = (((height_in*2) << 16) & 0xffff0000) + (width_in & 0xffff);
	}
	else
	{
		Data = ((height_in << 16) & 0xffff0000) + (width_in & 0xffff);
	}
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_SOURCE_VIDEO_SIZE,Data);
	  
	   //set x clip area pos (zero base)
    Data = (((width_in - 1) << 16) & 0xffff0000);
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_H_APERTURE_START_PIXEL,Data);
    
    if(is_interlace)
    {
	   //set y clip area pos (zero base)
	    Data = ((((height_in - 1)*2) << 16) & 0xffff0000);
	}
	else
	{
		Data = (((height_in - 1) << 16) & 0xffff0000);
	}
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_V_APERTURE_START_LINE,Data);
    
    if(is_interlace)
    {
	   //set output Size
	    Data = ((height_out << 16) & 0xffff0000) + (width_out & 0xffff);
	}
	else
	{
		Data = ((height_out << 16) & 0xffff0000) + (width_out & 0xffff);
	}	
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_OUTPUT_SIZE,Data);
	    
	   // constant 
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_CONSTANT1,0x4040);
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_CONSTANT2,0);
}

static void fpga_xilinx_vip_set_framerate(fpga_xilinx_cxt_t   *fpga_xilinx_cxt,int in_framerate,int out_framerate)
{
	U64_T framerate_target=148500000;
	U32_T fraterate_cnt;
	U32_T FPS=0;
	U32_T framerate_target_by100;
	
    if(out_framerate==0) //refer to input signal 
    {
        fraterate_cnt = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_FRAME_COUNT);
        #if 0 //special case
        if (fraterate_cnt !=0)
        {
            FPS = 125000000 / fraterate_cnt;  // 125000000 : crystal clock
            if ((FPS > 85) || (FPS < 24))
               FPS = 60; 
		}
        else
        {
            FPS = 60; 
		}
        #endif  
        
        FPS = 125000000 / (fraterate_cnt/100);
          
        framerate_target = framerate_target*100/FPS;
   
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_FRAME_RATE_TARGET,framerate_target);
        framerate_target = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_FRAME_RATE_TARGET);
        framerate_target_by100 = framerate_target/100;
        printk("%s Final FPS... %lld\n",__func__,148500000/framerate_target);
    }
    else //refer to output setting
    {		
        framerate_target/=out_framerate;
      
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_FRAME_RATE_TARGET,framerate_target);
        framerate_target = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_FRAME_RATE_TARGET);
	}
	
#if 1 //fix sync input mode
	if (in_framerate == out_framerate)
	    fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,0x20,0x20); // lock mode 20170503
	else
	    fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,0x20,0); // free run mode 20170503  
#endif
	
}

static void fpga_xilinx_vip_set_clip(fpga_xilinx_cxt_t *fpga_xilinx_cxt,int left,int top,int width,int height,BOOL_T is_interlace)
{
	U32_T clip_height,clip_top;
	int status0=0;
	int status1=0;
	
    if(is_interlace)
    {
		clip_height = height | (height << 16);
		clip_top = top/2;
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_LEFT_OFFSET,left);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_WIDTH,width);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_TOP_OFFSET,clip_top);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_HEIGHT,clip_height);
        //printk("%s interlaced %dx%d\n",__func__,width,clip_height & 0x0000ffff);
        status0 = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_HEIGHT) & 0x0000ffff;
        status1 = (fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_HEIGHT) >> 16) & 0xffff;
    }
    else
    {
        clip_height = height;
		clip_top = top;
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_LEFT_OFFSET,left);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_WIDTH,width);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_TOP_OFFSET,clip_top);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_HEIGHT,clip_height);
        //printk("%s progress... %dx%d\n",__func__,width,clip_height);
    }
    //sys_msleep(50);
}

static void fpga_xilinx_vip_bypass_scaler(fpga_xilinx_cxt_t *fpga_xilinx_cxt,BOOL_T disable,BOOL_T is_interlace)
{
	//scaler mode or non scaler mode
    if(disable)
    {
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,0x80,0);     
    }
    else
    {
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,0x80,0x80); 
    }    
}

static void fpga_xilinx_vip_enable_scaler_output(fpga_xilinx_cxt_t *fpga_xilinx_cxt,BOOL_T enable)
{
    if(enable)
    {
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_CTRL,0x03,3); 
    }
    else
    {
	    fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SCALER_CTRL,0x03,0); 
    }
}

static void fpga_xilinx_vip_enable_clip_output(fpga_xilinx_cxt_t *fpga_xilinx_cxt,BOOL_T enable)
{
	fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,0x40,0);
    
    if(enable)
    {  
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,CLIP_START_MASK,1); 
    }
    else
    {    
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,CLIP_START_MASK,0); 
    }    
}

static void fpga_xilinx_dual_pixel(handle_t fpga_xilinx_handle,int csc_dual_pixel)
{
	fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
	
    printk("%s dual pixel %d\n",__func__,csc_dual_pixel);
    
    fpga_xilinx_cxt->dual_pixel = csc_dual_pixel;
   
    if (csc_dual_pixel ==1)
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,CSC_DUAL_PIXEL_EN_MASK,CSC_DUAL_PIXEL_EN_MASK);
    else
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,CSC_DUAL_PIXEL_EN_MASK,0);   
        
    //printk("%s dual pixel (0x%2x)\n",__func__,fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL) & 0x20); 
}	

static void fpga_xilinx_video_bypass(handle_t fpga_xilinx_handle,int video_bypass)
{
	fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
	
    printk("%s video bypass %d\n",__func__,video_bypass);
   
    if (video_bypass ==1)
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_VIDEO_BYPASS,VIDEO_BYPASS_EN_MASK,1);
    else
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_VIDEO_BYPASS,VIDEO_BYPASS_EN_MASK,0); 
        
     //printk("%s video bypass (0x%2x)\n",__func__,fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_VIDEO_BYPASS) & 0x01);        
}	

static void fpga_xilinx_csc_in_out(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_colorspace_e colorspace_in,fpga_xilinx_colorspace_e colorspace_out,BOOL_T BT709_out,int ioctl_yuyv)
{
	int data=0;
    debug_msg("%s cs %d\n",__func__,colorspace_out);
    
    {  
        switch(colorspace_out)
        {
        case FPGA_XILINX_CS_YUV422:
        {  
			printk("%s FPGA_XILINX_CS_YUV\n",__func__);
			fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,VIDEO_OUT_FMT,0x00); //video out color format
			//fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV422_TO_444_EN_MASK,0x01); //move and check with infoframe
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,RGB_TO_YUV_EN_MASK,0);
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV_TO_YUV_EN_MASK,0); //video in keep ori color format
			
			if (colorspace_in == FPGA_XILINX_CS_RGB_LIMITED) //now ADV7619 only transfer the colorspace to default YUYV
			{
				fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV422_TO_444_EN_MASK,0);
                fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,RGB_TO_YUV_EN_MASK,RGB_TO_YUV_EN_MASK);	
                fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,RGB_TO_YUV_COLOR_FMT,BIT(8));
			}
            else
            {
                //printk("%s===== colorspace_in FPGA_XILINX_CS_YUV\n",__func__);
            }
            data=fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL);
        }
           break; 
        case FPGA_XILINX_CS_RGB_FULL: //YUV from ADV7619 
        case FPGA_XILINX_CS_RGB_LIMITED:
        {
			printk("%s FPGA_XILINX_CS_RGB\n",__func__);
			fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,VIDEO_OUT_FMT,BIT(9));
			//data=(fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL) & 0x0f00)>>8;
			data=fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL);
			fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV422_TO_444_EN_MASK,1);
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,RGB_TO_YUV_EN_MASK,0);
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV_TO_RGB_COLOR_FMT,0);
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV_TO_RGB_EN_MASK,BIT(3));
            
            //printk("%s>>>>>>>> cs %s(%x) BT709_out=%d\n",__func__,"FPGA_XILINX_CS_RGB",data,BT709_out);
        }
            break;
        default:
            break;
        }
      
        if (ioctl_yuyv == 0) //able to be updated from ioctl
        {
			printk("%s ioctl to uyvy\n",__func__);
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,0xff00,0x0900);   
		}
        else
        {
			printk("%s ioctl to yuyv\n",__func__);
            fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,0xff00,0x0000);      
		}
    }
    sys_msleep(50);
}

static void fpga_xilinx_audio_rate_input(fpga_xilinx_cxt_t *fpga_xilinx_cxt,fpga_xilinx_audio_e audio_rate)
{
     
    switch(audio_rate)
    {
		case FPGA_XILINX_AUDIO_32000:
		//fpga_xilinx_cxt->audio_info.audio_buf_size = 7680;
		fpga_xilinx_cxt->audio_info.audio_buf_size = 3840;
		break;
		case FPGA_XILINX_AUDIO_44100:
		//fpga_xilinx_cxt->audio_info.audio_buf_size = 10584;
		fpga_xilinx_cxt->audio_info.audio_buf_size = 5296;
		break;
		case FPGA_XILINX_AUDIO_48000:
		//fpga_xilinx_cxt->audio_info.audio_buf_size = 11520;
		fpga_xilinx_cxt->audio_info.audio_buf_size = 5760;
		break;
		case FPGA_XILINX_AUDIO_88200:
		fpga_xilinx_cxt->audio_info.audio_buf_size = 21168;
		break;
		case FPGA_XILINX_AUDIO_96000:
		fpga_xilinx_cxt->audio_info.audio_buf_size = 5760;  // avoid dma error
//		fpga_xilinx_cxt->audio_info.audio_buf_size = 23040;
		break;
		case FPGA_XILINX_AUDIO_176400:
		fpga_xilinx_cxt->audio_info.audio_buf_size = 42336;
		break;
		case FPGA_XILINX_AUDIO_192000:
		fpga_xilinx_cxt->audio_info.audio_buf_size = 46080;
		break;
		default:
		break;
	}
    printk("%s audio_buf_size=%d\n",__func__,fpga_xilinx_cxt->audio_info.audio_buf_size);  
}

void fpga_xilinx_config_video_process(handle_t fpga_xilinx_handle,fpga_xilinx_video_process_cfg_t *vip_cfg)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    BOOL_T BT709_out;
    int pixel_width_set;
    long int fraterate_cnt;
    printk("%s...\n",__func__);
    fraterate_cnt = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_FRAME_COUNT);
    if (fraterate_cnt !=0) //for reference
    {  
		fraterate_cnt = fraterate_cnt/100;     
	}
	
    fpga_xilinx_reset(fpga_xilinx_cxt);
    //fpga_xilinx_cxt->can_be_terminated=FALSE;
	
    if(vip_cfg)
    {
		fpga_xilinx_vip_enable_clip_output(fpga_xilinx_cxt, FALSE); 
		fpga_xilinx_vip_enable_scaler_output(fpga_xilinx_cxt, FALSE); 
		
        fpga_xilinx_audio_rate_input(fpga_xilinx_cxt,vip_cfg->audio_rate); 
        
        if ((vip_cfg->dual_pixel ==1)  && (vip_cfg->video_bypass ==0))
            pixel_width_set = vip_cfg->in_framesize.width/2;
        else
            pixel_width_set = vip_cfg->in_framesize.width;    
        
        printk("%s in:%dx%d%c out: %dx%d\n", __func__, pixel_width_set,vip_cfg->in_framesize.height,(vip_cfg->is_interlace) ?'i':'p',vip_cfg->out_framesize.width,vip_cfg->out_framesize.height);
        
        if (vip_cfg->out_framesize.height>576)
            BT709_out=TRUE;
        else
            BT709_out=FALSE;
            
        // support YUV422 and RGB444     
        if (vip_cfg->pixel_format == FPGA_XILINX_FMT_YUYV)
            vip_cfg->colorspace = FPGA_XILINX_CS_YUV422;
        if (vip_cfg->pixel_format == FPGA_XILINX_FMT_RGB3)
            vip_cfg->colorspace = FPGA_XILINX_CS_RGB_FULL; 
         
        if ((vip_cfg->packet_colorspace ==0) || (vip_cfg->dual_pixel ==0))  //CS_YUV
        {  
            fpga_xilinx_csc_in_out(fpga_xilinx_cxt,FPGA_XILINX_CS_YUV422/*input*/,vip_cfg->colorspace/*output*/,BT709_out,vip_cfg->ioctl_YUYV);  
            printk("%s======= input FPGA_XILINX_CS_YUV\n",__func__);
		}
        else
        {
            fpga_xilinx_csc_in_out(fpga_xilinx_cxt,FPGA_XILINX_CS_RGB_LIMITED,vip_cfg->colorspace,BT709_out,vip_cfg->ioctl_YUYV);
            printk("%s======== input FPGA_XILINX_CS_RGB_LIMITED\n",__func__);
		}
		
		if (vip_cfg->YUV422_flag ==1)
		    fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV422_TO_444_EN_MASK,1);
		else
	        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL,YUV422_TO_444_EN_MASK,0);
	        
		fpga_xilinx_dual_pixel(fpga_xilinx_cxt,vip_cfg->dual_pixel);        

        if (vip_cfg->valid_mask & CLIP_CFG_VALID_MASK)
        {
            fpga_xilinx_vip_set_clip(fpga_xilinx_cxt, vip_cfg->clip_start.x, vip_cfg->clip_start.y, vip_cfg->clip_size.width, vip_cfg->clip_size.height, vip_cfg->is_interlace);
        }
        else
        {
            fpga_xilinx_vip_set_clip(fpga_xilinx_cxt, 0, 0, /*vip_cfg->in_framesize.width*/pixel_width_set, vip_cfg->in_framesize.height, vip_cfg->is_interlace);
        }

        if (vip_cfg->valid_mask & SCALER_CFG_VALID_MASK) 
        {
            printk("%s ========scale to  (%d,%d) \n", __func__, vip_cfg->out_framesize.width, vip_cfg->out_framesize.height);
            fpga_xilinx_vip_set_scaler(fpga_xilinx_cxt, vip_cfg->out_framesize.width, vip_cfg->out_framesize.height, /*vip_cfg->in_framesize.width*/pixel_width_set, vip_cfg->in_framesize.height,vip_cfg->is_interlace);

        }
        else
        {
            fpga_xilinx_vip_set_scaler(fpga_xilinx_cxt, /*vip_cfg->in_framesize.width*/pixel_width_set, vip_cfg->in_framesize.height, vip_cfg->in_framesize.width, vip_cfg->in_framesize.height,vip_cfg->is_interlace);
        }
        
        fpga_xilinx_video_bypass(fpga_xilinx_cxt,vip_cfg->video_bypass);    

        if (vip_cfg->valid_mask & FRAMERATE_CFG_VALID_MASK)
        {
            fpga_xilinx_vip_set_framerate(fpga_xilinx_cxt, vip_cfg->in_framerate,vip_cfg->out_framerate);
        }
        else
        {
            fpga_xilinx_vip_set_framerate(fpga_xilinx_cxt, vip_cfg->in_framerate,vip_cfg->out_framerate=0);
        }
        
        if (vip_cfg->valid_mask & SCALER_CFG_VALID_MASK)
        {
            {
				if (vip_cfg->is_interlace ==1)
				{
                    if ((vip_cfg->in_framesize.width == vip_cfg->out_framesize.width) && ((2*vip_cfg->in_framesize.height) == vip_cfg->out_framesize.height))
                    {
                        fpga_xilinx_vip_bypass_scaler(fpga_xilinx_cxt, TRUE, vip_cfg->is_interlace);
                        printk("%s ==disable scaler \n", __func__);
                    }
                    else
                    {
                        fpga_xilinx_vip_bypass_scaler(fpga_xilinx_cxt, FALSE, vip_cfg->is_interlace); 
                        printk("%s == enable scaler \n", __func__);
                    }
				}
				else
				{
					if ((vip_cfg->in_framesize.width == vip_cfg->out_framesize.width) && (vip_cfg->in_framesize.height == vip_cfg->out_framesize.height))
                    {
                        fpga_xilinx_vip_bypass_scaler(fpga_xilinx_cxt, TRUE, vip_cfg->is_interlace);
                        printk("%s ===== disable scaler \n", __func__);
                    }
                    else
                    {
                        fpga_xilinx_vip_bypass_scaler(fpga_xilinx_cxt, FALSE, vip_cfg->is_interlace); 
                        printk("%s ===== enable scaler \n", __func__);
                    }
				}

            }

        }
        else
        {
            fpga_xilinx_vip_bypass_scaler(fpga_xilinx_cxt, TRUE, vip_cfg->is_interlace);
        }

        fpga_xilinx_vip_enable_scaler_output(fpga_xilinx_cxt, TRUE);
        fpga_xilinx_vip_enable_clip_output(fpga_xilinx_cxt, TRUE);

    } 
    
    //printk("%s color control reg=0x%02x\n",__func__,fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CSC_COLORCONTROL) & 0xffff);
}

void fpga_xilinx_reset(handle_t fpga_xilinx_handle)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
   
    int Status = 0;
	int cnt = 0;
    
    {
		fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,CLIP_START_MASK,0);
                //fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GLB_TX_AV_ENABLE,0x00); 
    
		do
		{
			sys_msleep(10);

			if (10 == ++cnt)   // delay
				break ;   

		} while(fpga_xilinx_cxt->can_be_terminated==FALSE);
        printk(">>>>>>>>>>>>>>>>cnt1=%d\n",cnt);
        fpga_xilinx_cxt->can_be_terminated=FALSE;
        //fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_CTRL,VIDEO_DMA_ENABLE,0);
        //sys_msleep(20);
        { 
                fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_RESET, 1);
	        cnt = 0;
	        do
	        {
		        Status = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_RESET);
		        Status &= 0x8;
                sys_msleep(20);
	        } while ((Status == 0) && ++cnt < 10);
	        sys_msleep(20);	// wait for at lease 20 ms
        }
        //if (fpga_xilinx_cxt->is_streaming)
        //    fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GLB_TX_AV_ENABLE,0x02); 
    } 
}

void fpga_xilinx_get_cardid(handle_t fpga_xilinx_handle, U32_T *card_id)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
   
	U32_T read_tmp=0xffffffff;
	U32_T read_version;
	
	read_version = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VERSION);
	
	if (read_version == 0x20171117)
	{
	
	    //read_tmp = fpga_xilinx_ReadReg(fpga_xilinx_cxt, REG_FPGA_BRIDGE_750_GPIO);

	    read_tmp = ((fpga_xilinx_ReadReg(fpga_xilinx_cxt, REG_FPGA_BRIDGE_750_GPIO) >> 6) & 0x03);
	}
	else if (read_version == 0x20181106)
	{
		read_tmp = ((fpga_xilinx_ReadReg(fpga_xilinx_cxt, REG_FPGA_BRIDGE_750_GPIO) >> 8) & 0x0f);
	}
	else
	{
		printk("%s fpga version not correct=%08x\n",__func__,read_tmp);
	}
	
	printk("%s card_id...=%x\n",__func__,read_tmp);
    
    *card_id = read_tmp;
}

static int fpga_xilinx_spi_done(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int wait)
{
	int addr;
	int data;
	int i;
	for (i=0; i < wait; i++)
	{
		addr = 0x404;//XILINX_SPI_CONTROL
		data  = 0x0;	// enable
		data=fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
		//REGReadN(dev, addr, &buf,4);
		if (data & 0x8)		// bit 3
		{
			//printk("fpga_xilinx_spi_done ok\n");
			sys_msleep(1);
			return BWFS_EVENT_NO_ERROR;
		}
		sys_msleep(1);
	}
	printk("fpga_xilinx_spi_done fail\n");
	return BWFS_EVENT_TIMEOUT;
}

static int fpga_xilinx_flash_pageread(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,U8_T *buffer,U32_T offset)
{
	int addr;
	int data;		
	//char  tmp[256];
	int ret=0;
	int i;

	// set clock speed
	addr = 0x400;//XILINX_SPI_CLK
	data  = 2;	// writ clock speed
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	// rx fifo reset
	addr = 0x418;//XILINX_SPI_RX_FIFO
	data  = 1;	// bit 0: rx fifo reset
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);


	addr = 0x408;//XILINX_SPI_COMMAND
	data  = 0x3;	// read Flash command
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);
	
	addr = 0x40c;//XILINX_SPI_ADDRESS
	data  = offset;	// 24 bit address
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	addr = 0x410;//XILINX_SPI_RX_COUNT
	data  = 256;	// read 256 bytes
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	addr = 0x414;//XILINX_SPI_TX_COUNT
	data  = 0;	// write 0 bytes
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	// start command
	addr = 0x404;//XILINX_SPI_CONTROL
	data  = 0x4 + 2;	//  start command, bit 1, // set address mode 1, using 24bit address mode - bit 2
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	// wait command done
	//BOOL ret = XilinxWaitSPIDone(dev, 500);
	ret = fpga_xilinx_spi_done(fpga_xilinx_cxt,500)	;

	if (ret==0)
	{
		for (i= 0; i<256; i+=4)
		{
			addr = 0x418;//XILINX_SPI_RX_FIFO
			data  = 2;	// issue read command
			fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
			//REGWriteN(dev, addr, &buf, 4);

			addr = 0x41c;//XILINX_SPI_RX_FIFO_DATA
			data=0;
			data=fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
			//REGReadN(dev, addr, &buf, 4);		
			*(&buffer[i+0]) = ((data>>24) & 0xff);
			*(&buffer[i+1]) = ((data>>16) & 0xff);
			*(&buffer[i+2]) = ((data>>8 ) & 0xff);
			*(&buffer[i+3]) = ((data>>0 ) & 0xff);
		}
	}
	else
	{
		printk("fpga_xilinx_flash_pageread, Xilinx SPI Fail !!!!!");
	}

    return ret;
}

int fpga_xilinx_flash_dump(handle_t fpga_xilinx_handle,int start_block, int blocks, U8_T *buffer)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    int i=0;
    int ret=0;
    
    for (i=0; i< blocks; i++)
    {
		ret = fpga_xilinx_flash_pageread(fpga_xilinx_cxt,buffer+i*256,i*256);
	}
    
    return 0;
}

static int fpga_xilinx_isp_enable(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int enable)
{
	int ret=0; 
	int addr;
	int data;
	
	addr = 0x408; //XILINX_SPI_COMMAND
	data = enable ? 0x06:0x04;
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	data = fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
	printk("%s spi_command =%x\n",__func__,data);
	
	addr = 0x40c; //XILINX_SPI_ADDRESS
	data = 0; //offset 0
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	addr = 0x410; //XILINX_SPI_RX_COUNT
	data = 0; 
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	addr = 0x414; //XILINX_SPI_TX_COUNT
	data = 0; 
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	addr = 0x404; //XILINX_SPI_CONTROL
	data = 0+2; //  start command, bit 1, not use address, 1: use 24bit address bit-2
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	data = fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
	printk("%s spi_control =%x\n",__func__,data);
	
	ret = fpga_xilinx_spi_done(fpga_xilinx_cxt,50);
	
	if (ret!=0)
	{
		printk("fpga_xilinx_isp_enable, Xilinx SPI Fail !!!!!\n");
	}
	return ret;
}

static int fpga_xilinx_rdsr_read(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,U32_T *tmp)
{
	int addr;
	int data;		
	int ret=0;
	
	// rx fifo reset
	addr = 0x418;//XILINX_SPI_RX_FIFO
	data  = 1;	// bit 0: rx fifo reset
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	addr = 0x408;//XILINX_SPI_COMMAND
	data  = 0x5;	// Read Status Register (RDSR)
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	addr = 0x40c;//XILINX_SPI_ADDRESS
	data  = 0;	// not using 24 bit address
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	addr = 0x410;//XILINX_SPI_RX_COUNT
	data  = 1;	// read 1 byte
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	addr = 0x414;//XILINX_SPI_TX_COUNT
	data  = 0;	// write 0 bytes
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	// start command
	addr = 0x404;//XILINX_SPI_CONTROL
	data  = 0 + 2;	//  start command, bit 1
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	// wait command done
	ret = fpga_xilinx_spi_done(fpga_xilinx_cxt,50);

	if (ret==0)
	{
		addr = 0x418;//XILINX_SPI_RX_FIFO
		data  = 2;	
		fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
		
		addr = 0x41c;//XILINX_SPI_RX_FIFO_DATA
		data  = 0;	
		data = fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
		*tmp = (U32_T)(data & 0xff);
        //printk("fpga_xilinx_rdsr_read, Xilinx SPI OK !!!!!\n");
		
	}
	else
	{
		printk("fpga_xilinx_rdsr_read, Xilinx SPI Fail !!!!!\n");
	}

    return ret;
}

static int fpga_xilinx_flash_pageerase(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,int type,U32_T address)
{
	int addr;
	int data;		
	int ret=0;
	int i;
	U8_T erase_cmd;
	U32_T wait=0;
	U32_T status=0;

    printk("fpga_xilinx_flash_pageerase start !!!!!\n");
	// set clock speed
	addr = 0x400;//XILINX_SPI_CLK
	data  = 2;	// writ clock speed
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	switch(type)
	{
		case 0:
		    erase_cmd = 0xc7; //address=0
		    wait = 0x10000; //10 sec
		    break;
		case 1:
		    erase_cmd = 0xd8; //address 64k boundary
		    wait = 0x3000; //10 sec
		    break;
		case 2:
		    erase_cmd = 0x20; //address 4k boundary
		    wait = 0x1000; //1 sec
		    break; 
		default:
		    return BWFS_EVENT_PARMS_ERROR;           
	}


	addr = 0x408;//XILINX_SPI_COMMAND
	data  = erase_cmd;	// read Flash command
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	data = fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
	printk("%s spi_command... =%x\n",__func__,data);
	
	addr = 0x40c;//XILINX_SPI_ADDRESS
	data  = address;	// 24 bit address
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	data = fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
	printk("%s spi_address... =%x\n",__func__,data);

	addr = 0x410;//XILINX_SPI_RX_COUNT
	data  = 0;	// read 256 bytes
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	addr = 0x414;//XILINX_SPI_TX_COUNT
	data  = 0;	// write 0 bytes
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	// start command
	addr = 0x404;//XILINX_SPI_CONTROL
	data  = 0x4 + 2;	//  using 24bit address, bit 2, Start
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	data = fpga_xilinx_ReadReg(fpga_xilinx_cxt,addr);
	printk("%s spi_control... =%x\n",__func__,data);

	// wait command done

	//ret = fpga_xilinx_spi_done(fpga_xilinx_cxt,500)	;

	//if (ret==0) 
	{
		for (i= 0; i<wait; i++)
		{
			ret = fpga_xilinx_rdsr_read(fpga_xilinx_cxt,&status);		

			if ((ret==0) && ((status & 0x1)==0))
			{
				printk("fpga_xilinx_flash_pageerase done !!!!!\n");
				return 0;
			}
			sys_msleep(1);
		}
		printk("fpga_xilinx_flash_pageerase fail !!!!!");
	}
	return ret;
}

int fpga_xilinx_flash_erase(fpga_xilinx_cxt_t  *fpga_xilinx_cxt)
{
	int block = (0x200000+(64*1024)) / (64*1024); //erase 2M, every 64k bytes
	//int block = 1; //test erase 64k, every 64k bytes
	int i;
	int ret=0;
	printk("%s start..................\n",__func__);
	for (i=0; i<block; i++)
	{
		fpga_xilinx_isp_enable(fpga_xilinx_cxt, 1); //enable
		ret = fpga_xilinx_flash_pageerase(fpga_xilinx_cxt, 1, i*(64*1024));
		fpga_xilinx_isp_enable(fpga_xilinx_cxt, 0); //disable
		if (ret !=0)
		   break;
	}
	printk("fpga_xilinx_flash_erase done !!!!!");
	return ret;
}

static int fpga_xilinx_flash_pagewrite(fpga_xilinx_cxt_t  *fpga_xilinx_cxt,U8_T *buffer,U32_T offset)
{
	int addr;
	int data;
	int status;		
	int ret=0;
	int i;
	U32_T value;

	// set clock speed
	addr = 0x400;//XILINX_SPI_CLK
	data  = 2;	// writ clock speed
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	// rx fifo reset
	addr = 0x418;//XILINX_SPI_RX_FIFO
	data  = 1;	// bit 0: rx fifo reset
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	//put data into FPGA buf
	for (i= 0; i<256; i+=4)
	{
		value= (buffer[i+3] << 24) | (buffer[i+2] << 16) | (buffer[i+1] << 8) | buffer[i] ;
		addr = 0x424;//XILINX_SPI_TX_FIFO_DATA
		fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,value);
					
		addr = 0x420;//XILINX_SPI_TX_FIFO;
		data  = 2;	// issue write command
		fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	}

	addr = 0x408;//XILINX_SPI_COMMAND
	data  = 0x2;	// write Flash command
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	
	addr = 0x40c;//XILINX_SPI_ADDRESS
	data  = offset;	// 24 bit address
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);
	//REGWriteN(dev, addr, &buf, 4);

	addr = 0x410;//XILINX_SPI_RX_COUNT
	data  = 0;	// 
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	addr = 0x414;//XILINX_SPI_TX_COUNT
	data  = 256;	// write 256 bytes
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	// start command
	addr = 0x404;//XILINX_SPI_CONTROL
	data  = 0x4 + 2;	//  start command, bit 1, // set address mode 1, using 24bit address mode - bit 2
	fpga_xilinx_WriteReg(fpga_xilinx_cxt,addr,data);

	// wait command done
	//BOOL ret = XilinxWaitSPIDone(dev, 500);
	ret = fpga_xilinx_spi_done(fpga_xilinx_cxt,2000)	;

	if (ret==0)
	{
		for (i= 0; i<2000; i++)
		{
			status =0;
			ret = fpga_xilinx_rdsr_read(fpga_xilinx_cxt,&status);	
			if ((ret==0) && ( (status & 0x1) ==0 ))
			{
				//printk("fpga_xilinx_flash_pagewrite done !!!!!\n");
				return 0;
			}
			sys_msleep(1);
		}
		printk("fpga_xilinx_flash_pagewrite fail !!!!!");
	}
	else
	{
		printk("fpga_xilinx_flash_pagewrite, Xilinx SPI Fail !!!!!");
	}

    return ret;
}

int fpga_xilinx_flash_update(handle_t fpga_xilinx_handle,int start_block, int blocks, U8_T *buffer)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
	char retbuf[256]; //64k
	int ret;
    int block = blocks;
    char *p = buffer;
    int i,j,k;
    
    fpga_xilinx_flash_erase(fpga_xilinx_cxt);
    
	//int block =  (data_size+255)/256;		// update 8M byte
	printk("%s start..................\n",__func__);
	// we need enable/desable evertime we do a page write
	// loop every 256 bytes
	for (i=0; i< block; i++)
	{
		
		fpga_xilinx_isp_enable(fpga_xilinx_cxt, 1);
		
		ret = fpga_xilinx_flash_pagewrite(fpga_xilinx_cxt, (buffer+i*256), i*256); // write 8M bytes from  8M byte positon
		sys_msleep(1);
		fpga_xilinx_isp_enable(fpga_xilinx_cxt, 0);
		printk("i=%d,ret=%d\n",i,ret);
		if (ret !=0)
			return ret;
	}
    printk("%s write completely..................\n",__func__);
	for (j=0; j< block; j++)
	{
		// the read command munst issue when ISP been disable
		ret = fpga_xilinx_flash_pageread(fpga_xilinx_cxt, /*(PBYTE)*/ &retbuf[0], j*256);// read 8M byte from  8M position
		sys_msleep(1);
		printk("j=%d,ret=%d\n",j,ret);
		if (ret !=0)
		{
			printk("read to compare: read fail\n");
			return ret;
		}

		for (k=0; k<256; k++)		// read to compare if error 
		{
			if (retbuf[k] != *p)
			{
				if (j < block)
				    printk("read to compare: fail\n");
				    
				if (j == block)
				    printk("last block compare: fail\n");    
				return BWFS_EVENT_DATA_ERROR;
			}
			p++;
		}
	}
	printk("%s verify completely..................\n",__func__);
    return 0;
}

void fpga_xilinx_enable_video_streaming(handle_t fpga_xilinx_handle,BOOL_T enable )
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    fpga_xilinx_desclist_item_t *desclist_item;
    int ch;
  
    ch=fpga_xilinx_cxt->last_tx_ch+1;
    if(ch==FPGA_XILINX_MAX_VIDEO_DESC_LIST_COUNT)
        ch=0;
    
    if(enable)
    {
        if (fpga_xilinx_cxt->is_streaming)
		return;
        do
        {
            desclist_item=fpga_xilinx_get_ready_desclist(fpga_xilinx_cxt);
            if(desclist_item)
                fpga_xilinx_tranfer_desclist(fpga_xilinx_cxt,desclist_item);
    
        }while(desclist_item);
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_SGDMA_CTRL,VIDEO_DMA_ENABLE,VIDEO_DMA_ENABLE);
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,CLIP_START_MASK,1);
        fpga_xilinx_cxt->is_streaming=TRUE;
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GLB_TX_AV_ENABLE,0x02); 
    }
    else
    {
        if (!fpga_xilinx_cxt->is_streaming)
		return;
	
		//fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_CLIPPER_CTRL,CLIP_START_MASK,0);
        
		fpga_xilinx_reset(fpga_xilinx_cxt);
        //fpga_xilinx_cxt->can_be_terminated=FALSE; //check by irq for fpga reset
        fpga_xilinx_cxt->is_streaming=FALSE;
    } 
}

void fpga_xilinx_start_audio_streaming(handle_t fpga_xilinx_handle,fpga_xilinx_audio_cbinfo_t *cb_info)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    int ch;
    
    if(cb_info)
    {
        fpga_xilinx_cxt->audio_info.audio_cb=cb_info->callback;
        fpga_xilinx_cxt->audio_info.audio_cb_cxt=cb_info->asso_cxt;
    }
    else
    {
        fpga_xilinx_cxt->audio_info.audio_cb=NULL;
        fpga_xilinx_cxt->audio_info.audio_cb_cxt=NULL;
    }
    fpga_xilinx_WriteReg(fpga_xilinx_cxt,0x00000218, 0);
    fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUDIO_MISC, 0);
        
    fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_SIZE,(fpga_xilinx_cxt->audio_info.audio_buf_size/4));
    
    //printk("%s %08x\n",__func__,fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_SIZE));
    for(ch=0;ch<FPGA_XILINX_AUDIO_BUF_COUNT;ch++)
    {
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_BUFFER_ADDR_L(ch),fpga_xilinx_cxt->audio_info.dma_buffer[ch]->phys_addr& 0xffffffff);
        fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_AUD_DMA_BUFFER_ADDR_H(ch),(fpga_xilinx_cxt->audio_info.dma_buffer[ch]->phys_addr >> 32));
    }
    
    fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GLB_TX_AV_ENABLE,0x02); 
}

void fpga_xilinx_stop_audio_streaming(handle_t fpga_xilinx_handle)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    
    fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GLB_TX_AV_ENABLE,0);
}

void fpga_xilinx_get_frameinfo(handle_t fpga_xilinx_handle,fpga_xilinx_frame_info_t *frameinfo,int *stable)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=fpga_xilinx_handle;
    U32_T height0;
    U32_T height1;
    int stable_check=0;
    
    if(fpga_xilinx_cxt)
    {
        if(frameinfo)
        {
            stable_check = fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_DETECTOR_STABLE) & 0x0001; 
            *stable = stable_check;
            #if 0
            if ((stable_check ==0)
                 fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_DETECTOR_STABLE, 2);
            #endif     
            frameinfo->width=fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_DETECTOR_WIDTH) & 0xffff;
            height0=fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_DETECTOR_HEIGHT_0) & 0xffff;
            height1=(fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_DETECTOR_HEIGHT_1) >> 16) & 0xffff;
            if(height1==height0)
            {
                frameinfo->height=height0;
                frameinfo->is_interlace=1;
            }
            else
            {
                
                frameinfo->height=height0;          
                frameinfo->is_interlace=0;
            }
            
            if (height0 == 240 || height0 == 288 || height0 == 540) //576i or 480i or 1080i
            {
				printk("%s ====set to interlace\n",__func__);
                frameinfo->is_interlace=1;
			}
			
		#if 1
            if (stable_check ==0)
                 fpga_xilinx_WriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VID_DETECTOR_STABLE, 0x02);
            #endif   
        }
    }
    
}

static BOOL_T fpga_xilinx_set_gpio(handle_t fpga_xilinx_handle,
				 BOOL_T dir, BOOL_T value)
{
	fpga_xilinx_cxt_t *fpga_xilinx_cxt = fpga_xilinx_handle;
	U32_T tmp;
	U32_T read_tmp;

	tmp = fpga_xilinx_ReadReg(fpga_xilinx_cxt, REG_FPGA_BRIDGE_750_GPIO);

	tmp = value ? tmp | (1 << dir) : tmp & ~(1 << dir);

	fpga_xilinx_WriteReg(fpga_xilinx_cxt, REG_FPGA_BRIDGE_750_GPIO, tmp);

	read_tmp = fpga_xilinx_ReadReg(fpga_xilinx_cxt, REG_FPGA_BRIDGE_750_GPIO);

	return TRUE;
}

void fpga_xilinx_hdmi_hotplug(handle_t fpga_xilinx_handle)
{
	fpga_xilinx_cxt_t *fpga_xilinx_cxt = fpga_xilinx_handle;
	BOOL_T tmp;
	
	tmp = fpga_xilinx_set_gpio(fpga_xilinx_cxt, 2, 1);

	if (!tmp) {
		printk("%s Error: Set HDP to low failed\n", __func__);
	}

	sys_msleep(50);

	tmp = fpga_xilinx_set_gpio(fpga_xilinx_cxt, 2, 0);

	if (!tmp) {
		printk("%s Error: Set HDP to hight failed\n", __func__);
	}

	sys_msleep(150);

	tmp = fpga_xilinx_set_gpio(fpga_xilinx_cxt, 2, 1);
        
        sys_msleep(50);

	if (!tmp) {
		printk("%s Error: Set HDP to low failed\n", __func__);
	}

}

handle_t fpga_xilinx_init(cxt_mgr_handle_t cxt_mgr,pci_model_handle_t pci_handle,fpga_xilinx_cfg_t *cfg)
{
    fpga_xilinx_cxt_t   *fpga_xilinx_cxt=NULL;
    handle_t gpio_mgr=NULL;
    handle_t mem_mgr=NULL;
    handle_t task_mgr=NULL;
    int i;
   
    enum
    {
        NO_ERROR,
        ERROR_CXT_MGR,
        ERROR_PCI_HANDLE,
        ERROR_NO_SETUP_INFO,
        ERROR_ALLOC_CXT,
        ERROR_ALLOC_SPINLOCK,
        ERROR_GET_GPIO_MGR,
        ERROR_GET_MEM_MGR,
        ERROR_GET_TASK_MGR,
        ERROR_ALLOC_TASKLET,
        ERROR_NO_DESC,
        ERROR_ALLOC_AUD_BUF,
        
    }err=NO_ERROR;
    
    do
    { 
        
        if(!cxt_mgr)
        {
            err=ERROR_CXT_MGR;
            break;
        }
        if(!pci_handle)
        {
            err=ERROR_PCI_HANDLE;
            break;
        }
        if(!cfg)
        {
            err=ERROR_NO_SETUP_INFO;
            break;
        }
        fpga_xilinx_cxt=cxt_manager_add_cxt(cxt_mgr,FPGA_XILINX_CXT_ID,fpga_xilinx_alloc,fpga_xilinx_release);
        fpga_xilinx_cxt->lock=sys_new_spinlock("fpga_xilinx",SYS_SPINLOCK_LEVEL_IRQ);
        if(!fpga_xilinx_cxt)
        {
            err=ERROR_ALLOC_CXT;
            break;
        }
        fpga_xilinx_cxt->pci_handle=pci_handle;
        cxt_manager_ref_context(pci_handle);
        
        if(!fpga_xilinx_cxt->lock)
        {
            err=ERROR_ALLOC_SPINLOCK;
            break;
        }
        
        fpga_xilinx_reset(fpga_xilinx_cxt);
        //fpga_xilinx_cxt->can_be_terminated=FALSE; //check by irq for fpga reset
        printk("****************************\n");
        printk("chip_version:%08x\n",fpga_xilinx_ReadReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_VERSION));
        printk("****************************\n");
        fpga_xilinx_setgpio(fpga_xilinx_cxt,cfg->gpio_mask,cfg->gpio_value);
        
        
        for(i=0;i<FPGA_XILINX_I2C_BUS_NUM;i++)
        {
             fpga_xilinx_set_i2c_speed(fpga_xilinx_cxt,i,cfg->i2c_bus_speed[i]);
        }
        

        for(i=0;i<FPGA_XILINX_I2C_BUS_NUM;i++)
        {
            unsigned short addr;
            for(addr=0;addr<0x80;addr++)
            {
                if(fpga_xilinx_I2CWriteData(fpga_xilinx_cxt,i,addr<<1,I2C_CMD_START_MASK | I2C_CMD_WRITE_MASK))
                {
                    debug_msg("i2c bus %d found device %02x\n",i,addr);
                }
                fpga_xilinx_I2CWriteData(fpga_xilinx_cxt,i,0,I2C_CMD_STOP_MASK | I2C_CMD_WRITE_MASK);
            }  
        }

        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GPIO,0x08,0x00);
        sys_msleep(100);
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GPIO,0x08,0x08);
        sys_msleep(100);
        fpga_xilinx_MaskWriteReg(fpga_xilinx_cxt,REG_FPGA_BRIDGE_750_GPIO,0x08,0x00);
        sys_msleep(100);
        
        gpio_mgr=cxt_manager_get_context(cxt_mgr,GPIO_CXT_ID,0);
        if(!gpio_mgr)
        {
            err=ERROR_GET_GPIO_MGR;
            break;
        }
        gpio_model_register_ic(gpio_mgr,FPGA_XILINX_GPIODRV_NAME,&fpga_xilinx_gpio_ops,fpga_xilinx_cxt);
                
        mem_mgr=cxt_manager_get_context(cxt_mgr,MEM_CXT_ID,0);
        if(!mem_mgr)
        {
            err=ERROR_GET_MEM_MGR;
            break;
        }
        fpga_xilinx_cxt->mem_mgr=mem_mgr;
        if(cfg->audio_buffer_size)
        {
            fpga_xilinx_cxt->audio_info.audio_buf_size=cfg->audio_buffer_size;
        }
        else
        {
            fpga_xilinx_cxt->audio_info.audio_buf_size=FPGA_XILINX_DEFAULT_AUDIO_BUF_SIZE;
        }
        task_mgr=cxt_manager_get_context(cxt_mgr,TASK_MODEL_CXT_ID,0);
        if(!task_mgr)
        {
            err=ERROR_GET_TASK_MGR;
            break;
        }
        fpga_xilinx_cxt->task_mgr=task_mgr;
        
        cxt_manager_ref_context(task_mgr);
        if(fpga_xilinx_cxt->audio_info.audio_buf_size)
        {
            mem_model_dma_pool_cfg_t audio_dma_pool_cfg =
            {
                .name = "fpga_xilinx_aud",
                .size = fpga_xilinx_cxt->audio_info.audio_buf_size,
                .align = 0,
                .boundary = 0,
            };
            BOOL_T alloc_ok;
        
            for(i=0,alloc_ok=TRUE;i<FPGA_XILINX_AUDIO_BUF_COUNT;i++)
            {
                fpga_xilinx_cxt->audio_info.dma_buffer[i]=mem_model_dma_pool_alloc_atomic(fpga_xilinx_cxt->mem_mgr, &audio_dma_pool_cfg);
                if(fpga_xilinx_cxt->audio_info.dma_buffer[i]==NULL)
                {
                    alloc_ok=FALSE;
                    break;
                }
            }
            if(!alloc_ok)
            {
                err=ERROR_ALLOC_AUD_BUF;
                break;
            }
            fpga_xilinx_cxt->audio_DPC_task=task_model_create_task(task_mgr,fpga_xilinx_audio_DPC_func,fpga_xilinx_cxt,"audio_DPC");
        }
        
        fpga_xilinx_cxt->last_tx_ch=-1;
        init_queue(&fpga_xilinx_cxt->free_desc_queue);
        init_queue(&fpga_xilinx_cxt->ready_desc_queue);
        init_queue(&fpga_xilinx_cxt->done_desc_queue);
        
        fpga_xilinx_cxt->video_DPC_task=task_model_create_task(task_mgr,fpga_xilinx_video_DPC_func,fpga_xilinx_cxt,"video_DPC");
        fpga_xilinx_cxt->video_term_DPC_task=task_model_create_task(task_mgr,fpga_xilinx_video_term_DPC_func,fpga_xilinx_cxt,"video_term_DPC");
        sys_atomic_set(fpga_xilinx_cxt->atomic_VDPC_scheduling,0);
        if(!fpga_xilinx_cxt->video_DPC_task)
        {
            err=ERROR_ALLOC_TASKLET;
            break;
        }
        fpga_xilinx_cxt->atomic_VDPC_scheduling=sys_new_atomic();
        
        
        fpga_xilinx_csc_in_out(fpga_xilinx_cxt,FPGA_XILINX_CS_YUV422,FPGA_XILINX_CS_YUV422,TRUE, 0/*uyvy*/);
        
        pci_model_register_isr(pci_handle,fpga_xilinx_irq_func,fpga_xilinx_cxt);
        
        fpga_xilinx_enable_int(fpga_xilinx_cxt,TRUE);
        
    }while(0);
    if(err!=NO_ERROR)
    {
        printk("%s err %d\n",__func__,err);
        switch(err)
        {
			case ERROR_GET_TASK_MGR:    
			case ERROR_GET_MEM_MGR:
			case ERROR_GET_GPIO_MGR:
				sys_release_spinlock(fpga_xilinx_cxt->lock);
			case ERROR_ALLOC_SPINLOCK:
				cxt_manager_unref_context(fpga_xilinx_cxt);
				fpga_xilinx_cxt = NULL;
			case ERROR_ALLOC_CXT:
			case ERROR_NO_SETUP_INFO:
			case ERROR_PCI_HANDLE:
			case ERROR_CXT_MGR:
            break;
			default:
            break;
        }
    }
    else
    {
		printk("%s ok\n",__func__);
	}
    
    return fpga_xilinx_cxt;
}
