 /*
 *Driver for the card pcie bridge.

 *Copyright (C) 2019 AVerMedia TECHNOLOGIES, Inc.

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#include <stdarg.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <media/v4l2-device.h>
#include "cxt_mgr.h"
#include "framegrabber.h"
#include "framegrabber_priv.h"
#include <media/videobuf2-vmalloc.h>

#include "v4l2_model.h"
#include "v4l2_model_device.h"

static void *framegrabber_alloc(void);
static void framegrabber_release(void *);


static const framegrabber_pixfmt_t support_pixfmts[] = {
	
	[FRAMEGRABBER_PIXFMT_YUYV]={ //YUYV index=0
		.name     = "4:2:2, packed, YUYV",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.depth    = 16,
		.is_yuv   = true,
		.pixfmt_out = YUYV,
	},
	[FRAMEGRABBER_PIXFMT_UYVY]={ //UYVY
		.name     = "4:2:2, packed, UYVY",
		.fourcc   = V4L2_PIX_FMT_UYVY,
		.depth    = 16,
		.is_yuv   = true,
		.pixfmt_out = UYVY,
	},
	[FRAMEGRABBER_PIXFMT_RGB24]={ //RGB3 index=8
		.name     = "RGB24 (LE)",
		.fourcc   = V4L2_PIX_FMT_RGB24, /* rgb */
		.depth    = 24,
		.is_yuv   = false,
		.pixfmt_out = RGB3,
	},
};

static const framegrabber_frame_size_t framegrabber_support_frame_size[FRAMEGRABBER_SUPPORT_FRAMESIZE_NUM] = {
	[FRAMESIZE_640x480]=	{ .width = 640, 	.height = 480 },
	[FRAMESIZE_720x480]= 	{ .width = 720, 	.height = 480 },
	[FRAMESIZE_720x576]= 	{ .width = 720, 	.height = 576 },
	//[FRAMESIZE_800x600]= 	{ .width = 800, 	.height = 600 },
	[FRAMESIZE_1024x768]= 	{ .width = 1024, 	.height = 768 },
	[FRAMESIZE_1280x720]=	{ .width = 1280, 	.height = 720 },
	//[FRAMESIZE_1280x800]=	{ .width = 1280, 	.height = 800 },
	[FRAMESIZE_1280x1024]=	{ .width = 1280, 	.height = 1024},
	[FRAMESIZE_1360x768]=	{ .width = 1360, 	.height = 768 },
	[FRAMESIZE_1440x900]=	{ .width = 1440, 	.height = 900 },
	[FRAMESIZE_1680x1050]=	{ .width = 1680, 	.height = 1050},
	[FRAMESIZE_1920x1080]=	{ .width = 1920, 	.height = 1080},
	[FRAMESIZE_3840x2160]=	{ .width = 3840, 	.height = 2160},
	[FRAMESIZE_4096x2160]=	{ .width = 4096, 	.height = 2160},
};


static const int framegrabber_support_refreshrate[]= {
	[REFRESHRATE_15]=15,
	[REFRESHRATE_24]=24,
	[REFRESHRATE_25]=25,
	[REFRESHRATE_30]=30,
	[REFRESHRATE_40]=40,
	[REFRESHRATE_50]=50,
	[REFRESHRATE_56]=56,
	[REFRESHRATE_60]=60,
	[REFRESHRATE_70]=70,
	[REFRESHRATE_72]=72,
	[REFRESHRATE_75]=75,
	[REFRESHRATE_85]=85,
};

typedef struct 
{
	U16_T width;
    U16_T   height;
    U8_T interlaced;
    U8_T   framerate;
    U8_T   framerate_decimal;
}edid_timing_info_t;

typedef struct
{
	int edid_timing_index;
	edid_timing_info_t edid_timing;
}edid_timing_list_t;

edid_timing_list_t standard_edid_timing[] = 
{
	{ 1, {640 , 480 , 0, 59, 94  }},
	{ 1, {640 , 480 , 0, 60, 0	}},
	{ 2, {720 , 480 , 0, 59, 94  }},
	{ 2, {720 , 480 , 0, 60, 0	}},
	{ 3, {720 , 480 , 0, 59, 94  }},
	{ 3, {720 , 480 , 0, 60, 0	}},
	{ 4, {1280, 720 , 0, 59, 94  }},
	{ 4, {1280, 720 , 0, 60, 0	}},
	{ 5, {1920, 1080, 1, 60, 0	}},
	{ 5, {1920, 1080, 1, 59, 94  }},
	{ 6, {1440, 480 , 1, 59, 94  }},
	{ 6, {1440, 480 , 1, 60, 0	}},
	{ 7, {1440, 480 , 1, 59, 94  }},
	{ 7, {1440, 480 , 1, 60, 0	}},
	{ 8, {1440, 240 , 0, 60, 5 }},
	{ 8, {1440, 240 , 0, 60, 0 }},
	{ 8, {1440, 240 , 0, 59, 94 }},
	{ 8, {1440, 240 , 0, 59, 83 }},
	{ 9, {1440, 240 , 0, 60, 5 }},
	{ 9, {1440, 240 , 0, 59, 83 }},
	{ 9, {1440, 240 , 0, 59, 94 }},
	{ 9, {1440, 240 , 0, 60, 0 }},
	{ 10, {2880, 480 , 1, 59, 94  }},
	{ 10, {2880, 480 , 1, 60, 0  }},
	{ 11, {2880, 480 , 1, 59, 94  }},
	{ 11, {2880, 480 , 1, 60, 0  }},
	{ 12, {2880, 240 , 0, 60, 5 }},
	{ 12, {2880, 240 , 0, 60, 0 }},
	{ 12, {2880, 240 , 0, 59, 94 }},
	{ 12, {2880, 240 , 0, 59, 83 }},
	{ 13, {2880, 240 , 0, 60, 5 }},
	{ 13, {2880, 240 , 0, 60, 0 }},
	{ 13, {2880, 240 , 0, 59, 94 }},
	{ 13, {2880, 240 , 0, 59, 83 }},
	{ 14, {1440, 480 , 0, 59, 94  }},
	{ 14, {1440, 480 , 0, 60, 0  }},
	{ 15, {1440, 480 , 0, 59, 94  }},
	{ 15, {1440, 480 , 0, 60, 0  }},
	{ 16, {1920, 1080, 0, 60, 0  }},
	{ 16, {1920,  1080, 0, 59, 94 }},
	{ 17, {720 ,  576 , 0, 50, 0 }},
	{ 18, {720 ,  576 , 0, 50, 0 }},
	{ 19, {1280,  720 , 0, 50, 0 }},
	{ 20, {1920,  1080, 1, 50, 0 }},
	{ 21, {1440,  576 , 1, 50, 0 }},
	{ 22, {1440,  576 , 1, 50, 0 }},
	{ 23, {1440,  288 , 0, 50, 80 }},
	{ 24, {1440,  288 , 0, 50, 80 }},
	{ 23, {1440,  288 , 0, 49, 92 }},
	{ 24, {1440,  288 , 0, 49, 92 }},
	{ 23, {1440,  288 , 0, 49, 76 }},
	{ 24, {1440,  288 , 0, 49, 76 }},
	{ 25, {2880,  576 , 1, 50, 0 }},
	{ 26, {2880,  576 , 1, 50, 0 }},
	{ 27, {2880,  288 , 0, 50, 80 }},
	{ 28, {2880,  288 , 0, 50, 80 }},
	{ 27, {2880,  288 , 0, 49, 92 }},
	{ 28, {2880,  288 , 0, 49, 92 }},
	{ 27, {2880,  288 , 0, 49, 76 }},
	{ 28, {2880,  288 , 0, 49, 76 }},
	{ 27, {2880,  288 , 0, 50, 0 }},
	{ 28, {2880,  288 , 0, 50, 0 }},
	{ 29, {1440,  576 , 0, 50, 0 }},
	{ 30, {1440,  576 , 0, 50, 0 }},
	{ 31, {1920,  1080, 0, 50, 0 }},
	{ 32, {1920 ,1080, 0, 24, 0  }},
	{ 32, {1920 ,1080, 0, 23, 98  }},
	{ 33, {1920 ,1080, 0, 25, 0  }},
	{ 34, {1920 ,1080, 0, 29, 97  }},
	{ 34, {1920 ,1080, 0, 30, 0  }},
	{ 35, {2880, 480 , 0, 59, 94  }},
	{ 35, {2880, 480 , 0, 60, 0  }},
	{ 36, {2880, 480 , 0, 59, 94  }},
	{ 36, {2880, 480 , 0, 60, 0  }},
	{ 37, {2880,  576 , 0, 50, 0 }},
	{ 38, {2880,  576 , 0, 50, 0 }},
	{ 39, {1920,  1080, 1, 50, 0 }},
	{ 40, {1920, 1080, 1, 100, 0 }},
	{ 41, {1280, 720 , 0, 100, 0 }},
	{ 42, {720 , 576 , 0, 100, 0 }},
	{ 43, {720 , 576 , 0, 100, 0 }},
	{ 44, {1440, 576 , 1, 100, 0 }},
	{ 45, {1440, 576 , 1, 100, 0 }},
	{ 46, {1920, 1080, 1, 120, 0 }},
	{ 46, {1920, 1080, 1, 119, 88 }},
	{ 47, {1280, 720 , 0, 120, 0 }},
	{ 47, {1280, 720 , 0, 119, 88 }},
	{ 48, {720 , 480 , 0, 119, 88 }},
	{ 48, {720 , 480 , 0, 120, 0 }},
	{ 49, {720 , 480 , 0, 119, 88 }},
	{ 49, {720 , 480 , 0, 120, 0 }},
	{ 50, {1440, 480 , 1, 119, 88 }},
	{ 50, {1440, 480 , 1, 120, 0 }},
	{ 51, {1440, 480 , 1, 119, 88 }},
	{ 51, {1440, 480 , 1, 120, 0 }},
	{ 52, {720 , 576 , 0, 200, 0 }},
	{ 53, {720 , 576 , 0, 200, 0 }},
	{ 54, {1440, 576 , 1, 200, 0 }},
	{ 55, {1440, 576 , 1, 200, 0 }},
	{ 56, {720 , 480 , 0, 239, 76 }},
	{ 56, {720 , 480 , 0, 240, 0 }},
	{ 57, {720 , 480 , 0, 239, 76 }},
	{ 57, {720 , 480 , 0, 240, 0 }},
	{ 58, {1440, 480 , 1, 239, 76 }},
	{ 58, {1440, 480 , 1, 240, 0 }},
	{ 59, {1440, 480 , 1, 239, 76 }},
	{ 59, {1440, 480 , 1, 240, 0 }},
	{ 60, {1280 ,720 , 0, 24, 0  }},
	{ 60, {1280 ,720 , 0, 23, 98  }},
	{ 61, {1280 ,720 , 0, 25, 0  }},
	{ 62, {1280 ,720 , 0, 30, 0  }},
	{ 62, {1280 ,720 , 0, 29, 97  }},
	{ 63, {1920, 1080, 0, 120, 0 }},
	{ 63, {1920, 1080, 0, 119, 88 }},
	{ 64, {1920, 1080, 0, 100, 0 }},
	{ 65, {1280 ,720 , 0, 24, 0  }},
	{ 65, {1280 ,720 , 0, 23, 98  }},
	{ 66, {1280 ,720 , 0, 25, 0  }},
	{ 67, {1280 ,720 , 0, 29, 97  }},
	{ 67, {1280 ,720 , 0, 30, 0  }},
	{ 68, {1280,  720 , 0, 50, 0 }},	
	{ 69, {1280, 720 , 0, 60, 0  }},
	{ 69, {1280, 720 , 0, 59, 94  }},		
	{ 70, {1280, 720 , 0, 100, 0 }},	
	{ 71, {1280, 720 , 0, 120, 0 }},
	{ 71, {1280, 720 , 0, 119, 88 }},		
	{ 72, {1920 ,1080, 0, 24, 0  }},
	{ 72, {1920 ,1080, 0, 23, 98  }},
	{ 73, {1920 ,1080, 0, 25, 0  }},
	{ 74, {1920 ,1080, 0, 30, 0  }},
	{ 74, {1920 ,1080, 0, 29, 97  }},
	{ 75, {1920 ,1080, 0, 50, 0  }},
	{ 76, {1920, 1080, 0, 60, 0  }},
	{ 76, {1920, 1080, 0, 59, 94  }},
	{ 77, {1920, 1080, 0, 100, 0 }},		
	{ 78, {1920, 1080, 0, 120, 0 }},
	{ 78, {1920, 1080, 0, 119, 88 }},	
	{ 79, {1680 ,720 , 0, 24, 0  }},
	{ 79, {1680 ,720 , 0, 23, 98  }},
	{ 80, {1680 ,720 , 0, 25, 0  }},
	{ 81, {1680 ,720 , 0, 30, 0  }},
	{ 81, {1680 ,720 , 0, 29, 97  }},
	{ 82, {1680,  720 , 0, 50, 0 }},
	{ 83, {1680, 720 , 0, 60, 0  }},
	{ 83, {1680, 720 , 0, 59, 94  }},
	{ 84, {1680, 720 , 0, 100, 0 }},
	{ 85, {1680, 720 , 0, 120, 0 }},
	{ 85, {1680, 720 , 0, 119, 88 }},
	{ 86, {2560 ,1080, 0, 24, 0  }},
	{ 86, {2560 ,1080, 0, 23, 98  }},
	{ 87, {2560 ,1080, 0, 25, 0  }},
	{ 88, {2560 ,1080, 0, 30, 0  }},
	{ 88, {2560 ,1080, 0, 29, 97  }},
	{ 89, {2560,  1080, 0, 50, 0 }},
	{ 90, {2560, 1080, 0, 60, 0  }},
	{ 90, {2560, 1080, 0, 59, 94  }},
	{ 91, {2560, 1080, 0, 100, 0 }},
	{ 92, {2560, 1080, 0, 120, 0 }},
	{ 92, {2560, 1080, 0, 119, 88 }},
	{ 93, {3840, 2160, 0, 23, 98  }},
	{ 93, {3840, 2160, 0, 24, 0  }},
	{ 94, {3840, 2160, 0, 25, 0  }},
	{ 95, {3840, 2160, 0, 29, 97  }},
	{ 95, {3840, 2160, 0, 30, 0  }},
	{ 96, {3840,  2160, 0, 50, 0 }},
	{ 97, {3840, 2160, 0, 60, 0  }},
	{ 97, {3840, 2160, 0, 59, 94  }},
	{ 98, {4096 ,2160, 0, 24, 0  }},
	{ 98, {4096 ,2160, 0, 23, 98  }},
	{ 99, {4096 ,2160, 0, 25, 0  }},
	{ 100, {4096 ,2160, 0, 30, 0  }},
	{ 100, {4096 ,2160, 0, 29, 97  }},
	{ 101, {4096,  2160, 0, 50, 0 }},
	{ 102, {4096, 2160, 0, 60, 0  }},
	{ 102, {4096, 2160, 0, 59, 94  }},
	{ 103, {3840, 2160, 0, 24, 0  }},
	{ 103, {3840, 2160, 0, 23, 98  }},
	{ 104, {3840, 2160, 0, 25, 0  }},
	{ 105, {3840, 2160, 0, 30, 0  }},
	{ 105, {3840, 2160, 0, 29, 97  }},
	{ 106, {3840,  2160, 0, 50, 0 }},
	{ 107, {3840, 2160, 0, 60, 0  }},
	{ 107, {3840, 2160, 0, 59, 94  }}
};


framegrabber_handle_t framegrabber_init(cxt_mgr_handle_t cxt_mgr,framegrabber_property_t *prop,framegrabber_interface_t *operation)
{
	framegrabber_context_t *fg_cxt=NULL;
	framegrabber_error_e err=FRAMEGRABBER_OK;

	do
	{
		int i,j;
		BOOL_T end_of_inputinfo;

		if(!cxt_mgr)
			break;
		fg_cxt=cxt_manager_add_cxt(cxt_mgr,FG_CXT_ID,framegrabber_alloc,framegrabber_release);
		if(!fg_cxt)
		{
			err=FRAMEGRABBER_ERROR_ALLOC_FAIL;
			break;
		}
		if(prop->input_setup_info==NULL)
		{
			err=FRAMEGRABBER_ERROR_INPUT_NULL;
			break;
		}

		fg_cxt->current_audiorate = AUDIO_RATE_96000;	// for unplugged start

		for(i=0,end_of_inputinfo=false;i<FRAMEGRABBER_MAX_INPUT_NUM;i++)
		{
			if(prop->input_setup_info[i].name)
			{
				fg_cxt->input_info[i].name=prop->input_setup_info[i].name;
				fg_cxt->input_info[i].framesize_cap=0;
				for(j=0;j<FRAMEGRABBER_SUPPORT_FRAMESIZE_NUM;j++)
				{
					if(prop->input_setup_info[i].support_framesize_info[j]!=0)
					{
						fg_cxt->input_info[i].framesize_cap |= (1 << j);
						fg_cxt->input_info[i].framesize_info[j]=prop->input_setup_info[i].support_framesize_info[j];
					}
				}
			}else
			{
				fg_cxt->input_no=i;
				end_of_inputinfo=true;
				break;
			}
		}
		if(!end_of_inputinfo)
		{
			err=FRAMEGRABBER_ERROR_INPUT_OVERFLOW;
			break;
		}
		fg_cxt->current_input=0;
		fg_cxt->current_input_status=FRAMEGRABBER_INPUT_STATUS_NO_POWER;
		if(prop->support_out_pixfmt_mask==0)
		{
			err=FRAMEGRABBER_ERROR_NO_PIXFMT_SETUP;
			break;
		}
		fg_cxt->pixfmt_cap=prop->support_out_pixfmt_mask;
		fg_cxt->max_supported_width=prop->max_supported_width;
                fg_cxt->max_frame_size=prop->max_frame_size;
		for(i=0;i<FRAMEGRABBER_PIXFMT_MAX;i++)
		{
			if(fg_cxt->pixfmt_cap & (1<<i) )
			{
				//printk("%s fg_cxt->current_out_pixfmt=%d\n",__func__,i);
				fg_cxt->current_out_pixfmt=i;
				break;
			}
		}
		memcpy(&fg_cxt->interface,operation,sizeof(*operation));
	}while(0);
	if(err!=FRAMEGRABBER_OK)
	{
		switch(err)
		{
			case FRAMEGRABBER_ERROR_INPUT_NULL:
			case FRAMEGRABBER_ERROR_ALLOC_FAIL:
				break;
			default:
				break;
		}
	}

	return (fg_cxt) ? &fg_cxt->interface : NULL;

}

static framegrabber_context_t *framegrabber_getcontext(framegrabber_handle_t handle)
{
	return container_of(handle,framegrabber_context_t,interface);
}

static void *framegrabber_alloc()
{
	framegrabber_context_t *context=NULL;

	context = kzalloc(sizeof(*context), GFP_KERNEL);

	return context;
}

static void framegrabber_release(void *context)
{
	framegrabber_context_t *fg_cxt=context;
	if(fg_cxt)
		kfree(fg_cxt);
}

void framegrabber_set_data(framegrabber_handle_t handle,void *data)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->data=data;
}
void *framegrabber_get_data(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->data;
}


void framegrabber_start(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	if(context->interface.s_input)
	{
		context->interface.s_input(handle,context->current_input);
	}
}

int framegrabber_g_input_num(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->input_no;
}

const char *framegrabber_g_input_name(framegrabber_handle_t handle,int input)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	if(input<0 || input > context->input_no)
	{
		return NULL;
	}
	return context->input_info[input].name;
}

int framegrabber_g_input(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->current_input;
}

int framegrabber_s_input(framegrabber_handle_t handle,int input)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	if(input <0 && input >= context->input_no)
	{
		return FRAMEGRABBER_ERROR_INPUT_INVALID;
	}
	if(context->current_input!=input)
	{
		if(context->interface.s_input)
			context->interface.s_input(handle,input);
		context->current_input=input;
	}

	return FRAMEGRABBER_OK;
}

framegrabber_input_status_bitmask_e framegrabber_g_input_status(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->current_input_status;
}

void framegrabber_s_input_status(framegrabber_handle_t handle,framegrabber_input_status_bitmask_e input_status)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_input_status=input_status;

}

void framegrabber_mask_s_status(framegrabber_handle_t handle,framegrabber_status_bitmask_e mask,framegrabber_status_bitmask_e status)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	framegrabber_status_bitmask_e current_status=context->current_status;
	framegrabber_status_bitmask_e diff;

	current_status &= ~mask;
	current_status |= (status & mask);
	diff=current_status ^ context->current_status;
	//printk("%s diff %08x status %08x=>%08x\n",__func__,diff,context->current_status,current_status);
	context->current_status=current_status;
    
	if(diff & (FRAMEGRABBER_STATUS_V4L_START_STREAMING_BIT|FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT) )
	{
		if((current_status & (FRAMEGRABBER_STATUS_V4L_START_STREAMING_BIT|FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT))==(FRAMEGRABBER_STATUS_V4L_START_STREAMING_BIT|FRAMEGRABBER_STATUS_SIGNAL_LOCKED_BIT))
		{
			if(handle->stream_on)
			{
				printk("%s stream_on\n",__func__);
				handle->stream_on(handle);
			}
		}else
		{
			if(handle->stream_off)
			{
				printk("%s stream_off\n",__func__);
				handle->stream_off(handle);
			}
		}
	}
	return;
}

framegrabber_status_bitmask_e framegrabber_g_status(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	return context->current_status;
}


const framegrabber_pixfmt_t *framegrabber_g_out_pixelfmt(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return &support_pixfmts[context->current_out_pixfmt];
}

const framegrabber_pixfmt_t *framegrabber_g_support_pixelfmt_by_index(framegrabber_handle_t handle,int index) 
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	int i,idx=-1;
	u32 mask=context->pixfmt_cap;

    
    if (index >= FRAMEGRABBER_PIXFMT_MAX)
	    return NULL;
	    
	for(i=0;i<FRAMEGRABBER_PIXFMT_MAX && mask;i++)
	{
		if( mask & (1<<i))
		{
			idx++;
			mask ^= (1<<i);
			if(idx==index)
			{
				return &support_pixfmts[i];
			}
		}
	}

	return NULL;
}

const framegrabber_pixfmt_t *framegrabber_g_support_pixelfmt_by_fourcc(framegrabber_handle_t handle,u32 fourcc)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	int i;
	int pixfmt_index=-1;

	for(i=0;i<FRAMEGRABBER_PIXFMT_MAX;i++)
	{
		if(support_pixfmts[i].fourcc==fourcc)
		{
			if(context->pixfmt_cap & (1<<i))
				pixfmt_index=i;
			break;
		}
	}
	if(pixfmt_index==-1)
			return NULL;

	return &support_pixfmts[pixfmt_index];
}


int framegrabber_s_out_pixelfmt(framegrabber_handle_t handle,u32 fourcc)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	int i;
	int pixfmt_index=-1;

	for(i=0;i<FRAMEGRABBER_PIXFMT_MAX;i++)
	{
		if(support_pixfmts[i].fourcc==fourcc)
		{
			if(context->pixfmt_cap & (1<<i))
				pixfmt_index=i;
			break;
		}
	}
	if(pixfmt_index==-1)
		return FRAMEGRABBER_ERROR_NO_SUPPORT_PIXFMT;
	context->current_out_pixfmt=pixfmt_index;
	if(context->interface.s_out_pixfmt)
		context->interface.s_out_pixfmt(handle,fourcc);

	return FRAMEGRABBER_OK;
}

void framegrabber_s_out_framesize(framegrabber_handle_t handle,int width,int height)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	if((context->current_out_framesize.width==width) && (context->current_out_framesize.height==height))
		return;
	context->current_out_framesize.width=width;
	context->current_out_framesize.height=height;
	if(context->interface.s_framesize)
		context->interface.s_framesize(handle,width,height);
}

unsigned framegrabber_g_max_framebuffersize(framegrabber_handle_t handle)
{
    framegrabber_context_t *context=framegrabber_getcontext(handle);
    const framegrabber_pixfmt_t *pixfmt=framegrabber_g_out_pixelfmt(handle);
    unsigned max_framebufsize=0;
        
    max_framebufsize=(context->max_frame_size * pixfmt->depth) >> 3;
                
    return max_framebufsize;
}

unsigned framegrabber_g_out_bytesperline(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	const framegrabber_pixfmt_t *pixfmt=framegrabber_g_out_pixelfmt(handle);
	unsigned bytesperline;

    if(context->current_out_framesize.width)
		bytesperline=(context->current_out_framesize.width * pixfmt->depth) >> 3;
    else
        bytesperline=(context->current_input_framesize.width * pixfmt->depth) >> 3;
        
	return bytesperline;
}

void framegrabber_g_out_framesize(framegrabber_handle_t handle,int *width,int *height)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	*width=context->current_out_framesize.width;
	*height=context->current_out_framesize.height;
}

void framegrabber_g_input_framesize(framegrabber_handle_t handle,int *width,int *height)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	*width=context->current_input_framesize.width;
	*height=context->current_input_framesize.height;
}

framegrabber_framemode_e framegrabber_g_input_framemode(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->framemode;
}

void framegrabber_s_input_framemode(framegrabber_handle_t handle,framegrabber_framemode_e framemode)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	if((framemode <0) || (framemode > FRAMEGRABBER_FRAMEMODE_COUNT))
		return;
	context->framemode=framemode;
}

void framegrabber_s_input_framesize(framegrabber_handle_t handle,int width,int height)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_input_framesize.width=width;
	context->current_input_framesize.height=height;
}

void framegrabber_s_fpga_bypassmode(framegrabber_handle_t handle,int bypass_disable)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_fpga_bypassmode=bypass_disable;
}

int framegrabber_g_fpga_bypassmode(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->current_fpga_bypassmode;
}

void framegrabber_s_output_pixelformat(framegrabber_handle_t handle,int output_pixelformat)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_output_pixelformat=output_pixelformat;
}

int framegrabber_g_output_pixelformat(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->current_output_pixelformat;
}

void framegrabber_s_input_dualmode(framegrabber_handle_t handle,int dual_pixel)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_dual_pixel=dual_pixel;
}

int framegrabber_g_input_dualmode(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
    //printk(">>>>>%s ..%d\n",__func__,context->current_dual_pixel);
	return context->current_dual_pixel;
}

int framegrabber_g_input_framerate(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
    //printk(">>>>>%s ..%d\n",__func__,context->current_framerate);
	return context->current_framerate;
}

int framegrabber_g_input_denominator(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	
	return context->current_denominator;
}

void framegrabber_s_input_framerate(framegrabber_handle_t handle,int framerate,int denominator)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_framerate=framerate;
    context->current_denominator=denominator;
    //printk(">>>>>%s ..%d\n",__func__,context->current_framerate);
}

void framegrabber_s_out_framerate(framegrabber_handle_t handle,int framerate)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_out_framerate=framerate;
    
    printk(">>>>>>%s ..%d\n",__func__,context->current_out_framerate);
}

int framegrabber_g_out_framerate(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	
    printk(">>>>>>>%s ..%d\n",__func__,context->current_out_framerate);
    
	return context->current_out_framerate;
}

void framegrabber_s_input_audiorate(framegrabber_handle_t handle,enum framegrabber_audio_sample audiorate)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_audiorate=audiorate;
  
    //printk(">>>>>%s ..%d\n",__func__,context->current_audiorate);
}

int framegrabber_g_input_audiorate(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	
    //printk(">>>>>>>%s ..%d\n",__func__,context->current_audiorate);
    
	return context->current_audiorate;
}

void framegrabber_s_input_aviinfoframe(framegrabber_handle_t handle,U8_T *fg_aviinfo)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

    memcpy(context->current_aviinfoframe,fg_aviinfo,sizeof(context->current_aviinfoframe)); 
    			    
#if 1        
    {
        int jj;
	    //check avi infoframe data
	    printk(" =======framegrabber_s_input_aviinfo======\n");
	    for (jj=0; jj< FRAMEGRABBER_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			   
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            context->current_aviinfoframe[jj-7],context->current_aviinfoframe[jj-6],context->current_aviinfoframe[jj-5],context->current_aviinfoframe[jj-4],
	            context->current_aviinfoframe[jj-3],context->current_aviinfoframe[jj-2],context->current_aviinfoframe[jj-1],context->current_aviinfoframe[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif
}

int framegrabber_g_input_aviinfoframe(framegrabber_handle_t handle,U8_T *fg_aviinfo)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	
	memcpy(fg_aviinfo,context->current_aviinfoframe,sizeof(context->current_aviinfoframe)); 
	
    #if 0        
    {
        int jj;
	    //check avi infoframe data
	    printk(" =====================framegrabber_g_input_aviinfoframe======================\n");
	    for (jj=0; jj< FRAMEGRABBER_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            fg_aviinfo[jj-7],fg_aviinfo[jj-6],fg_aviinfo[jj-5],fg_aviinfo[jj-4],
	            fg_aviinfo[jj-3],fg_aviinfo[jj-2],fg_aviinfo[jj-1],fg_aviinfo[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif
    return 0;
}

void framegrabber_s_input_vendorinfoframe(framegrabber_handle_t handle,U8_T *fg_vendorinfo)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

    memcpy(context->current_vendorinfoframe,fg_vendorinfo,sizeof(context->current_vendorinfoframe)); 
    			    
#if 0        
    {
        int jj;
	    //check vendor infoframe data
	    printk(" =======framegrabber_s_input_vendorinfoframe======\n");
	    for (jj=0; jj< FRAMEGRABBER_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			   
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            context->current_vendorinfoframe[jj-7],context->current_vendorinfoframe[jj-6],context->current_vendorinfoframe[jj-5],context->current_vendorinfoframe[jj-4],
	            context->current_vendorinfoframe[jj-3],context->current_vendorinfoframe[jj-2],context->current_vendorinfoframe[jj-1],context->current_vendorinfoframe[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif
}

int framegrabber_g_input_vendorinfoframe(framegrabber_handle_t handle,U8_T *fg_vendorinfo)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	
	memcpy(fg_vendorinfo,context->current_vendorinfoframe,sizeof(context->current_vendorinfoframe)); 
	
    #if 0        
    {
        int jj;
	    //check vendor infoframe data
	    printk(" =====================framegrabber_g_input_vendorinfoframe======================\n");
	    for (jj=0; jj< FRAMEGRABBER_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            fg_vendorinfo[jj-7],fg_vendorinfo[jj-6],fg_vendorinfo[jj-5],fg_vendorinfo[jj-4],
	            fg_vendorinfo[jj-3],fg_vendorinfo[jj-2],fg_vendorinfo[jj-1],fg_vendorinfo[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif
    return 0;
}

void framegrabber_s_input_audioinfoframe(framegrabber_handle_t handle,U8_T *fg_audioinfo)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

    memcpy(context->current_audioinfoframe,fg_audioinfo,sizeof(context->current_audioinfoframe)); 
    			    
#if 0        
    {
        int jj;
	    //check audio infoframe data
	    printk(" =======framegrabber_s_input_audioinfoframe======\n");
	    for (jj=0; jj< FRAMEGRABBER_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			   
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            context->current_audioinfoframe[jj-7],context->current_audioinfoframe[jj-6],context->current_audioinfoframe[jj-5],context->current_audioinfoframe[jj-4],
	            context->current_audioinfoframe[jj-3],context->current_audioinfoframe[jj-2],context->current_audioinfoframe[jj-1],context->current_audioinfoframe[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif
}

int framegrabber_g_input_audioinfoframe(framegrabber_handle_t handle,U8_T *fg_audioinfo)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	
	memcpy(fg_audioinfo,context->current_audioinfoframe,sizeof(context->current_audioinfoframe)); 
	
    #if 0        
    {
        int jj;
	    //check audio infoframe data
	    printk(" =====================framegrabber_g_input_audioinfoframe======================\n");
	    for (jj=0; jj< FRAMEGRABBER_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            fg_audioinfo[jj-7],fg_audioinfo[jj-6],fg_audioinfo[jj-5],fg_audioinfo[jj-4],
	            fg_audioinfo[jj-3],fg_audioinfo[jj-2],fg_audioinfo[jj-1],fg_audioinfo[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif
    return 0;
}

int framegrabber_g_input_bchs(framegrabber_handle_t handle,int *bchs_select)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	int *bchs_value=0;
    
    context->current_bchs_selection=*bchs_select;
    
	//printk("%s current_bchs_value(%d) = %x\n",__func__,context->current_bchs_selection,context->current_bchs_value);
	
	switch (*bchs_select)
	{
		case 0: //brightness
		{
			handle->brightness_get(handle,bchs_value); 
			bchs_value = &handle->fg_bchs_value;
			printk("%s brightness=%d\n",__func__,*bchs_value);  
		    break;
		}
		case 1: //contrast
		{
			handle->contrast_get(handle,bchs_value); 
			bchs_value = &handle->fg_bchs_value;
			printk("%s contrast=%d\n",__func__,*bchs_value);   
		    break;
		}
		case 2: //hue
		{
			handle->hue_get(handle,bchs_value); 
			bchs_value = &handle->fg_bchs_value;
			printk("%s hue=%d\n",__func__,*bchs_value);  
		    break;
		}
		case 3: //saturation
		{
			handle->saturation_get(handle,bchs_value); 
			bchs_value = &handle->fg_bchs_value;
			printk("%s saturation=%d\n",__func__,*bchs_value);  
		    break;
		}
		
	}
	
	context->current_bchs_value = *bchs_value;
	//printk("%s bchs_value(%d) = %d\n",__func__,context->current_bchs_selection,context->current_bchs_value);
	return context->current_bchs_value;
	
}

void framegrabber_s_input_bchs(framegrabber_handle_t handle,int bchs_value,int bchs_select)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	
    handle->fg_bchs_value = bchs_value;
    handle->fg_bchs_selection = bchs_select;
    context->current_bchs_selection=bchs_select;
    context->current_bchs_value=bchs_value;
    
    if(handle->bchs_set)
		handle->bchs_set(handle);   
	
	printk("%s current_bchs_value(%d) = %d\n",__func__,bchs_select,context->current_bchs_value);
}

int framegrabber_g_input_interlace(framegrabber_handle_t handle)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	return context->current_input_interlace;
}

void framegrabber_s_input_interlace(framegrabber_handle_t handle,BOOL_T interlace)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);

	context->current_input_interlace=interlace;
}

int framegrabber_g_supportframesize(framegrabber_handle_t handle,int index,int *width,int *height)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	int i,idx=-1;
	int in_width=0,in_height=0;
	framegrabber_g_input_framesize(handle,&in_width,&in_height);
	//printk("%s index %d %dx%d\n",__func__,index ,*width,*height);
	{
		u32 framesize_mask=context->input_info[context->current_input].framesize_cap;
		for(i=0;i<FRAMEGRABBER_SUPPORT_FRAMESIZE_NUM &&  framesize_mask;i++)
		{
			if(framesize_mask & (1<<i))
			{
				framesize_mask ^= (1<<i);
				idx++;
				if(idx==index)
				{
					*width=framegrabber_support_frame_size[i].width;
					*height=framegrabber_support_frame_size[i].height;
					return FRAMEGRABBER_OK;
				}
			}

		}
	}
	
	return FRAMEGRABBER_ERROR_INVALID_FRAMESIZE;
}

int framegrabber_g_framesize_supportrefreshrate(framegrabber_handle_t handle,int width,int height,int index)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	int i,idx=-1;
	u32 refreshrate_mask=0;

        
	for(i=0;i<FRAMEGRABBER_SUPPORT_FRAMESIZE_NUM ;i++)
	{
        
		if(framegrabber_support_frame_size[i].width==width && framegrabber_support_frame_size[i].height==height)
		{
        
                    if(context->input_info[context->current_input].framesize_cap & (1<<i) )
                    {
        
                        if(context->input_info[context->current_input].framesize_info[i])
                        {
                            refreshrate_mask=context->input_info[context->current_input].framesize_info[i];
                            break;
                        }
                    }
                }
	}

	for(i=0;i<FRAMEGRABBER_SUPPORT_REFERSHRATE_NUM && refreshrate_mask;i++)
	{
		if(refreshrate_mask & (1<<i))
		{
			refreshrate_mask ^= (1<<i);
			idx++;
			if(idx==index)
			{
				return framegrabber_support_refreshrate[i];
			}

		}
	}

	return 0;
}

void framegrabber_start_streaming(framegrabber_handle_t handle,BOOL_T on)
{
	if(on)
	{
		framegrabber_mask_s_status(handle,FRAMEGRABBER_STATUS_V4L_START_STREAMING_BIT ,FRAMEGRABBER_STATUS_V4L_START_STREAMING_BIT);

	}else
	{
		framegrabber_mask_s_status(handle,FRAMEGRABBER_STATUS_V4L_START_STREAMING_BIT,0);
	}
}

void framegrabber_notify(framegrabber_handle_t handle,char *fromname,int notification,void *arg)
{
	framegrabber_context_t *context=framegrabber_getcontext(handle);
	if(context->interface.notify!=NULL)
	{
		context->interface.notify(handle,fromname,notification,arg);
	}
}

int framegrabber_s_flash(framegrabber_handle_t handle, int start_block, int blocks, U8_T *flash_update)
{
	int ret=0;
	framegrabber_context_t *context = framegrabber_getcontext(handle);

	if (context->interface.flash_update)
	{
		ret = context->interface.flash_update(handle, start_block, blocks, flash_update);
	}
	return 0;
}

int framegrabber_g_flash(framegrabber_handle_t handle, int start_block, int blocks, U8_T *flash_dump)
{
	int ret=0;
	framegrabber_context_t *context = framegrabber_getcontext(handle);

	if (context->interface.flash_read)
	{
		ret = context->interface.flash_read(handle, start_block, blocks, flash_dump);
	}
	return 0;
}

int framegrabber_g_hdcp_flag(framegrabber_handle_t handle)
{
	framegrabber_context_t *context = framegrabber_getcontext(handle);
    int hdcp_flag;
	hdcp_flag = context->hdcp_flag;
	
	//printk("%s hdcp_flag =%u\n",__func__,hdcp_flag);
	return hdcp_flag;
}

void framegrabber_s_hdcp_flag(framegrabber_handle_t handle, unsigned int hdcp_flag)
{
	framegrabber_context_t *context = framegrabber_getcontext(handle);
	
	context->hdcp_flag = hdcp_flag;
	
	//printk("%s hdcp_flag =%u\n",__func__,hdcp_flag);
}

int framegrabber_g_input_ITU(framegrabber_handle_t handle)
{
	framegrabber_context_t *context = framegrabber_getcontext(handle);
    int ITU_flag=0;
	ITU_flag = context->ITU_flag;
	
	//printk("%s ITU_flag =%u\n",__func__,ITU_flag);
	return ITU_flag;
}

void framegrabber_s_input_ITU(framegrabber_handle_t handle, int ITU_flag)
{
	framegrabber_context_t *context = framegrabber_getcontext(handle);
	
	context->ITU_flag = ITU_flag;
	
	//printk("%s ITU709_flag =%u\n",__func__,ITU709_flag);
}

int framegrabber_g_edid(framegrabber_handle_t handle, void* edid_buf_info_param)
{
	framegrabber_context_t *context = framegrabber_getcontext(handle);

	struct edid_data* edid_datas = (struct edid_data*)edid_buf_info_param;
	if(edid_datas == NULL)
	{
		printk("[error] : edid_datas is null!\n");
		return -1;
	}

	printk("edid_datas->size = %d\n", edid_datas->size);
	if (context->interface.edid_val_get)
	{
		if(0 != context->interface.edid_val_get(handle, edid_datas->edid, &edid_datas->size))
		{
			return -1;
		}
	}

	return 0;
}

int framegrabber_s_edid(framegrabber_handle_t handle, void* val_edid)
{
	framegrabber_context_t *context = framegrabber_getcontext(handle);
	edid_data_t* val = (edid_data_t*)val_edid;
	if((NULL == context) || (NULL == val))
	{
		return -1;
	}

	if (context->interface.edid_val_set)
	{
		if(0 != context->interface.edid_val_set(handle, val->edid, val->size))
		{
			return -1;		//edid write failed
		}
	}

	return 0;
}

