 /*
 *Driver for the card pcie bridge.

 *Copyright (C) 2019 AVerMedia TECHNOLOGIES, Inc.

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/videodev2.h>
#include <linux/v4l2-dv-timings.h>
#include <linux/gpio.h>
#include <media/v4l2-device.h>
#include <media/v4l2-ioctl.h>
#include <media/videobuf2-vmalloc.h>

#include "queue.h"
#include "typedef.h"
#include "sys.h"
#include "v4l2_model.h"
#include "v4l2_model_device.h"
#include "v4l2_model_table.h"

#include "mem_model.h"

//defaults for ADV7619
#define BrightnessDefault  0x80
#define ContrastDefault    0x80
#define SaturationDefault  0x80
#define HueDefault         0x00
//static U32_T SharpnessDefault  = 0x00;

//To report color control range, define default color range here for 
//all video decoder. 

#define MIN_VAMP_BRIGHTNESS_UNITS   0
#define MAX_VAMP_BRIGHTNESS_UNITS   255

#define MIN_VAMP_CONTRAST_UNITS     0
#define MAX_VAMP_CONTRAST_UNITS     255

#define MIN_VAMP_SATURATION_UNITS   0
#define MAX_VAMP_SATURATION_UNITS   255

#define MIN_VAMP_HUE_UNITS          0
#define MAX_VAMP_HUE_UNITS          255

#define IOCTL_MAX_AVISIZE 30

int brightness_tmp=0;
int contrast_tmp=0;
int hue_tmp=0;
int saturation_tmp=0;
int mute_tmp=0;

struct v4l2_model_framesize 
{
	int width;
	int height;
};

enum v4l2_bchs_type {
	V4L2_BCHS_TYPE_BRIGHTNESS,
	V4L2_BCHS_TYPE_CONTRAST,
	V4L2_BCHS_TYPE_HUE,
	V4L2_BCHS_TYPE_SATURATION,
};

static struct v4l2_queryctrl g_cm311h_ctrls[] = 
{
	#if 1
	{
		V4L2_CID_BRIGHTNESS,           //id
		V4L2_CTRL_TYPE_INTEGER,        //type
		"Brightness",                  //name[32]
		MIN_VAMP_BRIGHTNESS_UNITS,     //minimum
		MAX_VAMP_BRIGHTNESS_UNITS,     //maximum
        1,                             //step
		BrightnessDefault,             //default_value
		0,                             //flags
	    { 0, 0 },                      //reserved[2]
	},
	{
		V4L2_CID_CONTRAST,             //id
		V4L2_CTRL_TYPE_INTEGER,        //type
		"Contrast",                    //name[32]
		MIN_VAMP_CONTRAST_UNITS,       //minimum
		MAX_VAMP_CONTRAST_UNITS,       //maximum
        1,                             //step
		ContrastDefault,               //default_value
		0,                             //flags
	    { 0, 0 },                      //reserved[2]
	},
	{
		V4L2_CID_SATURATION,           //id
		V4L2_CTRL_TYPE_INTEGER,        //type
		"Saturation",                  //name[32]
		MIN_VAMP_SATURATION_UNITS,     //minimum
		MAX_VAMP_SATURATION_UNITS,     //maximum
        1,                             //step
		SaturationDefault,             //default_value
		0,                             //flags
	    { 0, 0 },                      //reserved[2]
	},
	{
		V4L2_CID_HUE,                  //id
		V4L2_CTRL_TYPE_INTEGER,        //type
		"Hue",                         //name[32]
		MIN_VAMP_HUE_UNITS,            //minimum
		MAX_VAMP_HUE_UNITS,            //maximum
        1,                             //step
		HueDefault,                    //default_value
		0,                             //flags
	    { 0, 0 },                      //reserved[2]
	},
	#endif
};

#define ARRAY_SIZE_OF_CTRL		(sizeof(g_cm311h_ctrls)/sizeof(g_cm311h_ctrls[0]))

static struct v4l2_queryctrl *find_ctrl(unsigned int id)
{
	int i;
	//scan supported queryctrl table
	for( i=0; i<ARRAY_SIZE_OF_CTRL; ++i )
		if( g_cm311h_ctrls[i].id==id )
			return &g_cm311h_ctrls[i];

	return 0;
}

u32 v4l2_model_to_v4l2_caps(u32 v4l2m_caps)
{
	u32 caps=0;
	int i;

	for(i=0;i<V4L2_MODEL_CAPS_MAX && v4l2m_caps;i++)
	{
		if(v4l2m_caps & (1 << i))
		{
			switch(i)
			{
				case V4L2_MODEL_CAPS_CAPTURE:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_VIDEO_CAPTURE;
					break;
				case V4L2_MODEL_CAPS_CAPTURE_MPLANE:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_VIDEO_CAPTURE_MPLANE;
					break;
				case V4L2_MODEL_CAPS_VBI_CAPTURE:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_VBI_CAPTURE;
					break;
				case V4L2_MODEL_CAPS_TUNER:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_TUNER;
					break;
				case V4L2_MODEL_CAPS_AUDIO:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_AUDIO;
					break;
				case V4L2_MODEL_CAPS_RADIO:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_RADIO;
					break;
				case V4L2_MODEL_CAPS_READWRITE:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_READWRITE;
					break;
				case V4L2_MODEL_CAPS_STREAMING:
					v4l2m_caps ^=(1 << i);
					caps |= V4L2_CAP_STREAMING;
					break;
			}
		}
	}
	if(v4l2m_caps)
	{
		//V4L2_MODEL_DEBUG("Unsupport capabilities!!!\n");	
		printk("Unsupport capabilities!!!\n");	
	}
	

	return caps;
}


int v4l2_model_ioctl_querycap(struct file *file, void *fh, struct v4l2_capability *cap)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	
    //printk("%s...\n",__func__);

	if(v4l2m_context)
	{
		strncpy(cap->driver, v4l2m_context->device_info.driver_name,sizeof(cap->driver));
#if 0		
		if (subsystem_id == 0x3112)
		    strncpy(cap->card, v4l2m_context->device_info.card_name_1,sizeof(cap->card));
		else
		    strncpy(cap->card, v4l2m_context->device_info.card_name,sizeof(cap->card));    
#endif        
        strncpy(cap->card, v4l2m_context->device_info.card_name,sizeof(cap->card));    
        sprintf(cap->bus_info, "%s", dev_name(v4l2m_context->dev));
		if(v4l2m_context->device_info.capabilities)
		{
			cap->device_caps = v4l2_model_to_v4l2_caps(v4l2m_context->device_info.capabilities) ;	
		}
		
		cap->capabilities = cap->device_caps | V4L2_CAP_DEVICE_CAPS;
	}
	return 0;
}

int v4l2_model_ioctl_enum_fmt_vid_cap(struct file *file, void *fh, struct v4l2_fmtdesc *f)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	U32_T index = f->index;

    //printk("%s...\n",__func__);
	if(f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
	{
		printk("%s.\n",__func__);
		return -EINVAL;
	}

	if(v4l2m_context)
	{
		const framegrabber_pixfmt_t *pixfmt;
		if(f->index <0)
		{
			return -EINVAL;
		}
		if(f->index >= FRAMEGRABBER_PIXFMT_MAX)
		{
			return -EINVAL;
		}
		else
		{
		    pixfmt=framegrabber_g_support_pixelfmt_by_index(v4l2m_context->framegrabber_handle,f->index);
		    //printk("%s..pixfmt=%d.\n",__func__,f->index);
		    f->index = index;
		    f->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		    strlcpy(f->description, pixfmt->name, sizeof(f->description));
		    f->pixelformat=pixfmt->fourcc;
		}
			
		if(pixfmt==NULL)
			return -EINVAL;
        //printk("%s....pixfmt=%d.\n",__func__,f->index);
		strlcpy(f->description, pixfmt->name, sizeof(f->description));
		f->pixelformat=pixfmt->fourcc;
	}

	return 0;
}

int v4l2_model_ioctl_g_fmt_vid_cap(struct file *file, void *fh,struct v4l2_format *f)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	const framegrabber_pixfmt_t *pixfmt;
	BOOL_T interlace_mode;
	int ITU_flag;

	pixfmt=framegrabber_g_out_pixelfmt(v4l2m_context->framegrabber_handle);
	interlace_mode = framegrabber_g_input_interlace(v4l2m_context->framegrabber_handle); 
	ITU_flag = framegrabber_g_input_ITU(v4l2m_context->framegrabber_handle); //get from infoframe
	if(pixfmt)
	{
		int width,height;
		//unsigned bytesperline;
        //printk("%s...\n",__func__);
		framegrabber_g_input_framesize(v4l2m_context->framegrabber_handle,&width,&height);
		//bytesperline=framegrabber_g_out_bytesperline(v4l2m_context->framegrabber_handle);
        //printk("%s..f->fmt.pix.width=%d.f->fmt.pix.height=%d.\n",__func__,f->fmt.pix.width,f->fmt.pix.height);

		f->fmt.pix.width=width;
		f->fmt.pix.height=height;
		f->fmt.pix.field=V4L2_FIELD_NONE; //Field
		f->fmt.pix.pixelformat  = pixfmt->fourcc;

		//f->fmt.pix.bytesperline = bytesperline;
		f->fmt.pix.bytesperline = (f->fmt.pix.width * pixfmt->depth) >> 3;
		f->fmt.pix.sizeimage =	f->fmt.pix.height * f->fmt.pix.bytesperline;
		
		if (ITU_flag ==2) //get from infoframe
		{
		    f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709;
		    f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_709;
		}
		else if (ITU_flag ==1)
		{
		    f->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M;  
		    f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_601;
		}
		else
		{
			f->fmt.pix.colorspace = V4L2_COLORSPACE_DEFAULT;
			f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_DEFAULT;
		}
		
		f->fmt.pix.quantization = V4L2_QUANTIZATION_LIM_RANGE;
        
        if (interlace_mode)  
        {
			f->fmt.pix.height = height*2; //for AP like VLC
			f->fmt.pix.field=V4L2_FIELD_INTERLACED;
		}
		//printk("%s....f->fmt.pix.width=%d.f->fmt.pix.height=%d.\n",__func__,f->fmt.pix.width,f->fmt.pix.height);
		return 0;
	} 

	return -EINVAL;
}

int v4l2_model_ioctl_try_fmt_vid_cap(struct file *file, void *fh, struct v4l2_format *f) //bypass mode customized 
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	const framegrabber_pixfmt_t *fmt;
	int width,height;
	BOOL_T interlace_mode;
	//int ITU_flag;
    //printk("%s>>f->fmt.pix.width=%d.f->fmt.pix.height=%d.f->fmt.pix.pixelformat=%d\n",__func__,f->fmt.pix.width,f->fmt.pix.height,f->fmt.pix.pixelformat);                  
	fmt = framegrabber_g_support_pixelfmt_by_fourcc(v4l2m_context->framegrabber_handle, f->fmt.pix.pixelformat);
        
	framegrabber_g_input_framesize(v4l2m_context->framegrabber_handle,&width,&height);
	
    interlace_mode = framegrabber_g_input_interlace(v4l2m_context->framegrabber_handle); 
    
    //ITU_flag = framegrabber_g_input_ITU(v4l2m_context->framegrabber_handle);
        
    if(!fmt)
    {
		printk("%s..\n",__func__);
        return 0;
	}

	//if(width!=0 && height!=0) 
	{
		if ((f->fmt.pix.width <320) || (f->fmt.pix.width >4096) || (f->fmt.pix.height <240) || (f->fmt.pix.height >2160))
		{
		    if (interlace_mode)
		    {
		        f->fmt.pix.width=width;
		        f->fmt.pix.height=height*2;
		        f->fmt.pix.field =V4L2_FIELD_INTERLACED; //field order
		        //printk("%s>f->fmt.pix.width=%d.f->fmt.pix.height=%d.\n",__func__,f->fmt.pix.width,f->fmt.pix.height); 
		    }
		    else
		    {
			    f->fmt.pix.width=width;
		        f->fmt.pix.height=height;
		        f->fmt.pix.field = V4L2_FIELD_NONE;
		    }
		}
		else
		{
			if ((f->fmt.pix.height == height) && interlace_mode)
			{
			    f->fmt.pix.height=height*2;  
			    f->fmt.pix.field =V4L2_FIELD_INTERLACED;
			    printk("%s f->fmt.pix.width=%d.f->fmt.pix.height=%d.\n",__func__,f->fmt.pix.width,f->fmt.pix.height);   
			}
			else
			{
				f->fmt.pix.field = V4L2_FIELD_NONE;
			}
		}
	}
	if ((f->fmt.pix.width ==0) || (f->fmt.pix.height ==0)) //protection
	{
		f->fmt.pix.width = 1920;
		f->fmt.pix.height = 1080;
	}
	
	f->fmt.pix.bytesperline = (f->fmt.pix.width * fmt->depth) >> 3;
	f->fmt.pix.sizeimage =f->fmt.pix.height * f->fmt.pix.bytesperline;
	
	if (fmt!=NULL && fmt->is_yuv)
		f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709;//V4L2_COLORSPACE_SMPTE170M;
	else
		f->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
		
#if 0        
	if (ITU_flag ==2) //get from infoframe
	{
	    f->fmt.pix.colorspace = V4L2_COLORSPACE_REC709;
		f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_709;
	}
	else if (ITU_flag ==1)
	{
		f->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M;  
		f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_601;
    }
	else
	{
		f->fmt.pix.colorspace = V4L2_COLORSPACE_DEFAULT;
		f->fmt.pix.ycbcr_enc = V4L2_YCBCR_ENC_DEFAULT;
	}
#endif    
	f->fmt.pix.priv = 0;
	//f->fmt.pix.quantization = V4L2_QUANTIZATION_LIM_RANGE;
        
    //printk("%s<<f->fmt.pix.width=%d.f->fmt.pix.height=%d.\n",__func__,f->fmt.pix.width,f->fmt.pix.height);      
	return 0;

}

int v4l2_model_ioctl_s_fmt_vid_cap(struct file *file, void *fh,struct v4l2_format *f)
{

	v4l2_model_context_t *v4l2m_context = video_drvdata(file);

	int ret = v4l2_model_ioctl_try_fmt_vid_cap(file, fh, f);
	if (ret < 0)
	{
		printk("%s..\n",__func__);
		return ret;
	}

	if (vb2_is_busy(&v4l2m_context->queue)) {
		printk("%s.\n",__func__);
	    return -EBUSY;
	}

	/* TODO check setting format is supported */

   printk("%s...\n",__func__);

   if(framegrabber_g_support_pixelfmt_by_fourcc(v4l2m_context->framegrabber_handle,f->fmt.pix.pixelformat)==NULL)
   {
	   printk("..%s..\n",__func__);
	   return -EINVAL;
   }
   if(f->fmt.pix.width==0 || f->fmt.pix.height==0)
   {
	   printk("..%s.\n",__func__);
	   return -EINVAL;
   }
   //printk("%s.f->fmt.pix.width=%d..f->fmt.pix.height=%d\n",__func__,f->fmt.pix.width,f->fmt.pix.height);
   framegrabber_s_out_framesize(v4l2m_context->framegrabber_handle,f->fmt.pix.width,f->fmt.pix.height); 
   framegrabber_s_out_pixelfmt(v4l2m_context->framegrabber_handle,f->fmt.pix.pixelformat); 

    return 0;
}

int v4l2_model_ioctl_enum_framesizes(struct file *file, void *fh, struct v4l2_frmsizeenum *fsize)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	const framegrabber_pixfmt_t *pixfmt;
	int width=0,height=0;

    //printk("%s...\n",__func__);
	pixfmt=framegrabber_g_support_pixelfmt_by_fourcc(v4l2m_context->framegrabber_handle,fsize->pixel_format);
	//printk("%s %08x %p\n",__func__,fsize->pixel_format,pixfmt);
	if(pixfmt==NULL)
	{
		printk("%s..\n",__func__);
		return -EINVAL;
	}
	else
	{
		//printk("%s %08x %x\n",__func__,fsize->pixel_format,pixfmt->fourcc);
	}

	if(framegrabber_g_supportframesize(v4l2m_context->framegrabber_handle,fsize->index,&width,&height)!=FRAMEGRABBER_OK)
	{
		//printk("%s. invalid framesize\n",__func__);
		return -EINVAL;
	}

	//printk("%s...supportframesize width=%d height=%d..\n",__func__,width,height); 
	fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
	fsize->pixel_format=pixfmt->fourcc;
	fsize->discrete.width=width;
	fsize->discrete.height=height;

	return 0;
}

int v4l2_model_ioctl_enum_input(struct file *file, void *fh, struct v4l2_input *inp)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    //printk("%s...\n",__func__);
	if (inp->index >= framegrabber_g_input_num(v4l2m_context->framegrabber_handle))
		return -EINVAL;
	inp->type = V4L2_INPUT_TYPE_CAMERA;
	sprintf(inp->name, "%s", framegrabber_g_input_name(v4l2m_context->framegrabber_handle,inp->index));
	inp->status=0;

	if(inp->index==framegrabber_g_input(v4l2m_context->framegrabber_handle))
	{
		u32 input_status=framegrabber_g_input_status(v4l2m_context->framegrabber_handle);
        //printk("%s..\n",__func__);
		if(input_status == FRAMEGRABBER_INPUT_STATUS_OK)
		{
			//printk("%s.FRAMEGRABBER_INPUT_STATUS_OK\n",__func__);
			inp->status=0;
		}else
		{
			//printk("..%s..\n",__func__);
			if(input_status & FRAMEGRABBER_INPUT_STATUS_NO_POWER)
				inp->status |= V4L2_IN_ST_NO_POWER;

			if(input_status & FRAMEGRABBER_INPUT_STATUS_NO_SIGNAL)
				inp->status |= V4L2_IN_ST_NO_SIGNAL;

			if(input_status & FRAMEGRABBER_INPUT_STATUS_NO_SYNC)
				inp->status |= V4L2_IN_ST_NO_SYNC;
		}
	}else
	{
		//printk("..%s.\n",__func__);
		inp->status |= V4L2_IN_ST_NO_POWER;
	}
	
	return 0;
}

int v4l2_model_ioctl_g_input(struct file *file, void *fh, unsigned int *i)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	*i=framegrabber_g_input(v4l2m_context->framegrabber_handle);
	//printk("%s..input=%d.\n",__func__,*i);
	return 0;
}

int v4l2_model_ioctl_s_input(struct file *file, void *fh, unsigned int i)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	//printk("%s...\n",__func__);
	if(i < 0 || i > framegrabber_g_input_num(v4l2m_context->framegrabber_handle))
	{
		printk("..%s..\n",__func__);
		return -EINVAL;
	}

	framegrabber_s_input(v4l2m_context->framegrabber_handle,i);

	return 0;
}

int v4l2_model_ioctl_enum_frameintervals(struct file *file, void *fh, struct v4l2_frmivalenum *fival)
{
	int frameinterval=0;
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    BOOL_T interlace_mode;

	
	interlace_mode = framegrabber_g_input_interlace(v4l2m_context->framegrabber_handle); 
    //printk("%s...\n",__func__);

	frameinterval=framegrabber_g_framesize_supportrefreshrate(v4l2m_context->framegrabber_handle,fival->width,fival->height,fival->index);
	if(frameinterval)
	{
		//printk("%s..\n",__func__);
		fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;
		fival->discrete.numerator=1;
		fival->discrete.denominator=frameinterval;
	}else
	{
		//printk("%s.\n",__func__);
        if(fival->index==0)
        {
			//frameinterval = framegrabber_g_out_framerate(v4l2m_context->framegrabber_handle);
			//if (frameinterval ==0)
			{
                frameinterval=framegrabber_g_input_framerate(v4l2m_context->framegrabber_handle);
			}
            fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;
		    fival->discrete.numerator=1;
		    //fival->discrete.numerator=1000;
		    fival->discrete.denominator=frameinterval;
                
        }else{
		    return -EINVAL;
        }
	}
	
	if (interlace_mode) fival->discrete.denominator /=2;
	
	
    //printk("%s.frameinterval =%d.fival->width=%d.fival->height=%d\n",__func__,frameinterval,fival->width,fival->height);
	return 0;
}

int v4l2_model_ioctl_g_parm(struct file *file, void *fh,struct v4l2_streamparm *a)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	//int current_denominator;

	if (a->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
	{
		printk("%s..\n",__func__);
	    return -EINVAL;
	}

	a->parm.capture.capability   = V4L2_CAP_TIMEPERFRAME;
	a->parm.capture.capturemode  = V4L2_MODE_HIGHQUALITY;
	a->parm.capture.timeperframe.numerator = 1000;

	a->parm.capture.readbuffers  = 1;
	
	// bypass mode only
	a->parm.capture.timeperframe.denominator = framegrabber_g_input_denominator(v4l2m_context->framegrabber_handle);
	
	return 0;
}

int v4l2_model_ioctl_s_parm(struct file *file, void *fh,struct v4l2_streamparm *a)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	U32_T io_frame_rate;
	U32_T in_frame_rate;

	io_frame_rate = a->parm.capture.timeperframe.denominator/a->parm.capture.timeperframe.numerator;
	in_frame_rate = framegrabber_g_input_framerate(v4l2m_context->framegrabber_handle);
    if(io_frame_rate  !=0)
    {
        framegrabber_s_out_framerate(v4l2m_context->framegrabber_handle,io_frame_rate);
        //printk("%s set io_framerate= %u\n", __func__, io_frame_rate);
	}
	else
	{
		framegrabber_s_out_framerate(v4l2m_context->framegrabber_handle,in_frame_rate);
		//printk("%s set in_framerate= %u\n", __func__, in_frame_rate);
	}
   
	return 0;
}

int v4l2_model_ioctl_g_ctrl(struct file *file, void *fh,struct v4l2_control *a)
{
	//v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	
    struct v4l2_control *ctrl = a;
	int ret = -EINVAL;
	int bchs_select=0;
	 
	switch( ctrl->id ) {
		case V4L2_CID_BRIGHTNESS: //0x00980900
			
		    bchs_select = V4L2_BCHS_TYPE_BRIGHTNESS;
		    //adv7619_get_bchs(v4l2m_context->adv7619_handle,&BCHSinfo,bchs_select);
		    //printk("%s...brightness(%d)\n",__func__,bchs_select);
            //ctrl->value = framegrabber_g_input_bchs(v4l2m_context->framegrabber_handle,bchs_select);
			ret = 0;
			break;

		case V4L2_CID_CONTRAST:
	
			bchs_select = V4L2_BCHS_TYPE_CONTRAST;
			
			ret = 0;
			break;

		case V4L2_CID_SATURATION:
			
			bchs_select = V4L2_BCHS_TYPE_SATURATION;
			
			ret = 0;
			break;

		case V4L2_CID_HUE:
		
			bchs_select = V4L2_BCHS_TYPE_HUE;
			
			ret = 0;
			break; 

		default:
		    ctrl->value =0;
			printk("control id %d not handled\n", ctrl->id);
		    break;	
		
	}
	#if 0
	if (ret ==0)
	{
		ctrl->value = framegrabber_g_input_bchs(v4l2m_context->framegrabber_handle,&bchs_select);	
	}
	#endif
	//printk("%s...ctrl->value(%d)=%x\n",__func__,bchs_select,ctrl->value);
	return ret;
}

int v4l2_model_ioctl_s_ctrl(struct file *file, void *fh,struct v4l2_control *a)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	
    struct v4l2_control *ctrl = a;
	struct v4l2_queryctrl *found_ctrl = find_ctrl(ctrl->id);
	int ret = -EINVAL;
	int io_bchs_value=0;
	int io_bchs_select=0;
	
	if( found_ctrl ) {
		switch( found_ctrl->type ) {
		case V4L2_CTRL_TYPE_INTEGER:
			if( ctrl->value >= found_ctrl->minimum 
				|| ctrl->value <= found_ctrl->maximum ) {
				{
					switch( ctrl->id ) {
	                case V4L2_CID_BRIGHTNESS:
		            io_bchs_value = ctrl->value;
		            io_bchs_select = V4L2_BCHS_TYPE_BRIGHTNESS;
		            break;

	                case V4L2_CID_CONTRAST:
		            io_bchs_value = ctrl->value;
		            io_bchs_select = V4L2_BCHS_TYPE_CONTRAST;
		            break;

	                case V4L2_CID_HUE:
		            io_bchs_value = ctrl->value;
		            io_bchs_select = V4L2_BCHS_TYPE_HUE;
		            break;

	                case V4L2_CID_SATURATION:
		            io_bchs_value = ctrl->value;
		            io_bchs_select = V4L2_BCHS_TYPE_SATURATION;
		            break;

	                default:
		            break;
	            }
								
			}
				framegrabber_s_input_bchs(v4l2m_context->framegrabber_handle,io_bchs_value,io_bchs_select);
				ret = 0;
			} else {
				printk("control %s out of range\n", found_ctrl->name);
			}
			break;
#if 0
		case V4L2_CTRL_TYPE_BOOLEAN:
			mute_tmp = ctrl->value;
			ret = 0;
			break;
#endif
		default:
			printk("control type %d not handled\n", found_ctrl->type);
			
		}
	}

	return ret;
}

int v4l2_model_ioctl_queryctrl(struct file *file, void *fh,struct v4l2_queryctrl *a)
{
	struct v4l2_queryctrl *ctrl = a;
	struct v4l2_queryctrl *found_ctrl = find_ctrl(ctrl->id);
	int ret = -EINVAL;

	if( 0==found_ctrl ) {
		unsigned int id = ctrl->id;
        mem_model_memset(ctrl, 0, sizeof(*ctrl));
        ctrl->id = id;
		ctrl->flags = V4L2_CTRL_FLAG_DISABLED;
	} else {
		*ctrl = *found_ctrl;
		printk("control %s\n", found_ctrl->name);
		ret = 0;
	}

	return ret;
}

int v4l2_model_ioctl_cropcap(struct file *file, void *fh,struct v4l2_cropcap *a)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	int width,height;

	if (a->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
	{
		printk("%s..\n",__func__);
		return -EINVAL;
	}
  
	framegrabber_g_input_framesize(v4l2m_context->framegrabber_handle,&width,&height);
	
	a->bounds.left = 0;
	a->bounds.top = 0;
	a->bounds.width = width;
	a->bounds.height = height;

	a->defrect = a->bounds;
	a->pixelaspect.numerator = 16;	/* 4:3 FIXME: remove magic numbers */
	a->pixelaspect.denominator = 9;


	return 0;
}

int v4l2_model_ioctl_g_aviinfoframe(struct file *file,U8_T *aviinfo)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    int ret=0;
    unsigned char ioctl_aviinfo[IOCTL_MAX_AVISIZE];

	ret = framegrabber_g_input_aviinfoframe(v4l2m_context->framegrabber_handle,ioctl_aviinfo);
		
	memcpy(aviinfo,ioctl_aviinfo,sizeof(ioctl_aviinfo)); 
	
#if 1        
    {
        int jj;
	    printk(" =====================v4l2_model_ioctl_g_aviinfoframe======================\n");
	    for (jj=0; jj< IOCTL_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            aviinfo[jj-7],aviinfo[jj-6],aviinfo[jj-5],aviinfo[jj-4],
	            aviinfo[jj-3],aviinfo[jj-2],aviinfo[jj-1],aviinfo[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif

	return ret;
}

int v4l2_model_ioctl_g_vendorinfoframe(struct file *file,U8_T *vendorinfo)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    int ret=0;
    unsigned char ioctl_vendorinfo[IOCTL_MAX_AVISIZE];

	ret = framegrabber_g_input_vendorinfoframe(v4l2m_context->framegrabber_handle,ioctl_vendorinfo);
		
	memcpy(vendorinfo,ioctl_vendorinfo,sizeof(ioctl_vendorinfo)); 
	
#if 1        
    {
        int jj;
	    printk(" =====================v4l2_model_ioctl_g_vendorinfoframe======================\n");
	    for (jj=0; jj< IOCTL_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            vendorinfo[jj-7],vendorinfo[jj-6],vendorinfo[jj-5],vendorinfo[jj-4],
	            vendorinfo[jj-3],vendorinfo[jj-2],vendorinfo[jj-1],vendorinfo[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif

	return ret;
}

int v4l2_model_ioctl_g_audioinfoframe(struct file *file,U8_T *audioinfo)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    int ret=0;
    unsigned char ioctl_audioinfo[IOCTL_MAX_AVISIZE];

	ret = framegrabber_g_input_audioinfoframe(v4l2m_context->framegrabber_handle,ioctl_audioinfo);
		
	memcpy(audioinfo,ioctl_audioinfo,sizeof(ioctl_audioinfo)); 
	
#if 1        
    {
        int jj;
	    printk(" =====================v4l2_model_ioctl_g_audioinfoframe======================\n");
	    for (jj=0; jj< IOCTL_MAX_AVISIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            audioinfo[jj-7],audioinfo[jj-6],audioinfo[jj-5],audioinfo[jj-4],
	            audioinfo[jj-3],audioinfo[jj-2],audioinfo[jj-1],audioinfo[jj]);
		    }
	    }
	    printk("\n");           
    }
#endif

	return ret;
}

int v4l2_model_ioctl_g_audiorate(struct file *file, int *audio_rate)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	
#if 1 // S_MODIFIED
	framegrabber_status_bitmask_e framegrabber_status = framegrabber_g_status(v4l2m_context->framegrabber_handle);
	if (!(framegrabber_status & FRAMEGRABBER_STATUS_V4L_START_STREAMING_BIT)) {
		const int INVALID_AUDIO_RATE_FOR_CLIENT = 96000;
		*audio_rate = INVALID_AUDIO_RATE_FOR_CLIENT;
		return 0;
	}
#endif
	
	*audio_rate=framegrabber_g_input_audiorate(v4l2m_context->framegrabber_handle);
	
	if (*audio_rate ==0)
        *audio_rate = 32000;
	else if (*audio_rate ==1)
		*audio_rate = 44100;
	else if (*audio_rate ==2)
		*audio_rate = 48000; 
    else  
		*audio_rate = 96000;  
	
	//printk("%s..audio_rate=%d.\n",__func__,*audio_rate);
	return 0;
}

int v4l2_model_ioctl_g_hdcp(struct file *file, int *hdcp_flag)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	*hdcp_flag=framegrabber_g_hdcp_flag(v4l2m_context->framegrabber_handle);
	//printk("%s..hdcp_flag=%d.\n",__func__,*hdcp_flag);
	return 0;
}

int v4l2_model_ioctl_g_ITU(struct file *file, int *ITU_flag)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	*ITU_flag=framegrabber_g_input_ITU(v4l2m_context->framegrabber_handle);
	printk("%s..ITU_flag=%d.\n",__func__,*ITU_flag);
	return 0;
}

int v4l2_model_ioctl_s_bypassmode(struct file *file,int *bypassmode)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    int ret=0;
    int bypass_disable = *bypassmode;

    printk("%s bypassmode=%d\n", __func__,*bypassmode);

	framegrabber_s_fpga_bypassmode(v4l2m_context->framegrabber_handle,bypass_disable);
	
	return ret;
}

int v4l2_model_ioctl_s_pixelformat(struct file *file,int *pixel_format)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    int ret=0;
    int pixelformat_YUYV = *pixel_format;

    //printk("%s is_YUYV=%d\n", __func__,*pixel_format);
    
    if (pixelformat_YUYV==0)
        printk("%s set pixel format to UYVY\n", __func__);
    else
         printk("%s set pixel format to YUYV\n", __func__);    

	framegrabber_s_output_pixelformat(v4l2m_context->framegrabber_handle,pixelformat_YUYV);
	
	return ret;
}

int v4l2_model_ioctl_g_flash(struct file *file,struct v4l2_dump_flash *flash_dump)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    int ret=0;

    printk("%s pad=%d, sb=%d, bs=%d\n", __func__, 
		flash_dump->pad, flash_dump->start_block, flash_dump->blocks);

	ret = framegrabber_g_flash(v4l2m_context->framegrabber_handle, flash_dump->start_block, flash_dump->blocks, flash_dump->flash);

	return ret;
}

int v4l2_model_ioctl_s_flash(struct file *file,struct v4l2_dump_flash *flash_update)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
    int ret=0;

    printk("%s pad=%d, sb=%d, bs=%d\n", __func__, 
		flash_update->pad, flash_update->start_block, flash_update->blocks);

	ret = framegrabber_s_flash(v4l2m_context->framegrabber_handle, flash_update->start_block, 
	                           flash_update->blocks, flash_update->flash);

	return ret;
}

int v4l2_model_ioctl_s_customedid(struct file *file, struct edid_data* custom_edid)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);

    if((NULL == custom_edid) || (NULL == v4l2m_context))
    {
    	return -1;
    }

	return framegrabber_s_edid(v4l2m_context->framegrabber_handle, (void*)custom_edid);
}

int v4l2_model_ioctl_g_customedid(struct file *file, struct edid_data* edid_buf_info)
{
	v4l2_model_context_t *v4l2m_context = video_drvdata(file);
	
	if((NULL == v4l2m_context) || (NULL == edid_buf_info))
	{
		return -1;
	}

	edid_buf_info->size = sizeof(edid_buf_info->edid);
	return framegrabber_g_edid(v4l2m_context->framegrabber_handle, (void*)edid_buf_info);
}

int v4l2_model_ioctl_s_dv_timings(struct file *file, void *fd,
                                struct v4l2_dv_timings *timings)
{
    return 0;
}


int v4l2_model_ioctl_g_dv_timings(struct file *file, void *fd,
                                struct v4l2_dv_timings *timings)
{
    return 0;
}

int v4l2_model_ioctl_enum_std(struct file *file, void *fd,
                                struct v4l2_standard *std)
{
    return -EINVAL;
}

int v4l2_model_ioctl_s_std(struct file *file, void *fd,
            v4l2_std_id std)
{
    return 0;
}

int v4l2_model_ioctl_g_std(struct file *file, void *fd,
            v4l2_std_id *std)
{
    return 0;
}

int v4l2_model_ioctl_querystd(struct file *file, void *fd,
            v4l2_std_id *std)
{
    return 0;
}

int v4l2_model_g_tuner(struct file *file, void *fd,
            struct v4l2_tuner *tuner)
{
    return 0;
}

int v4l2_model_s_tuner(struct file *file, void *fd,
            const struct v4l2_tuner *tuner)
{
    return 0;
}
